/* memory management */
/* We use st_mm.h package by default unless SMALLOC is defined. */
#ifdef USE_SMALLOC
#  include <smllc.h>
#endif /* SMALLOC */

/* Memory allocations */
#ifdef USE_SMALLOC
#  define Cilk_malloc smalloc
#  define Cilk_free sfree
#  ifdef PP_ONLY
#    define alloca smalloc_internal
#    define alloca_exp_free(a,b) sfree_internal(a)
#    define Cilk_internal_malloc smalloc_internal
#    define Cilk_internal_free sfree_internal
#    define Cilk_internal_malloc_fast smalloc_internal
#    define Cilk_internal_free_fast sfree_internal
#  else /* PP_ONLY */
#    define Cilk_internal_malloc smalloc
#    define Cilk_internal_free sfree
#    define Cilk_internal_malloc_fast smalloc
#    define Cilk_internal_free_fast sfree
#    include <alloca.h>
#    define Cilk_alloca alloca
#    define alloca_exp_free(a,b)
#  endif /* PP_ONLY */
#else /* USE_SMALLOC */
#  define Cilk_malloc malloc
#  define Cilk_free free
#  define Cilk_internal_malloc malloc
#  define Cilk_internal_free free
#  define Cilk_internal_malloc_fast malloc
#  define Cilk_internal_free_fast free
#  ifdef PP_ONLY
#    define Cilk_alloca malloc
#    define alloca malloc
#    define alloca_exp_free(a,b) free(a)
#  else /* PP_ONLY */
#    include <alloca.h>
#    define Cilk_alloca alloca
#    define alloca_exp_free(a,b)
#  endif /* PP_ONLY */
#endif /* USE_SMALLOC */

/* Mains */
#ifdef PP_ONLY
#  define MAIN st_main
#  define cilk_main st_main
#else /* PP_ONLY */
#  ifdef USE_SMALLOC
     int main(int argc, char *argv[]){
       smalloc_init(1);
       allc_main(argc, argv);
     }
#    define MAIN allc_main
#    define cilk_main allc_main
#  else /* USE_SMALLOC */
#    define MAIN main
#    define cilk_main main
#  endif /* USE_SMALLOC */
#endif /* PP_ONLY */

/* For Cilk_malloc without linking StackThreads runtime library */
#if defined(USE_SMALLOC)&&!defined(PP_ONLY)
void st_wg_die(void* x)
{
  exit((int)x);
}
void asm_membar_write_write()
{
#if defined(sparc)
  asm volatile ("! REAL INSN BEGINS");
  asm volatile ("retl");
  asm volatile ("membar #StoreStore");

#elif defined(__alpha)
  asm volatile ("# REAL INSN BEGINS");
  asm volatile ("wmb");
  asm volatile ("ret $31,($26),1");

#else
#error "unknown CPU that has membar"
#endif
}
#define HAVE_SWAP 1
#if HAVE_SWAP
long asm_swap_long(long volatile * l, long v)
{
#if defined(i386)
  asm volatile ("/ REAL INSN BEGINS");
  asm volatile ("movl 4(%esp),%ecx"); /* %ecx = l */
  asm volatile ("movl 8(%esp),%eax"); /* %eax = v */
  asm volatile ("xchg 0(%ecx),%eax"); /* swap l[0] and v */
  asm volatile ("ret");

#elif defined(sparc)
  asm volatile ("! REAL INSN BEGINS");
  asm volatile ("swap [%o0],%o1");
  asm volatile ("retl");
  asm volatile ("mov %o1,%o0");

#else

#error "unknown CPU"

#endif
  return 0;			/* to make gcc happy */
}
#endif /* HAVE_SWAP */
#endif /* USE_SMALLOC && !PP_ONLY */

/* For cilk_time without ST runtime */
#ifndef PP_ONLY
#  include <sys/time.h>
   /* returns current time in ms */
   long st_current_time_ms()
   {
#    if defined(sgi) || defined(sun) || defined(__linux) || defined(__osf__) || defined(__CYGWIN32__)
      struct timeval tp[1];
      gettimeofday (tp, 0);
      return tp->tv_sec * 1000 + tp->tv_usec / 1000;
#    elif defined(_WIN32)
#     error "do this on win32"
#    else
#     warning "no way to get a time"
#    endif
   }
#endif /* PP_ONLY */

#define CILK_PROC_ARGS(type)
#define CILK_PROC_ARGS0(type)
#define CILK_VOID_PROC_ARGS     
#define CILK_VOID_PROC_ARGS0    
#define CILK_INLET_PROC_ARGS(type)
#define USE_SPAWN
#define USE_SPAWN_IF_CILK
#define VAR_SPAWN(var, f, args...) \
   var  = f(## args)
#define VAR_ADD_SPAWN(var, f, args...) \
   var += f(## args)
#define VOID_SPAWN(f, args...) \
          f(## args)
#define VAR_SPAWN0(var, f, args...) \
   var  = f(## args)
#define VOID_SPAWN0(f, args...) \
          f(## args)
#define VOID_SPAWN_IF_CILK(f, args...) \
          f(## args)
#define VOID_SPAWN0_IF_CILK(f, args...) \
          f(## args)
#define VAR_SPAWN_IF_CILK(var, f, args...) \
          var = f(## args)
#define SYNC_IF_CILK /* hack to update timers */
#define LAST_SYNC_IF_CILK /* hack to update timers */
#define ST_POLLING()        
#define SYNC 
#define LAST_SYNC
#define RETURN(x) return x
#define ATOMIC_ADD_RETURN(x) return x
#define VOID_RETURN return

#define CILK_ASSERT assert       
#define Cilk_active_size 1
#define Cilk_die(a) do { fprintf(stderr,a); exit(-1); } while(0)
#define Cilk_lockvar long volatile
#define Cilk_lock(a)
#define Cilk_unlock(a)
#define Cilk_lock_init(a)
#define SYNCHED 1 /* always synched */

/* commonly used macros */
#define cilk
typedef long Cilk_time;
#define Cilk_get_wall_time() st_current_time_ms()
#define Cilk_get_time() st_current_time_ms() /* used by cilksort */
#define Cilk_wall_time_to_sec(e) (((e)+0.0)/1000.0)
#define Cilk_time_to_sec(e) (((e)+0.0)/1000.0)

/* Perhaps not implemented in STHREADS */
#define Cilk_user_work 0          /* Current number of generated works */
#define Cilk_user_critical_path 0 /* Current number of critical paths  */

