/* for high performance SYNC */
#  define THREAD_STOLEN_HOOK() __i__->waiter_stolen = 1

/* include files for parallel Sthreads */
#  include <st.h>
#  include <cilkcnt.h>

/* memory management */
/* We use st_mm.h package by default unless SMALLOC is defined. */
#ifdef USE_SMALLOC
#  include <smllc.h>
#else /* USE_SMALLOC */
/* use st_mm.h package by default */
#  ifndef UNDERLYING_MM
#    define UNDERLYING_MM UNDERLYING_MM_SGC_EXP_FREE
#  endif /* !defined(UNDERLYING_MM) */
#  ifndef RALLOC
#    define RALLOC 1
#  endif /* !defined(RALLOC) */
#  include <st_mm.h>
#endif /* WITH_SGC */

/* commonly used macros */
#  define cilk
typedef long Cilk_time;
#  define Cilk_get_wall_time() st_current_time_ms()
#  define Cilk_get_time() st_current_time_ms() /* used by cilksort */
#  define Cilk_wall_time_to_sec(e) (((e)+0.0)/1000.0)
#  define Cilk_time_to_sec(e) (((e)+0.0)/1000.0)

/* Perhaps not implemented in STHREADS */
#  define Cilk_user_work 0          /* Current number of generated works */
#  define Cilk_user_critical_path 0 /* Current number of critical paths  */

/* extra arg to pass a pointer of the variable which hold return value,
   and to pass a pointer of sync counter */
#  define CILK_PROC_ARGS(type)    , csc_t __o__, type * __r__
#  define CILK_PROC_ARGS0(type)     csc_t __o__, type * __r__
#  define CILK_VOID_PROC_ARGS     , csc_t __o__
#  define CILK_VOID_PROC_ARGS0      csc_t __o__
#  define CILK_INLET_PROC_ARGS(type) , csc_t __o__, type * __r__

/* spawns */
#  define USE_SPAWN int use_spawn; csc __i__[1]; inl_init_csc(__i__, 0)
#  define USE_SPAWN_IF_CILK\
     int use_spawn; csc __i__[1]; __i__->waiter_stolen = 0;

#  define VAR_SPAWN(var, f, args...) \
     do { use_spawn; inl_csc_spawn(__i__);\
     ST_THREAD_CREATE(f(## args, __i__, &var)); } while(0)
#  define VAR_ADD_SPAWN(var, f, args...) \
     do { use_spawn; inl_csc_spawn(__i__);\
     ST_THREAD_CREATE(f(## args, __i__, &var)); } while(0)
#  define VOID_SPAWN(f, args...) \
     do { use_spawn; inl_csc_spawn(__i__);\
     ST_THREAD_CREATE(f(## args, __i__)); } while(0)
#  define VAR_SPAWN0(var, f, args...) \
     do { use_spawn; inl_csc_spawn(__i__);\
     ST_THREAD_CREATE(f(## args __i__, &var)); } while(0)
#  define VOID_SPAWN0(f, args...) \
     do { use_spawn; inl_csc_spawn(__i__);\
     ST_THREAD_CREATE(f(## args  __i__)); } while(0)
#  define VOID_SPAWN0_IF_CILK(f, args...) \
       f(## args  __i__)
#  define VOID_SPAWN_IF_CILK(f, args...) \
       f(## args,__i__)
#  define VAR_SPAWN_IF_CILK(var, f, args...) \
        f(## args,__i__, &var)
#  define VAR_ADD_SPAWN_IF_CILK(var, f, args...) \
       f(## args,__i__, &var)

/* sync */
#  define SYNC_IF_CILK inl_init_csc(__i__, 0)/* hack to update timers */
#  define LAST_SYNC_IF_CILK /* hack to update timers */
#  define SYNC  \
    do { use_spawn; inl_csc_wait(__i__); inl_init_csc(__i__, 0); } while(0)
#  define LAST_SYNC inl_csc_wait(__i__)

/* return */
#  define RETURN(x)   do { *__r__ = x; inl_csc_finish(__o__); return;} while(0)
#  define ATOMIC_ADD_RETURN(x, TYPE)\
do { inl_csc_add_finish(__o__, __r__, x, TYPE); return; } while(0)
#  define VOID_RETURN do{ inl_csc_finish(__o__); return;}while(0)

/* main wrappers */
#  define MAIN st_main
#  define cilk_main st_main

/* other primitives */
#  undef alloca  
#  define CILK_ASSERT st_assert       
#  ifdef USE_SMALLOC
#    define Cilk_malloc smalloc
#    define Cilk_free sfree
#    define Cilk_alloca smalloc_internal
#    define alloca smalloc_internal
#    define alloca_exp_free(a,b) sfree_internal(a)
#    define Cilk_internal_malloc smalloc_internal
#    define Cilk_internal_free(a,b) sfree_internal(a)
#    define Cilk_internal_malloc_fast smalloc_internal
#    define Cilk_internal_free_fast(a,b) sfree_internal(a)
#  else /* USE_SMALLOC */
  /* use st_mm.h package by default */
#    define Cilk_malloc MM_ALLOC
#    define Cilk_free MM_FREE
#    define Cilk_internal_malloc MM_ALLOC
#    define Cilk_internal_free MM_FREE
#    define Cilk_internal_malloc_fast MM_ALLOC
#    define Cilk_internal_free_fast MM_FREE
#    define Cilk_alloca MM_ALLOC
#    define alloca MM_ALLOC
#    define alloca_exp_free(a, b) MM_FREE(a)
#  endif  
#  define Cilk_active_size tls(gopts)->n_workers
#  define Cilk_die(a) do { fprintf(stderr,a); st_wg_die((void*)(-1)); } while(0)
#  define Cilk_lockvar long volatile
#  define Cilk_lock spin_lock
#  define Cilk_unlock(a) do{ MEMBAR_WRITE_WRITE();*(a)=0; }while(0)
#  define Cilk_lock_init(a) *(a)=0;
#  define SYNCHED \
  ( __i__->waiter_stolen==0 || __i__->n_spawned==__i__->n_finished.v )

#  define shared 
#  define private 

#ifdef sparc
#  define CACHE_LINE 64
#elif defined(mips)
#  define CACHE_LINE 128
#endif

/* padding, to be used in structures to avoid false sharing */
#define CACHE_LINE_ALIGNED(decl) decl
#define Self tls(thread_id)
