
#include <sys/time.h>

#define N 256

typedef double matrix[N][N];

matrix gA;
matrix gB;
matrix gC;

void init_mult(matrix m, double x)
{
  int i, j;
  for (i = 0; i < N; i++) {
    for (j = 0; j < N; j++) {
      m[i][j] = x;
    }
  }
}

void check_mult(matrix m, double x)
{
  int i, j;
  for (i = 0; i < N; i++) {
    for (j = 0; j < N; j++) {
      if (m[i][j] != x) {
	printf("m[%d][%d] = %f != %f\n", i, j, m[i][j], x);
	exit(1);
      }
    }
  }
}

void mat_mult(matrix A, matrix B, matrix C)
{
  int i, j, k;
  printf("mat_mult\n");

  parallel for (i = 0; i < N; i++) {
    parallel for (j = 0; j < N; j++) {
      double s = 0.0;
      ST_POLLING();
      for (k = 0; k < N; k++) {
	s += A[i][k] * B[k][j];
      }
      C[i][j] = s;
    }
  }
}

int main(int argc, char ** argv)
{
  struct timeval tp[2];
  int ms;

  init_mult(gA, 1.0);
  init_mult(gB, 1.0);
  gettimeofday(tp, 0);
  mat_mult(gA, gB, gC);
  gettimeofday(tp + 1, 0);
  check_mult(gC, (double)N);
  
  ms = (tp[1].tv_sec - tp[0].tv_sec) * 1000 
    + (tp[1].tv_usec - tp[0].tv_usec) / 1000;

  printf("%d ms for %dx%d mat mult\n", ms, N, N);
}
