/*
 * Copyright (c) 1998 by Toshio Endo, Kenjiro Taura. All rights reserved.
 * Copyright (c) 1998 by Mitsubishi Research Institute.  All rights reserved.
 * Copyright (c) 1998 by Information-technology Promotion Agency.  All rights reserved.
 *
 * THIS MATERIAL IS PROVIDED AS IS, WITH ABSOLUTELY NO WARRANTY EXPRESSED
 * OR IMPLIED.  ANY USE IS AT YOUR OWN RISK.
 *
 * Permission is hereby granted to use or copy this program
 * for any purpose,  provided the above notices are retained on all copies.
 * Permission to modify the code and to distribute modified code is granted,
 * provided the above notices are retained, and a notice that the code was
 * modified is included with the above copyright notice.
 *
 */
/* Endo, February 2, 1999 */

#include <stdio.h>
#include "gc_priv.h"
#include <sys/types.h>

/********** Blocking synchronization operation. ************/
/* The functions provided by this file are more efficient */
/* than spin-wait, if the number of processes is larger than */
/* that of actual processors. */

/* #define BS_DEBUGPRINT */

#if defined SGC_FORK_THREAD || defined SGC_SPROC_THREAD

#if 1
/* WARNING: GC_fork_bsbar_XXX has bug!!! */

#include <sys/ipc.h>
#include <sys/sem.h>

#if defined LINUX || defined IRIX5
#  define SEMUN_ALREADY_DEFINED
#endif

#ifndef SEMUN_ALREADY_DEFINED
union semun {
    int val;
    struct semid_ds *buf;
    ushort *array;
};
#endif

/************************************************ module blocksync */
/* must be called by only once. */
int GC_fork_blocksync_init(sh_blocksync_t *bsp, int maxsem, int maxbar)
{
    int semid;
    int n;
    int i;
    int nsem;
    sh_semgroup_t *sgp;

    /* how many semaphores can we allocate at once? */
    nsem = 256;
    for (nsem = 256; nsem >= 1; nsem = nsem/2) {
	semid = semget(IPC_PRIVATE, nsem, IPC_CREAT | 0600);
	if (semid >= 0) {
	    break;
	}
    };

    if (nsem < 1) {
	GC_err_printf0("Error: Failed to semget.\n");
	return -1;
    }

    bsp->nsem_per_group = nsem;

    sgp = (sh_semgroup_t *)GC_scratch_g_alloc(sizeof(sh_semgroup_t));
    sgp->next = NULL;
    sgp->semid = semid;
    sgp->used_sem_num =
	(char *)GC_scratch_g_alloc(sizeof(char)*bsp->nsem_per_group);
    BZERO(sgp->used_sem_num, sizeof(char)*bsp->nsem_per_group);

    bsp->semgroup_list = sgp;
    bsp->nsem = 0;
    bsp->ngroup = 1;

#ifdef BS_DEBUGPRINT
    GC_err_printf1("[GC_blocksync_init]: succeeded, nsems-per-group is %d\n",
		   bsp->nsem_per_group);
#endif

    return 0;
}

/* must be called by only once. */
void GC_fork_blocksync_end(sh_blocksync_t *bsp)
{
    sh_semgroup_t *sgp;
    union semun dummy;
    dummy.val = 0;

    /* remove all `semid'. */
    /* this operation is necessary, because semaphores are NOT removed */
    /* automatically even if the creator process exits. */
    for (sgp = bsp->semgroup_list; sgp; sgp = sgp->next) {
	semctl(sgp->semid, 0, IPC_RMID, dummy);
    }
    bsp->nsem = 0;
    bsp->ngroup = 0;

    return;
}

/********************** auxiliary functions */
static int sem_setval(sh_bssem_t *semp, int val)
{
    union semun arg;
    arg.val = val;
    if (semctl(semp->sgp->semid, semp->sem_num, SETVAL, arg) < 0) {
	GC_err_printf0("ERROR: Failed to semctl(...SETVAL...)\n");
	return -1;
    }
    return 0;
}

/******************************* Blocking Semaphore */
int GC_fork_bssem_init(sh_bssem_t *semp, sh_blocksync_t *bsp)
{
    sh_semgroup_t *sgp;
    int i;

    if (bsp->nsem >= bsp->ngroup * bsp->nsem_per_group) {
	/* now new semgroup is required */
	int semid;
	semid = semget(IPC_PRIVATE, bsp->nsem_per_group, IPC_CREAT | 0600);

	if (semid < 0) {
	    GC_err_printf1("GC_bssem_init Error: Failed to semget %d.\n",
			   bsp->nsem_per_group);
	    return -1;
	}
	else {
#ifdef BS_DEBUGPRINT
	    GC_err_printf1("GC_bssem_init: Succeeded to semget %d.\n",
			   bsp->nsem_per_group);
#endif
	}
	sgp = (sh_semgroup_t *)GC_scratch_g_alloc(sizeof(sh_semgroup_t));
	sgp->next = bsp->semgroup_list;
	sgp->semid = semid;
	sgp->used_sem_num =
	    (char *)GC_scratch_g_alloc(sizeof(char)*bsp->nsem_per_group);
	BZERO(sgp->used_sem_num, sizeof(char)*bsp->nsem_per_group);
	
	bsp->semgroup_list = sgp;
	bsp->ngroup++;
    }

    for (sgp = bsp->semgroup_list; sgp; sgp = sgp->next) {
	/* allocate new sem_num */
	for (i = 0; i < bsp->nsem_per_group; i++) {
	    if (sgp->used_sem_num[i] == 0) {
		sgp->used_sem_num[i] = 1;
		bsp->nsem++;
		break;
	    }
	}
	if (i < bsp->nsem_per_group) {
	    /* found */
	    break;
	}
    }

    if (sgp == NULL) {
	GC_err_printf0("ERROR: GC_bssem_init failed to allocate semaphore.\n");
	return -1;
    }

    semp->bsp = bsp;
    semp->sgp = sgp;
    semp->sem_num = i;

    /* initial value */
    sem_setval(semp, 1);

#if 0
    {
	char buf[64];
	sprintf(buf, "BSSEM_INIT: allocated sem_num %d\n", i);
	GC_err_printf0(buf);
    }
#endif

    return 0;
}

void GC_fork_bssem_end(sh_bssem_t *semp)
{
    sh_blocksync_t *bsp = semp->bsp;

    /* free sem_num */
    semp->sgp->used_sem_num[semp->sem_num] = 0;
    bsp->nsem--;

    return;
}

void GC_fork_bssem_clear(sh_bssem_t *semp)
{
    /* initial value */
    sem_setval(semp, 1);
#if 0
    sh_blocksync_t *bsp = semp->bsp;
    int semid = bsp->semid;
    struct sembuf sop;
    sop.sem_num = semp->sem_num;
    sop.sem_op = 1;
    sop.sem_flg = 0;
    semop(semid, &sop, 1);
#endif
    return;
}

void GC_fork_bssem_signal(sh_bssem_t *semp)
{
    sh_blocksync_t *bsp = semp->bsp;
    int semid = semp->sgp->semid;
    struct sembuf sop;
#ifdef BS_DEBUGPRINT
    {
	char buf[64];
	sprintf(buf, "BSSEM_SIGNAL: signal to sem_num %d\n", semp->sem_num);
	fprintf(stderr, buf);
    }
#endif
    sop.sem_num = semp->sem_num;
    sop.sem_op = -1;
    sop.sem_flg = 0;
    semop(semid, &sop, 1);
    return;
}

void GC_fork_bssem_wait(sh_bssem_t *semp)
{
    sh_blocksync_t *bsp = semp->bsp;
    int semid = semp->sgp->semid;
    struct sembuf sop;
#ifdef BS_DEBUGPRINT
    {
	char buf[64];
	sprintf(buf, "BSSEM_Wait: wait sem_num... %d\n", semp->sem_num);
	fprintf(stderr, buf);
    }
#endif
    /* Wait */
    sop.sem_num = semp->sem_num;
    sop.sem_op = 0;
    sop.sem_flg = 0;
    semop(semid, &sop, 1);
    /* Setup for next signal */
    sop.sem_num = semp->sem_num;
    sop.sem_op = 1;
    sop.sem_flg = 0;
    semop(semid, &sop, 1);
#ifdef BS_DEBUGPRINT
    {
	char buf[64];
	sprintf(buf, "BSSEM_Wait: Awaken sem_num %d\n", semp->sem_num);
	fprintf(stderr, buf);
    }
#endif
    return;
}

/******************************* Blocking Barrier */
int GC_fork_bsbar_init(sh_bsbar_t *barp, sh_blocksync_t *bsp, int np)
{
    int i;

    barp->bsp = bsp;
    barp->np = np;
    barp->sense = 0;
    GC_bssem_init(&barp->esems[0], bsp);
    GC_bssem_init(&barp->lsems[0], bsp);
    GC_bssem_init(&barp->esems[1], bsp);
    GC_bssem_init(&barp->lsems[1], bsp);

    sem_setval(&barp->esems[0], np);
    sem_setval(&barp->lsems[0], 1);
    sem_setval(&barp->esems[1], np);
    sem_setval(&barp->lsems[1], 1);

    return 0;
}

void GC_fork_bsbar_sync(sh_bsbar_t *barp, int isleader)
{
    sh_blocksync_t *bsp = barp->bsp;
    struct sembuf sop;
    int local_sense = barp->sense;

#ifdef BS_DEBUGPRINT
    fprintf(stderr, "*** START BARRIER \n");
    {
	int semval;
	union semun dummy;
	char buf[64];
	semval = semctl(barp->esems[local_sense].sgp->semid,
			barp->esems[local_sense].sem_num, GETVAL, dummy);
	sprintf(buf, "semval is %d\n", semval);
	fprintf(stderr, buf);
    }
#endif

    /* Awake leader */
    sop.sem_num = barp->esems[local_sense].sem_num;
    sop.sem_op = -1;
    sop.sem_flg = 0;
    semop(barp->esems[local_sense].sgp->semid, &sop, 1);
    
    if (isleader) {
	int i;
	int np = barp->np;
	int new_sense = 1-local_sense;

	/* wait for all */
	sop.sem_num = barp->esems[local_sense].sem_num;
	sop.sem_op = 0;
	sop.sem_flg = 0;
	semop(barp->esems[local_sense].sgp->semid, &sop, 1);

	/* Awake all */
	sop.sem_num = barp->lsems[local_sense].sem_num;
	sop.sem_op = -1;
	sop.sem_flg = 0;
	semop(barp->lsems[local_sense].sgp->semid, &sop, 1);
    }

    /* Wait for leader */
    sop.sem_num = barp->lsems[local_sense].sem_num;
    sop.sem_op = 0;
    sop.sem_flg = 0;
    semop(barp->lsems[local_sense].sgp->semid, &sop, 1);

#ifdef BS_DEBUGPRINT
    fprintf(stderr, "*** END BARRIER \n");
#endif

    return; 
}

void GC_fork_bsbar_end(sh_bsbar_t *barp)
{
    GC_bssem_end(&barp->esems[0]);
    GC_bssem_end(&barp->lsems[0]);
    GC_bssem_end(&barp->esems[1]);
    GC_bssem_end(&barp->lsems[1]);
    return;
}

#else

/* currently process version uses pthread facility. */

int GC_pthread_blocksync_init(sh_blocksync_t *bsp, int maxsem, int maxbar);
void GC_pthread_blocksync_end(sh_blocksync_t *bsp);
int GC_pthread_bssem_init(sh_bssem_t *semp, sh_blocksync_t *bsp);
void GC_pthread_bssem_end(sh_bssem_t *semp);
void GC_pthread_bssem_clear(sh_bssem_t *semp);
void GC_pthread_bssem_signal(sh_bssem_t *semp);
void GC_pthread_bssem_wait(sh_bssem_t *semp);
int GC_pthread_bsbar_init(sh_bsbar_t *barp, sh_blocksync_t *bsp, int np);
void GC_pthread_bsbar_sync(sh_bsbar_t *barp, int isleader);
void GC_pthread_bsbar_end(sh_bsbar_t *barp);

int GC_fork_blocksync_init(sh_blocksync_t *bsp, int maxsem, int maxbar)
{
    return GC_pthread_blocksync_init(bsp, maxsem, maxbar);
}

void GC_fork_blocksync_end(sh_blocksync_t *bsp)
{
    GC_pthread_blocksync_end(bsp);
}

int GC_fork_bssem_init(sh_bssem_t *semp, sh_blocksync_t *bsp)
{
    return GC_pthread_bssem_init(semp, bsp);
}

void GC_fork_bssem_end(sh_bssem_t *semp)
{
    GC_pthread_bssem_end(semp);
}

void GC_fork_bssem_clear(sh_bssem_t *semp)
{
    GC_pthread_bssem_clear(semp);
}

void GC_fork_bssem_signal(sh_bssem_t *semp)
{
    GC_pthread_bssem_signal(semp);
}

void GC_fork_bssem_wait(sh_bssem_t *semp)
{
    GC_pthread_bssem_wait(semp);
}

int GC_fork_bsbar_init(sh_bsbar_t *barp, sh_blocksync_t *bsp, int np)
{
    return GC_pthread_bsbar_init(barp, bsp, np);
}

void GC_fork_bsbar_sync(sh_bsbar_t *barp, int isleader)
{
    GC_pthread_bsbar_sync(barp, isleader);
}

void GC_fork_bsbar_end(sh_bsbar_t *barp)
{
    GC_pthread_bsbar_end(barp);
}
#endif /* 0 */

#endif /* defined defined SGC_FORK_THREAD || defined SGC_SPROC_THREAD */

/*******************************************************************/

#if defined SGC_SOLARIS_THREAD || \
    defined SGC_SOLARIS_THREAD_PTHREAD

#include <synch.h>

/************************************************ module blocksync */
/* must be called by only once. */
int GC_solaris_thread_blocksync_init(sh_blocksync_t *bsp, int maxsem, int maxbar)
{
    return 0;
}

/* must be called by only once. */
void GC_solaris_thread_blocksync_end(sh_blocksync_t *bsp)
{
    return;
}

/******************************* Blocking Semaphore */
int GC_solaris_thread_bssem_init(sh_bssem_t *semp, sh_blocksync_t *bsp)
{
#ifndef SDSM_TEST
    mutex_init(&semp->st_ml, USYNC_THREAD, NULL);
    mutex_lock(&semp->st_ml);
#else
    S_mutex_init(&semp->sdsm_ml);
    S_mutex_lock(&semp->sdsm_ml);
#endif
    return 0;
}

void GC_solaris_thread_bssem_end(sh_bssem_t *semp)
{
#ifndef SDSM_TEST
    mutex_destroy(&semp->st_ml);
#else
    S_mutex_destroy(&semp->sdsm_ml);
#endif
    return;
}

void GC_solaris_thread_bssem_clear(sh_bssem_t *semp)
{
    return;
}

void GC_solaris_thread_bssem_signal(sh_bssem_t *semp)
{
#ifndef SDSM_TEST
    mutex_unlock(&semp->st_ml);
#else
    S_mutex_unlock(&semp->sdsm_ml);
#endif
    return;
}

void GC_solaris_thread_bssem_wait(sh_bssem_t *semp)
{
#ifndef SDSM_TEST
    mutex_lock(&semp->st_ml);
#else
    S_mutex_lock(&semp->sdsm_ml);
#endif
    return;
}

/******************************* Blocking Barrier */
int GC_solaris_thread_bsbar_init(sh_bsbar_t *barp, sh_blocksync_t *bsp, int np)
{
    barp->np = np;
    barp->count = np;
    barp->sense = 0;
    cond_init(&barp->st_cv, USYNC_THREAD, NULL);
    mutex_init(&barp->st_ml, USYNC_THREAD, NULL);
#ifdef SDSM_TEST
    GC_err_printf0("currently SDSM_TEST doesn't work with blocking barrier!");
#endif

    return 0;
}

void GC_solaris_thread_bsbar_sync(sh_bsbar_t *barp, int isleader)
{
    int local_sense;

#ifdef BS_DEBUGPRINT
    fprintf(stderr, "*** START BARRIER \n");
#endif

    mutex_lock(&barp->st_ml);
    local_sense = barp->sense;
    barp->count--;

    if (barp->count == 0) {
	/* last thread */
	barp->count = barp->np;
	barp->sense = 1-local_sense;
	cond_broadcast(&barp->st_cv);
    }
    else {
	while (barp->sense == local_sense) {
	    cond_wait(&barp->st_cv, &barp->st_ml);
	}
    }

    mutex_unlock(&barp->st_ml);

#ifdef BS_DEBUGPRINT
    fprintf(stderr, "*** END BARRIER \n");
#endif

    return; 
}

void GC_solaris_thread_bsbar_end(sh_bsbar_t *barp)
{
    cond_destroy(&barp->st_cv);
    mutex_destroy(&barp->st_ml);
    return;
}

#endif /* SOLARIS_THREAD */

/*******************************************************************/

#if defined SGC_PTHREAD || \
    defined SGC_SOLARIS_THREAD_PTHREAD

/************************************************ module blocksync */
/* must be called by only once. */
int GC_pthread_blocksync_init(sh_blocksync_t *bsp, int maxsem, int maxbar)
{
    return 0;
}

/* must be called by only once. */
void GC_pthread_blocksync_end(sh_blocksync_t *bsp)
{
    return;
}

/******************************* Blocking Semaphore */
int GC_pthread_bssem_init(sh_bssem_t *semp, sh_blocksync_t *bsp)
{
#ifndef SDSM_TEST
    pthread_mutex_init(&semp->pt_ml, NULL);
    pthread_mutex_lock(&semp->pt_ml);
#else
    S_mutex_init(&semp->sdsm_ml);
    S_mutex_lock(&semp->sdsm_ml);
#endif
    return 0;
}

void GC_pthread_bssem_end(sh_bssem_t *semp)
{
#ifndef SDSM_TEST
    pthread_mutex_destroy(&semp->pt_ml);
#else
    S_mutex_destroy(&semp->sdsm_ml);
#endif
    return;
}

void GC_pthread_bssem_clear(sh_bssem_t *semp)
{
    return;
}

void GC_pthread_bssem_signal(sh_bssem_t *semp)
{
#ifndef SDSM_TEST
    pthread_mutex_unlock(&semp->pt_ml);
#else
    S_mutex_unlock(&semp->sdsm_ml);
#endif
    return;
}

void GC_pthread_bssem_wait(sh_bssem_t *semp)
{
#ifndef SDSM_TEST
    pthread_mutex_lock(&semp->pt_ml);
#else
    S_mutex_lock(&semp->sdsm_ml);
#endif
    return;
}

/******************************* Blocking Barrier */
int GC_pthread_bsbar_init(sh_bsbar_t *barp, sh_blocksync_t *bsp, int np)
{
    barp->np = np;
    barp->count = np;
    barp->sense = 0;
    pthread_cond_init(&barp->pt_cv, NULL);
    pthread_mutex_init(&barp->pt_ml, NULL);
#ifdef SDSM_TEST
    GC_err_printf0("currently SDSM_TEST doesn't work with blocking barrier!");
#endif

    return 0;
}

void GC_pthread_bsbar_sync(sh_bsbar_t *barp, int isleader)
{
    int local_sense;

    pthread_mutex_lock(&barp->pt_ml);
    local_sense = barp->sense;
    barp->count--;

    if (barp->count == 0) {
	/* last thread */
	barp->count = barp->np;
	barp->sense = 1-local_sense;
	pthread_cond_broadcast(&barp->pt_cv);
    }
    else {
	while (barp->sense == local_sense) {
	    pthread_cond_wait(&barp->pt_cv, &barp->pt_ml);
	}
    }

    pthread_mutex_unlock(&barp->pt_ml);

    return; 
}

void GC_pthread_bsbar_end(sh_bsbar_t *barp)
{
    pthread_cond_destroy(&barp->pt_cv);
    pthread_mutex_destroy(&barp->pt_ml);
    return;
}

#endif /* PTHREAD */

int GC_blocksync_init(sh_blocksync_t *bsp, int maxsem, int maxbar)
{
#if defined SGC_FORK_THREAD || defined SGC_SPROC_THREAD
    return GC_fork_blocksync_init(bsp, maxsem, maxbar);
#endif
#if defined SGC_SOLARIS_THREAD || \
    defined SGC_SOLARIS_THREAD_PTHREAD
    return GC_solaris_thread_blocksync_init(bsp, maxsem, maxbar);
#endif
#if defined SGC_PTHREAD
    return GC_pthread_blocksync_init(bsp, maxsem, maxbar);
#endif
}

void GC_blocksync_end(sh_blocksync_t *bsp)
{
#if defined SGC_FORK_THREAD || defined SGC_SPROC_THREAD
    GC_fork_blocksync_end(bsp);
    return;
#endif
#if defined SGC_SOLARIS_THREAD || \
    defined SGC_SOLARIS_THREAD_PTHREAD
    GC_solaris_thread_blocksync_end(bsp);
    return;
#endif
#if defined SGC_PTHREAD
    GC_pthread_blocksync_end(bsp);
    return;
#endif
}

int GC_bssem_init(sh_bssem_t *semp, sh_blocksync_t *bsp)
{
#if defined SGC_FORK_THREAD || defined SGC_SPROC_THREAD
    return GC_fork_bssem_init(semp, bsp);
#endif
#if defined SGC_SOLARIS_THREAD || \
    defined SGC_SOLARIS_THREAD_PTHREAD
    return GC_solaris_thread_bssem_init(semp, bsp);
#endif
#if defined SGC_PTHREAD
    return GC_pthread_bssem_init(semp, bsp);
#endif
}

void GC_bssem_end(sh_bssem_t *semp)
{
#if defined SGC_FORK_THREAD || defined SGC_SPROC_THREAD
    GC_fork_bssem_end(semp);
    return;
#endif
#if defined SGC_SOLARIS_THREAD || \
    defined SGC_SOLARIS_THREAD_PTHREAD
    GC_solaris_thread_bssem_end(semp);
    return;
#endif
#if defined SGC_PTHREAD
    GC_pthread_bssem_end(semp);
    return;
#endif
}

void GC_bssem_clear(sh_bssem_t *semp)
{
#if defined SGC_FORK_THREAD || defined SGC_SPROC_THREAD
    GC_fork_bssem_clear(semp);
    return;
#endif
#if defined SGC_SOLARIS_THREAD || \
    defined SGC_SOLARIS_THREAD_PTHREAD
    GC_solaris_thread_bssem_clear(semp);
    return;
#endif
#if defined SGC_PTHREAD
    GC_pthread_bssem_clear(semp);
    return;
#endif
}

void GC_bssem_signal(sh_bssem_t *semp)
{
#if defined SGC_FORK_THREAD || defined SGC_SPROC_THREAD
    GC_fork_bssem_signal(semp);
    return;
#endif
#if defined SGC_SOLARIS_THREAD || \
    defined SGC_SOLARIS_THREAD_PTHREAD
    GC_solaris_thread_bssem_signal(semp);
    return;
#endif
#if defined SGC_PTHREAD
    GC_pthread_bssem_signal(semp);
    return;
#endif
}

void GC_bssem_wait(sh_bssem_t *semp)
{
#if defined SGC_FORK_THREAD || defined SGC_SPROC_THREAD
    GC_fork_bssem_wait(semp);
    return;
#endif
#if defined SGC_SOLARIS_THREAD || \
    defined SGC_SOLARIS_THREAD_PTHREAD
    GC_solaris_thread_bssem_wait(semp);
    return;
#endif
#if defined SGC_PTHREAD
    GC_pthread_bssem_wait(semp);
    return;
#endif
}

int GC_bsbar_init(sh_bsbar_t *barp, sh_blocksync_t *bsp, int np)
{
#if defined SGC_FORK_THREAD || defined SGC_SPROC_THREAD
    return GC_fork_bsbar_init(barp, bsp, np);
#endif
#if defined SGC_SOLARIS_THREAD || \
    defined SGC_SOLARIS_THREAD_PTHREAD
    return GC_solaris_thread_bsbar_init(barp, bsp, np);
#endif
#if defined SGC_PTHREAD
    return GC_pthread_bsbar_init(barp, bsp, np);
#endif
}

void GC_bsbar_sync(sh_bsbar_t *barp, int isleader)
{
#if defined SGC_FORK_THREAD || defined SGC_SPROC_THREAD
    GC_fork_bsbar_sync(barp, isleader);
    return;
#endif
#if defined SGC_SOLARIS_THREAD || \
    defined SGC_SOLARIS_THREAD_PTHREAD
    GC_solaris_thread_bsbar_sync(barp, isleader);
    return;
#endif
#if defined SGC_PTHREAD
    GC_pthread_bsbar_sync(barp, isleader);
    return;
#endif
}

void GC_bsbar_end(sh_bsbar_t *barp)
{
#if defined SGC_FORK_THREAD || defined SGC_SPROC_THREAD
    GC_fork_bsbar_end(barp);
    return;
#endif
#if defined SGC_SOLARIS_THREAD || \
    defined SGC_SOLARIS_THREAD_PTHREAD
    GC_solaris_thread_bsbar_end(barp);
    return;
#endif
#if defined SGC_PTHREAD
    GC_pthread_bsbar_end(barp);
    return;
#endif
}

