/* 
 * Copyright 1988, 1989 Hans-J. Boehm, Alan J. Demers
 * Copyright (c) 1991-1994 by Xerox Corporation.  All rights reserved.
 * Copyright (c) 1998 by Toshio Endo, Kenjiro Taura. All rights reserved.
 * Copyright (c) 1998 by Mitsubishi Research Institute.  All rights reserved.
 * Copyright (c) 1998 by Information-technology Promotion Agency.  All rights reserved.
 *
 * THIS MATERIAL IS PROVIDED AS IS, WITH ABSOLUTELY NO WARRANTY EXPRESSED
 * OR IMPLIED.  ANY USE IS AT YOUR OWN RISK.
 *
 * This program was derived from a program originally developed by
 * Hans-J. Boehm and Alan J. Demers. Extensions have been made by Toshio
 * Endo.
 *
 * Permission is hereby granted to use or copy this program
 * for any purpose,  provided the above notices are retained on all copies.
 * Permission to modify the code and to distribute modified code is granted,
 * provided the above notices are retained, and a notice that the code was
 * modified is included with the above copyright notice.
 */
/* Endo, July 21, 1999 */

# ifndef GC_PRIVATE_H
# define GC_PRIVATE_H

#if defined(mips) && defined(SYSTYPE_BSD43)
    /* MIPS RISCOS 4 */
#   define BSD_TIME
#error
#endif

#ifdef BSD_TIME
#   include <sys/types.h>
#   include <sys/time.h>
#   include <sys/resource.h>
#endif /* BSD_TIME */

# include "gcconfig.h"
# include "sgcconfig.h"
# include "gc_tls.h"
# include "sgc.h"

typedef GC_word word;
typedef GC_signed_word signed_word;

typedef int GC_bool;
# define TRUE 1
# define FALSE 0

typedef char * ptr_t;	/* A generic pointer to which we can add	*/
			/* byte displacements.				*/
			/* Preferably identical to caddr_t, if it 	*/
			/* exists.					*/
			
#if defined(__STDC__)
#   include <stdlib.h>
#   if !(defined( sony_news ) )
#       include <stddef.h>
#   endif
#   define VOLATILE volatile
#else
#   ifdef MSWIN32
#   	include <stdlib.h>
#   endif
#   define VOLATILE
#endif

#ifdef AMIGA
#   define GC_FAR __far
#else
#   define GC_FAR
#endif

/*********************************/
/*                               */
/* Definitions for conservative  */
/* collector                     */
/*                               */
/*********************************/

/*********************************/
/*                               */
/* Easily changeable parameters  */
/*                               */
/*********************************/

#define STUBBORN_ALLOC	/* Define stubborn allocation primitives	*/

# undef STUBBORN_ALLOC

/* #define ALL_INTERIOR_POINTERS */
		    /* Forces all pointers into the interior of an 	*/
		    /* object to be considered valid.  Also causes the	*/
		    /* sizes of all objects to be inflated by at least 	*/
		    /* one byte.  This should suffice to guarantee	*/
		    /* that in the presence of a compiler that does	*/
		    /* not perform garbage-collector-unsafe		*/
		    /* optimizations, all portable, strictly ANSI	*/
		    /* conforming C programs should be safely usable	*/
		    /* with malloc replaced by GC_malloc and free	*/
		    /* calls removed.  There are several disadvantages: */
		    /* 1. There are probably no interesting, portable,	*/
		    /*    strictly ANSI	conforming C programs.		*/
		    /* 2. This option makes it hard for the collector	*/
		    /*    to allocate space that is not ``pointed to''  */
		    /*    by integers, etc.  Under SunOS 4.X with a 	*/
		    /*    statically linked libc, we empiricaly		*/
		    /*    observed that it would be difficult to 	*/
		    /*	  allocate individual objects larger than 100K.	*/
		    /* 	  Even if only smaller objects are allocated,	*/
		    /*    more swap space is likely to be needed.       */
		    /*    Fortunately, much of this will never be	*/
		    /*    touched.					*/
		    /* If you can easily avoid using this option, do.	*/
		    /* If not, try to keep individual objects small.	*/
		    
#define PRINTSTATS   /* Print garbage collection statistics          	*/
		    /* For less verbose output, undefine in reclaim.c 	*/

/* #undef PRINTSTATS */

/* #define PRINTTIMES */ /* Print the amount of time consumed by each garbage   */
		    /* collection.                                         */

#define PRINTBLOCKS /* Print object sizes associated with heap blocks,     */
		    /* whether the objects are atomic or composite, and    */
		    /* whether or not the block was found to be empty      */
		    /* duing the reclaim phase.  Typically generates       */
		    /* about one screenful per garbage collection.         */
#undef PRINTBLOCKS

#define PRINTBLACKLIST 	/* Print black listed blocks, i.e. values that     */
			/* cause the allocator to avoid allocating certain */
			/* blocks in order to avoid introducing "false	   */
			/* hits".					   */
#undef PRINTBLACKLIST

#ifdef SILENT
#  ifdef PRINTSTATS
#    undef PRINTSTATS
#  endif
#  ifdef PRINTTIMES
#    undef PRINTTIMES
#  endif
#  ifdef PRINTNBLOCKS
#    undef PRINTNBLOCKS
#  endif

#  ifdef DO_MARKLOG
#    undef DO_MARKLOG
#  endif
#  ifdef DO_PICLOG
#    undef DO_PICLOG
#  endif
#  ifdef DO_LDLOG
#    undef DO_LDLOG
#  endif
#  ifdef DO_ALLOCLOG
#    undef DO_ALLOCLOG
#  endif
#  ifdef DO_GCRLOG
#    undef DO_GCRLOG
#  endif
#  ifdef DO_MSLOG
#    undef DO_MSLOG
#  endif
#  ifdef DO_CACHELOG
#    undef DO_CACHELOG
#  endif

#endif


#if defined(PRINTSTATS) && !defined(GATHERSTATS)
#   define GATHERSTATS
#endif

#if 0 /* now these definitions are in gcconfig.h. (Jun 3, 99 endo) */
#if defined(SPARC)
#   define ALIGN_DOUBLE  /* Align objects of size > 1 word on 2 word   */
			 /* boundaries.  Wasteful of memory, but       */
			 /* apparently required by SPARC architecture. */
#   define ASM_CLEAR_CODE	/* Stack clearing is crucial, and we 	*/
				/* include assembly code to do it well.	*/
#endif

#ifdef HP_PA
#   define ALIGN_DOUBLE
#endif

#ifdef MIPS /* add by Endo */
#   define ALIGN_DOUBLE
#endif
#endif /* 0 */

#define MERGE_SIZES /* Round up some object sizes, so that fewer distinct */
		    /* free lists are actually maintained.  This applies  */
		    /* only to the top level routines in misc.c, not to   */
		    /* user generated code that calls GC_allocobj and     */
		    /* GC_allocaobj directly.                             */
		    /* Slows down average programs slightly.  May however */
		    /* substantially reduce fragmentation if allocation   */
		    /* request sizes are widely scattered.                */
		    /* May save significant amounts of space for obj_map  */
		    /* entries.						  */

/* ALIGN_DOUBLE requires MERGE_SIZES at present. */
# if defined(ALIGN_DOUBLE) && !defined(MERGE_SIZES)
#   define MERGE_SIZES
# endif

#if defined(ALL_INTERIOR_POINTERS) && !defined(DONT_ADD_BYTE_AT_END)
# define ADD_BYTE_AT_END
#endif


# ifndef LARGE_CONFIG
#   define MINHINCR 16	/* Minimum heap increment, in blocks of HBLKSIZE  */
#   define MAXHINCR 512	/* Maximum heap increment, in blocks              */
# else
#   define MINHINCR 64
#   define MAXHINCR 4096
# endif

# define TIME_LIMIT 50	   /* We try to keep pause times from exceeding	 */
			   /* this by much. In milliseconds.		 */

# define BL_LIMIT GC_black_list_spacing
			   /* If we need a block of N bytes, and we have */
			   /* a block of N + BL_LIMIT bytes available, 	 */
			   /* and N > BL_LIMIT,				 */
			   /* but all possible positions in it are 	 */
			   /* blacklisted, we just use it anyway (and	 */
			   /* print a warning, if warnings are enabled). */
			   /* This risks subsequently leaking the block	 */
			   /* due to a false reference.  But not using	 */
			   /* the block risks unreasonable immediate	 */
			   /* heap growth.				 */

/*********************************/
/*                               */
/* Stack saving for debugging	 */
/*                               */
/*********************************/

/*
 * Number of frames and arguments to save in objects allocated by
 * debugging allocator.
 */
#   define NFRAMES 6	/* Number of frames to save. Even for		*/
			/* alignment reasons.				*/
#   define NARGS 2	/* Mumber of arguments to save for each call.	*/


#ifdef SAVE_CALL_CHAIN
    struct callinfo {
	word ci_pc;
	word ci_arg[NARGS];	/* bit-wise complement to avoid retention */
    };

/* Fill in the pc and argument information for up to NFRAMES of my	*/
/* callers.  Ignore my frame and my callers frame.			*/
void GC_save_callers GC_PROTO((struct callinfo *info));

void GC_print_callers GC_PROTO((struct callinfo *info));

#endif


/*********************************/
/*                               */
/* OS interface routines	 */
/*                               */
/*********************************/

#ifdef BSD_TIME
#   undef CLOCK_TYPE
#   undef GET_TIME
#   undef MS_TIME_DIFF
#   define CLOCK_TYPE struct timeval
#   define GET_TIME(x) { struct rusage rusage; \
			 getrusage (RUSAGE_SELF,  &rusage); \
			 x = rusage.ru_utime; }
#   define MS_TIME_DIFF(a,b) ((double) (a.tv_sec - b.tv_sec) * 1000.0 \
                               + (double) (a.tv_usec - b.tv_usec) / 1000.0)
#else /* !BSD_TIME */
#   include <time.h>
#   if !defined(__STDC__) && defined(SPARC) && defined(SUNOS4)
      clock_t clock();	/* Not in time.h, where it belongs	*/
#   endif
#   if defined(FREEBSD) && !defined(CLOCKS_PER_SEC)
#     include <machine/limits.h>
#     define CLOCKS_PER_SEC CLK_TCK
#   endif
#   if !defined(CLOCKS_PER_SEC)
#     define CLOCKS_PER_SEC 1000000
/*
 * This is technically a bug in the implementation.  ANSI requires that
 * CLOCKS_PER_SEC be defined.  But at least under SunOS4.1.1, it isn't.
 * Also note that the combination of ANSI C and POSIX is incredibly gross
 * here. The type clock_t is used by both clock() and times().  But on
 * some machines these use different notions of a clock tick,  CLOCKS_PER_SEC
 * seems to apply only to clock.  Hence we use it here.  On many machines,
 * including SunOS, clock actually uses units of microseconds (which are
 * not really clock ticks).
 */
#   endif
#   define CLOCK_TYPE clock_t
#   define GET_TIME(x) x = clock()
#   define MS_TIME_DIFF(a,b) ((unsigned long) \
		(1000.0*(double)((a)-(b))/(double)CLOCKS_PER_SEC))
#endif /* !BSD_TIME */

/* We use bzero and bcopy internally.  They may not be available.	*/
# if defined(SPARC) && defined(SUNOS4)
#   define BCOPY_EXISTS
# endif
# if defined(M68K) && defined(AMIGA)
#   define BCOPY_EXISTS
# endif
# if defined(M68K) && defined(NEXT)
#   define BCOPY_EXISTS
# endif
# if defined(VAX)
#   define BCOPY_EXISTS
# endif
# if defined(AMIGA)
#   include <string.h>
#   define BCOPY_EXISTS
# endif

#ifndef SDSM_TEST
# ifndef BCOPY_EXISTS
#   include <string.h>
#   define BCOPY(x,y,n) memcpy(y, x, (size_t)(n))
#   define BZERO(x,n)  memset(x, 0, (size_t)(n))
# else
#   define BCOPY(x,y,n) bcopy((char *)(x),(char *)(y),(int)(n))
#   define BZERO(x,n) bzero((char *)(x),(int)(n))
# endif
#else /* !SDSM_TEST */
#  define BCOPY(x,y,n) {int i;for (i=0;i<n;i++) \
	  ((char*)(y))[i] = ((char*)(x))[i];}
#  define BZERO(x,n) {int i;for (i=0;i<n;i++) ((char*)(x))[i] = (char)0;}
#endif /* !SDSM_TEST */

/* HBLKSIZE aligned allocation.  0 is taken to mean failure 	*/
/* space is assumed to be cleared.				*/
#   ifdef OS2
      void * os2_alloc(size_t bytes);
#     define GET_MEM(bytes) HBLKPTR((ptr_t)os2_alloc((size_t)bytes + HBLKSIZE) \
                                    + HBLKSIZE-1)
#   else
#     if defined(AMIGA) || defined(NEXT)
#       define GET_MEM(bytes) HBLKPTR((size_t) \
				      calloc(1, (size_t)bytes + HBLKSIZE) \
                                      + HBLKSIZE-1)
#     else
#	ifdef MSWIN32
          extern ptr_t GC_win32_get_mem();
#         define GET_MEM(bytes) (struct hblk *)GC_win32_get_mem(bytes)
#	else
#	  ifdef MACOS
#	    if defined(USE_TEMPORARY_MEMORY)
		extern Ptr GC_MacTemporaryNewPtr(size_t size,
						 Boolean clearMemory);
#               define GET_MEM(bytes) HBLKPTR( \
		    GC_MacTemporaryNewPtr(bytes + HBLKSIZE, true) + HBLKSIZE-1)
#	    else
#         	    define GET_MEM(bytes) HBLKPTR( \
			NewPtrClear(bytes + HBLKSIZE) + HBLKSIZE-1)
#	    endif
#	  else
            extern ptr_t GC_sh_get_mem GC_PROTO((word bytes));
#           define GET_MEM(bytes) (struct hblk *)GC_sh_get_mem(bytes)
/*#           define GET_MEM(bytes) (struct hblk *)GC_unix_get_mem(bytes)*/
#	  endif
#	endif
#     endif
#   endif

/* thread package */

# if !defined BSD_TIME && defined TIMEVAL_SUPPORTED
#   include <sys/time.h>
# endif

# include "psm.h"

# include "gc_hdrs.h"
# ifdef DO_PICLOG
#   include "perfc_util.h"
# endif /* DO_PICLOG */
# include "gc_shlog.h"

# if defined SGC_SOLARIS_THREAD || \
     defined SGC_SOLARIS_THREAD_PTHREAD
#   include <thread.h>
    typedef thread_t GC_thread_id;
# endif
# if defined SGC_PTHREAD || \
     defined SGC_SOLARIS_THREAD_PTHREAD
#   include <pthread.h>
# endif
# if defined SGC_PTHREAD && !defined SGC_SOLARIS_THREAD_PTHREAD
    typedef pthread_t GC_thread_id;
# endif
# if defined SGC_SPROC_THREAD
#   include <sys/types.h>
#   include <sys/prctl.h>
    typedef pid_t GC_thread_id;
# endif
# if defined SGC_FORK_THREAD
#   include <sys/types.h>
#   include <unistd.h>
    typedef pid_t GC_thread_id;
# endif

# include "gc_bsync.h"

#ifdef DO_LDLOG
#define MUTEX_LOCK_LOG(mutex_p, kind) \
if (!PSM_mutex_trylock(mutex_p)) {\
  GC_LDLOG_PUT(GC_LDC_S, kind);\
  PSM_mutex_lock(mutex_p);\
  GC_LDLOG_PUT(GC_LDC_E, kind);\
}
#else
#define MUTEX_LOCK_LOG(mutex_p, kind) PSM_mutex_lock(mutex_p)
#endif

#  define LOCK_MB(hhdr, off) MUTEX_LOCK_LOG(&hhdr->mutex[off], GC_LDK_WMB)
#  define UNLOCK_MB(hhdr, off) PSM_mutex_unlock(&hhdr->mutex[off])

/* Delay any interrupts or signals that may abort this thread.  Data	*/
/* structures are in a consistent state outside this pair of calls.	*/
/* ANSI C allows both to be empty (though the standard isn't very	*/
/* clear on that point).  Standard malloc implementations are usually	*/
/* neither interruptable nor thread-safe, and thus correspond to	*/
/* empty definitions.							*/
#   if defined(SRC_M3) || defined(AMIGA) || defined(SOLARIS_THREADS) \
	|| defined(MSWIN32) || defined(MACOS) || defined(DJGPP) \
	|| defined(NO_SIGNALS)
			/* Also useful for debugging.		*/
	/* Should probably use thr_sigsetmask for SOLARIS_THREADS. */
#     define DISABLE_SIGNALS()
#     define ENABLE_SIGNALS()
#   else
#     define DISABLE_SIGNALS() GC_disable_signals()
        void GC_disable_signals GC_PROTO((void)); 
#     define ENABLE_SIGNALS() GC_enable_signals()
        void GC_enable_signals GC_PROTO((void)); 
#   endif

/* Abandon ship */
#   ifdef SMALL_CONFIG
#	define ABORT(msg) abort();
#   else
        void GC_abort GC_PROTO((char *msg)); 
#       define ABORT(msg) GC_abort(msg);
#   endif

/* Exit abnormally, but without making a mess (e.g. out of memory) */
#   define EXIT() (void)exit(1)

/* Print warning message, e.g. almost out of memory.	*/
# define WARN(msg,arg) (*GC_current_warn_proc)(msg, (GC_word)(arg))
extern GC_warn_proc GC_current_warn_proc;

/*********************************/
/*                               */
/* Word-size-dependent defines   */
/*                               */
/*********************************/

#if CPP_WORDSZ == 32
#  define WORDS_TO_BYTES(x)   ((x)<<2)
#  define BYTES_TO_WORDS(x)   ((x)>>2)
#  define LOGWL               ((word)5)    /* log[2] of CPP_WORDSZ */
#  define modWORDSZ(n) ((n) & 0x1f)        /* n mod size of word	    */
#  if ALIGNMENT != 4
#	define UNALIGNED
#  endif
#endif

#if CPP_WORDSZ == 64
#  define WORDS_TO_BYTES(x)   ((x)<<3)
#  define BYTES_TO_WORDS(x)   ((x)>>3)
#  define LOGWL               ((word)6)    /* log[2] of CPP_WORDSZ */
#  define modWORDSZ(n) ((n) & 0x3f)        /* n mod size of word	    */
#  if ALIGNMENT != 8
#	define UNALIGNED
#  endif
#endif

#define WORDSZ ((word)CPP_WORDSZ)
#define SIGNB  ((word)1 << (WORDSZ-1))
#define BYTES_PER_WORD      ((word)(sizeof (word)))
#define ONES                ((word)(-1))
#define divWORDSZ(n) ((n) >> LOGWL)	   /* divide n by size of word      */

/*********************/
/*                   */
/*  Size Parameters  */
/*                   */
/*********************/

/*  heap block size, bytes. Should be power of 2 */

#ifdef SMALL_CONFIG
#   define CPP_LOG_HBLKSIZE 10
#else
# if CPP_WORDSZ == 32
#   define CPP_LOG_HBLKSIZE 12
# else
/* GC_reclaim1 in reclaim.c requires that
   CPP_LOG_HBLKSIZE must greater than 8. */
#   define CPP_LOG_HBLKSIZE 13
# endif
#endif
#define LOG_HBLKSIZE   ((word)CPP_LOG_HBLKSIZE)
#define CPP_HBLKSIZE (1 << CPP_LOG_HBLKSIZE)
#define HBLKSIZE ((word)CPP_HBLKSIZE)


/*  max size objects supported by freelist (larger objects may be   */
/*  allocated, but less efficiently)                                */

#define CPP_MAXOBJSZ    BYTES_TO_WORDS(CPP_HBLKSIZE/2)
#define MAXOBJSZ ((word)CPP_MAXOBJSZ)
		
# define divHBLKSZ(n) ((n) >> LOG_HBLKSIZE)

# define HBLK_PTR_DIFF(p,q) divHBLKSZ((ptr_t)p - (ptr_t)q)
	/* Equivalent to subtracting 2 hblk pointers.	*/
	/* We do it this way because a compiler should	*/
	/* find it hard to use an integer division	*/
	/* instead of a shift.  The bundled SunOS 4.1	*/
	/* o.w. sometimes pessimizes the subtraction to	*/
	/* involve a call to .div.			*/
 
# define modHBLKSZ(n) ((n) & (HBLKSIZE-1))
 
# define HBLKPTR(objptr) ((struct hblk *)(((word) (objptr)) & ~(HBLKSIZE-1)))

# define HBLKDISPL(objptr) (((word) (objptr)) & (HBLKSIZE-1))

/* Round up byte allocation requests to integral number of words, etc. */
# ifdef ADD_BYTE_AT_END
#   define ROUNDED_UP_WORDS(n) BYTES_TO_WORDS((n) + WORDS_TO_BYTES(1))
#   ifdef ALIGN_DOUBLE
#       define ALIGNED_WORDS(n) (BYTES_TO_WORDS((n) + WORDS_TO_BYTES(2)) & ~1)
#   else
#       define ALIGNED_WORDS(n) ROUNDED_UP_WORDS(n)
#   endif
#   define SMALL_OBJ(bytes) ((bytes) < WORDS_TO_BYTES(MAXOBJSZ))
#   define ADD_SLOP(bytes) ((bytes)+1)
# else
#   define ROUNDED_UP_WORDS(n) BYTES_TO_WORDS((n) + (WORDS_TO_BYTES(1) - 1))
#   ifdef ALIGN_DOUBLE
#       define ALIGNED_WORDS(n) \
			(BYTES_TO_WORDS((n) + WORDS_TO_BYTES(2) - 1) & ~1)
#   else
#       define ALIGNED_WORDS(n) ROUNDED_UP_WORDS(n)
#   endif
#   define SMALL_OBJ(bytes) ((bytes) <= WORDS_TO_BYTES(MAXOBJSZ))
#   define ADD_SLOP(bytes) (bytes)
# endif

/*
 * Hash table representation of sets of pages.  This assumes it is
 * OK to add spurious entries to sets.
 * Used by black-listing code, and perhaps by dirty bit maintenance code.
 */
 
# ifdef LARGE_CONFIG
#   define LOG_PHT_ENTRIES  17
# else
#   define LOG_PHT_ENTRIES  14	/* Collisions are likely if heap grows	*/
				/* to more than 16K hblks = 64MB.	*/
				/* Each hash table occupies 2K bytes.   */
# endif
# define PHT_ENTRIES ((word)1 << LOG_PHT_ENTRIES)
# define PHT_SIZE (PHT_ENTRIES >> LOGWL)
typedef word page_hash_table[PHT_SIZE];

# define PHT_HASH(addr) ((((word)(addr)) >> LOG_HBLKSIZE) & (PHT_ENTRIES - 1))

# define get_pht_entry_from_index(bl, index) \
		(((bl)[divWORDSZ(index)] >> modWORDSZ(index)) & 1)
# define set_pht_entry_from_index(bl, index) \
		(bl)[divWORDSZ(index)] |= (word)1 << modWORDSZ(index)
# define clear_pht_entry_from_index(bl, index) \
		(bl)[divWORDSZ(index)] &= ~((word)1 << modWORDSZ(index))

/********************************************/
/*                                          */
/*    H e a p   B l o c k s                 */
/*                                          */
/********************************************/

/*  heap block header */
#define HBLKMASK   (HBLKSIZE-1)

#define BITS_PER_HBLK (HBLKSIZE * 8)

#define MARK_BITS_PER_HBLK (BITS_PER_HBLK/CPP_WORDSZ)
	   /* upper bound                                    */
	   /* We allocate 1 bit/word.  Only the first word   */
	   /* in each object is actually marked.             */

# ifdef ALIGN_DOUBLE
#   define MARK_BITS_SZ (((MARK_BITS_PER_HBLK + 2*CPP_WORDSZ - 1) \
			  / (2*CPP_WORDSZ))*2)
# else
#   define MARK_BITS_SZ ((MARK_BITS_PER_HBLK + CPP_WORDSZ - 1)/CPP_WORDSZ)
# endif
	   /* Upper bound on number of mark words per heap block  */

#ifdef USE_DEMOWIN
#  define DEMOWIN_MB_UPDATE(obj, hhdr) GC_demowin_mb_update(obj, hhdr)
#else
#  define DEMOWIN_MB_UPDATE(obj, hhdr)
#endif

struct hblkhdr {
    word hb_sz;  /* If in use, size in words, of objects in the block. */
		 /* if free, the size in bytes of the whole block      */
    struct hblk * hb_next; 	/* Link field for hblk free list	 */
    				/* and for lists of chunks waiting to be */
    				/* reclaimed.				 */
    word hb_descr;   		/* object descriptor for marking.  See	*/
    				/* mark.h.				*/
    char* hb_map;	/* A pointer to a pointer validity map of the block. */
    		      	/* See GC_obj_map.				     */
    		     	/* Valid for all blocks with headers.		     */
    		     	/* Free blocks point to GC_invalid_map.		     */
    unsigned char hb_obj_kind;
    			 /* Kind of objects in the block.  Each kind 	*/
    			 /* identifies a mark procedure and a set of 	*/
    			 /* list headers.  Sometimes called regions.	*/
    unsigned char hb_flags;
#	define IGNORE_OFF_PAGE	1	/* Ignore pointers that do not	*/
					/* point to the first page of 	*/
					/* this object.			*/
    unsigned short hb_last_reclaimed;
    				/* Value of GC_gc_no when block was	*/
    				/* last allocated or swept. May wrap.   */
    word hb_marks[MARK_BITS_SZ];
			    /* Bit i in the array refers to the             */
			    /* object starting at the ith word (header      */
			    /* INCLUDED) in the heap block.                 */
			    /* The lsb of word 0 is numbered 0.		    */

    struct GC_Thread_Rep *home_envp;
        /* home_envp means the process who allocated the hblk. */
        /* One of App/SPMD processes. */
    PSM_mutex_t mutex[MARK_BITS_SZ]; /* used if SETMB_KIND == 1, 2 */
#ifdef SGC_BALANCED_HDR
    struct GC_Thread_Rep *hdrhome_envp;
        /* hdrhome_envp means the process who allocated this hblkhdr. */
        /* One of GC/SPMD processes. */
#endif
    int hdr_allocator; /* for statistics. SPMD process No. (OBSOLETE) */
#ifdef USE_DEMOWIN
    void *rectid[MARK_BITS_PER_HBLK];
    short marker[MARK_BITS_PER_HBLK];
    int dirty;
#endif
#ifdef COUNT_CROWD
    word crowd_count;
    word crowd_max;
#endif
#ifdef FIND_LEAK
    ptr_t gfl;
    ptr_t gfl_top;
    PSM_mutex_t gfl_ml;
#endif
};

/*  heap block body */

# define DISCARD_WORDS 0
	/* Number of words to be dropped at the beginning of each block	*/
	/* Must be a multiple of WORDSZ.  May reasonably be nonzero	*/
	/* on machines that don't guarantee longword alignment of	*/
	/* pointers, so that the number of false hits is minimized.	*/
	/* 0 and WORDSZ are probably the only reasonable values.	*/

# define BODY_SZ ((HBLKSIZE-WORDS_TO_BYTES(DISCARD_WORDS))/sizeof(word))

struct hblk {
#   if (DISCARD_WORDS != 0)
        word garbage[DISCARD_WORDS];
#   endif
    word hb_body[BODY_SZ];
};

# define HDR_WORDS ((word)DISCARD_WORDS)
# define HDR_BYTES ((word)WORDS_TO_BYTES(DISCARD_WORDS))
# define OBJ_SZ_TO_BLOCKS(sz) \
    divHBLKSZ(HDR_BYTES + WORDS_TO_BYTES(sz) + HBLKSIZE-1)
/* Size of block (in units of HBLKSIZE) needed to hold objects of	*/
/* given sz (in words).						*/
     
/* Object free list link */
# define obj_link(p) (*(ptr_t *)(p))

/***********************************************************/
# define MAXOBJKINDS 6

#   ifdef ALL_INTERIOR_POINTERS
#	define map_entry_type short
#       define OBJ_INVALID 0x7fff
#	define MAP_ENTRY(map, bytes) \
		(((map_entry_type *)(map))[BYTES_TO_WORDS(bytes)])
#	define MAP_ENTRIES BYTES_TO_WORDS(HBLKSIZE)
#	define MAP_SIZE (MAP_ENTRIES * sizeof(map_entry_type))
#	define OFFSET_VALID(displ) TRUE
#	define CPP_MAX_OFFSET (HBLKSIZE - HDR_BYTES - 1)
#	define MAX_OFFSET ((word)CPP_MAX_OFFSET)
#   else
#	define map_entry_type char
#       define OBJ_INVALID 0x7f
#	define MAP_ENTRY(map, bytes) \
		(map)[bytes]
#	define MAP_ENTRIES HBLKSIZE
#	define MAP_SIZE MAP_ENTRIES
#	define CPP_MAX_OFFSET (WORDS_TO_BYTES(OBJ_INVALID) - 1)	
#	define MAX_OFFSET ((word)CPP_MAX_OFFSET)
# 	define VALID_OFFSET_SZ \
	  (CPP_MAX_OFFSET > WORDS_TO_BYTES(CPP_MAXOBJSZ)? \
	   CPP_MAX_OFFSET+1 \
	   : WORDS_TO_BYTES(CPP_MAXOBJSZ)+1)
#	define OFFSET_VALID(displ) GC_valid_offsets[displ]
#   endif

# ifndef GC_MARK_H
#   include "gc_mark.h"
# endif

typedef struct {
    mse * buf;
    word size;
    mse * top;
    char dummy1[124];
    mse * bottom;
    mse * watermark;
    PSM_mutex_t ml;
} mark_stack_t;

typedef struct {
    char * buf;
    word size;
    char * in;
    char * out;
    PSM_mutex_t ml;
} mark_queue_t;

#ifdef DO_REQ_REMOTE
struct reqbuf_entry {
    word obj;
    hdr * hhdr;
};

struct reqbuf {
    struct reqbuf_entry *buf;
    int validsize;
    int bufsize;
    PSM_mutex_t ml;
};
#endif

typedef struct {
  signed_word mem_found;
			/* Number of longwords of memory GC_reclaimed     */
# ifdef GATHERSTATS
    word composite_in_use;
   		/* Number of words in accessible composite	*/
		/* objects.					*/
    word atomic_in_use;
   		/* Number of words in accessible atomic		*/
		/* objects.					*/
# endif
} gcproc_stat_t;

typedef struct {
  word words_allocd_before_gc;
		/* Number of words allocated before this	*/
		/* collection cycle.				*/
  word words_allocd;
  	/* Number of words allocated during this collection cycle */
  word words_wasted;
  	/* Number of words wasted due to internal fragmentation	 */
  	/* in large objects allocated since last gc. Approximate.*/
  word words_finalized;
  	/* Approximate number of words in objects (and headers)	*/
  	/* That became ready for finalization in the last 	*/
  	/* collection.						*/
  word non_gc_bytes_at_gc;
  	/* Number of explicitly managed bytes of storage 	*/
  	/* at last collection.					*/
  word non_gc_bytes;
        /* Number of bytes not intended to be collected */
  word mem_freed;
  	/* Number of explicitly deallocated words of memory	*/
  	/* since last collection.				*/
} appproc_stat_t;

/* Object kinds: */
struct obj_kind {
   word ok_descriptor;  /* Descriptor template for objects in this	*/
   			/* block.					*/
   GC_bool ok_relocate_descr;
   			/* Add object size in bytes to descriptor 	*/
   			/* template to obtain descriptor.  Otherwise	*/
   			/* template is used as is.			*/
   GC_bool ok_init;     /* Clear objects before putting them on the free list. */
};

/* Predefined kinds: */
# define PTRFREE 0
# define NORMAL  1
# define UNCOLLECTABLE 2
# ifdef ATOMIC_UNCOLLECTABLE
#   define AUNCOLLECTABLE 3
#   define STUBBORN 4
#   define IS_UNCOLLECTABLE(k) (((k) & ~1) == UNCOLLECTABLE)
# else
#   define STUBBORN 3
#   define IS_UNCOLLECTABLE(k) ((k) == UNCOLLECTABLE)
# endif

# ifdef LARGE_CONFIG
#   define MAX_ROOT_SETS 4096
# else
#   ifdef PCR
#     define MAX_ROOT_SETS 1024
#   else
#     ifdef MSWIN32
#	define MAX_ROOT_SETS 512
	    /* Under NT, we add only written pages, which can result 	*/
	    /* in many small root sets.					*/
#     else
#       define MAX_ROOT_SETS 64
#     endif
#   endif
# endif

/* Data structure for list of root sets.				*/
/* We keep a hash table, so that we can filter out duplicate additions.	*/
/* Under Win32, we need to do a better job of filtering overlaps, so	*/
/* we resort to sequential search, and pay the price.			*/
struct roots {
	ptr_t r_start;
	ptr_t r_end;
#	ifndef MSWIN32
	  struct roots * r_next;
#	endif
	GC_bool r_tmp;
	  	/* Delete before registering new dynamic libraries */
};

struct scratch {
    PSM_mutex_t ml;
    ptr_t free_ptr;
    ptr_t end_ptr;
};

struct shm_handler {
#ifdef ADDRS_SEPARATE
    char filename[64];
    int filedesc;
#else
    int dummy;
#endif
};

/* begin finalizer types */

/* Type of mark procedure used for marking from finalizable object.	*/
/* This procedure normally does not mark the object, only its		*/
/* descendents.								*/
typedef void finalization_mark_proc GC_PROTO((ptr_t finalizable_obj_ptr)); 

# define HASH3(addr,size,log_size) \
    ((((word)(addr) >> 3) ^ ((word)(addr) >> (3+(log_size)))) \
    & ((size) - 1))
#define HASH2(addr,log_size) HASH3(addr, 1 << log_size, log_size)

struct hash_chain_entry {
    word hidden_key;
    struct hash_chain_entry * next;
};

struct disappearing_link {
    struct hash_chain_entry prolog;
#   define dl_hidden_link prolog.hidden_key
				/* Field to be cleared.		*/
#   define dl_next(x) (struct disappearing_link *)((x) -> prolog.next)
#   define dl_set_next(x,y) (x) -> prolog.next = (struct hash_chain_entry *)(y)

    word dl_hidden_obj;		/* Pointer to object base	*/
};

struct finalizable_object {
    struct hash_chain_entry prolog;
#   define fo_hidden_base prolog.hidden_key
				/* Pointer to object base.	*/
				/* No longer hidden once object */
				/* is on finalize_now queue.	*/
#   define fo_next(x) (struct finalizable_object *)((x) -> prolog.next)
#   define fo_set_next(x,y) (x) -> prolog.next = (struct hash_chain_entry *)(y)
    GC_finalization_proc fo_fn;	/* Finalizer.			*/
    ptr_t fo_client_data;
    word fo_object_size;	/* In bytes.			*/
    finalization_mark_proc * fo_mark_proc;	/* Mark-through procedure */
};

/* end finalizer types */

struct gcproc_info {
    gcproc_stat_t stat;
    long pno; /* Sequent No. */

    /* For Load Distributing */
    struct {
# if DO_LOAD_DIST == 3
	int busy_flag; /* Used if DO_OPT_NSB == 1 */
	int nonempty_flag; /* Used if DO_OPT_NSB == 1 */
	mark_queue_t smq; /* Stealable mark queue */
# endif

# if DO_LOAD_DIST == 4 || DO_LOAD_DIST == 5
	int busy_flag; /* Used if DO_OPT_NSB == 1 */
# endif
    } ld;

#ifdef DO_REQ_REMOTE
    struct reqbuf reqbuf;
#endif
    int spmd_start_flag;

    /* for parallel reclaim */
    struct {
#ifdef DO_PRECLAIM
	SH_HBCHUNK * chunkp;
	struct GC_Thread_Rep *temp_rlist_home_envp;
	struct hblk * temp_rlist_start;
#endif
#ifdef FIND_LEAK
	ptr_t *freelist_top[MAXOBJKINDS];
#endif
    } pr;

#ifdef SGC_BALANCED_HDR
    hdr * _hdr_free_list;
    PSM_mutex_t _hdr_free_list_ml;
    int _n_free_hdr;
#endif

#ifdef LOCAL_HDR_COPY
    bottom_index * _all_bottom_indices_copy;
    hdr * _copied_hdr_free_list;
    hdr * _copied_hdr_used_list;
#endif

    /* structure for debug log (local) */
#ifdef DO_LDLOG
    struct lg_ldlog *ldlogp;
#endif
#ifdef DO_MSLOG
    struct lg_mslog *mslogp;
#endif
};

struct appproc_info {
    appproc_stat_t stat;
    void *gcregvar; /* envp can be retrieve from gcregvar */
    ptr_t _stackbottom;
    ptr_t *freelist[MAXOBJKINDS];
    struct hblk **reclaim_list[MAXOBJKINDS];
    /* Avoid contention between local allocation and signal */
    PSM_mutex_t alloc_ml;
    /* for parallel reclaim */
    struct {
#ifdef DO_PRECLAIM
	PSM_mutex_t link_rl_ml;
#endif
    } pr;

/*#define SGC_FINISHED 1*/
#define SGC_DETACHED 2
#define SGC_MAIN_THREAD 4
    int flags;
    /* Used only if `use_polling' of SGC_attr_t structure is set. */
    int polling_flag;
    /* Simple barrier synchronization for App processes. */
    long barrier_stage;

    struct roots static_roots[MAX_ROOT_SETS];
    int n_root_sets;
    /* static_roots[0..n_root_sets) contains the valid root sets. */

    struct {
	word* buf;
	int validsize;
	int bufsize;
    } rootcopy;

#ifdef DO_ALLOCLOG
    struct la_alloclog *alloclogp;
#endif
};

struct GC_Thread_Rep {
    /* Common members BEGIN */
    GC_thread_id pid; /* Process (Thread) ID in OS. if <0, not used */
    int app_or_gc; /* 0 ... App process, 1 ... GC process */
    int initialized;
    int terminated;
    struct GC_Thread_Rep *next;
    sh_bssem_t sem;
    /* For debug */
    int debug_state;
    struct scratch scratch;
    mark_stack_t mark_stack;
    GC_bool _mark_stack_too_small; /* We need a larger mark stack.  May be */
				/* set by client supplied mark routines.*/
    mark_state_t _mark_state;
    GC_bool _objects_are_marked;	/* Are there collectable marked	*/
                                /* objects in the heap?		*/
    /* Although all threads have this member, */
    /* statistics only on GC threads are printed. */
#ifdef DO_MARKLOG
    struct l_marklog *marklogp;
#endif
#ifdef DO_PICLOG
    struct l_piclog *piclogp;
#endif
#ifdef DO_CACHELOG
    struct l_cachelog *cachelogp;
#endif
    /* Common members END */
    
    struct gcproc_info gc;
    struct appproc_info app;
};

typedef struct GC_Thread_Rep * GC_thread;

struct envp_list {
    GC_thread head;
    PSM_mutex_t ml;
};

/*  lists of all heap blocks and free lists	*/
/* These are grouped together in a struct	*/
/* so that they can be easily skipped by the	*/
/* GC_mark routine.				*/
/* The ordering is weird to make GC_malloc	*/
/* faster by keeping the important fields	*/
/* sufficiently close together that a		*/
/* single load of a base register will do.	*/
/* Scalars that could easily appear to		*/
/* be pointers are also put here.		*/
/* The main fields should precede any 		*/
/* conditionally included fields, so that	*/
/* gc_inl.h will work even if a different set	*/
/* of macros is defined when the client is	*/
/* compiled.					*/

struct _GC_arrays {
  word _heapsize;
  word _max_heapsize;
  ptr_t _last_heap_addr;
  ptr_t _prev_heap_addr;

    gcproc_stat_t sum_gc_stat;
    appproc_stat_t sum_app_stat;

# ifdef MERGE_SIZES
    unsigned _size_map[WORDS_TO_BYTES(MAXOBJSZ+1)];
    	/* Number of words to allocate for a given allocation request in */
    	/* bytes.							 */
    PSM_mutex_t _size_map_ml;
# endif 

    ptr_t _obj_map[MAXOBJSZ+1];
                       /* If not NIL, then a pointer to a map of valid  */
    		       /* object addresses. _obj_map[sz][i] is j if the	*/
    		       /* address block_start+i is a valid pointer      */
    		       /* to an object at				*/
    		       /* block_start+i&~3 - WORDS_TO_BYTES(j).		*/
    		       /* (If ALL_INTERIOR_POINTERS is defined, then	*/
    		       /* instead ((short *)(hbh_map[sz])[i] is j if	*/
    		       /* block_start+WORDS_TO_BYTES(i) is in the	*/
    		       /* interior of an object starting at		*/
    		       /* block_start+WORDS_TO_BYTES(i-j)).		*/
    		       /* It is OBJ_INVALID if				*/
    		       /* block_start+WORDS_TO_BYTES(i) is not		*/
    		       /* valid as a pointer to an object.              */
    		       /* We assume all values of j <= OBJ_INVALID.	*/
    		       /* The zeroth entry corresponds to large objects.*/
    ptr_t _invalid_map;
			/* Pointer to the nowhere valid hblk map */
			/* Blocks pointing to this map are free. */
    PSM_mutex_t _obj_map_ml;

# ifdef STUBBORN_ALLOC
      page_hash_table _changed_pages;
        /* Stubborn object pages that were changes since last call to	*/
	/* GC_read_changed.						*/
      page_hash_table _prev_changed_pages;
        /* Stubborn object pages that were changes before last call to	*/
	/* GC_read_changed.						*/
# endif
# if defined(PROC_VDB) || defined(MPROTECT_VDB)
      page_hash_table _grungy_pages; /* Pages that were dirty at last 	   */
				     /* GC_read_dirty.			   */
# endif
# ifdef LARGE_CONFIG
#   if CPP_WORDSZ > 32
#     define MAX_HEAP_SECTS 4096 	/* overflows at roughly 64 GB	   */
#   else
#     define MAX_HEAP_SECTS 768		/* Separately added heap sections. */
#   endif
# else
#   define MAX_HEAP_SECTS 256
# endif
  struct HeapSect {
      ptr_t hs_start; word hs_bytes;
  } _heap_sects[MAX_HEAP_SECTS];
# ifdef MSWIN32
    ptr_t _heap_bases[MAX_HEAP_SECTS];
    		/* Start address of memory regions obtained from kernel. */
# endif
    PSM_mutex_t _heap_sects_ml;

  /* Block header index */
  bottom_index * _all_nils;
  bottom_index * _top_index [TOP_SZ];
				/* Each entry points to a bottom_index.	*/
				/* On a 32 bit machine, it points to 	*/
				/* the index for a set of high order	*/
				/* bits equal to the index.  For longer	*/
				/* addresses, we hash the high order	*/
				/* bits to compute the index in 	*/
				/* GC_top_index, and each entry points	*/
				/* to a hash chain.			*/
				/* The last entry in each chain is	*/
				/* GC_all_nils.				*/
#ifdef SAVE_CALL_CHAIN
  struct callinfo _last_stack[NFRAMES];	/* Stack at last garbage collection.*/
  					/* Useful for debugging	mysterious  */
  					/* object disappearances.	    */
  					/* In the multithreaded case, we    */
  					/* currently only save the calling  */
  					/* stack.			    */
#endif
    
#   ifndef ALL_INTERIOR_POINTERS
  	char _valid_offsets[VALID_OFFSET_SZ];
				/* GC_valid_offsets[i] == TRUE ==> i 	*/
				/* is registered as a displacement.	*/
  	char _modws_valid_offsets[sizeof(word)];
				/* GC_valid_offsets[i] ==>		  */
				/* GC_modws_valid_offsets[i%sizeof(word)] */
#   endif

    struct hblk * _scan_ptr;
    int _scan_ptr_flag;
    PSM_mutex_t _scan_ptr_ml;

    struct {
	PSM_mutex_t ml;
	int ospagesize;
	struct shm_handler handler;
#ifdef SHM_FIXED
	ptr_t shmaddr;
	int shmsize;
	ptr_t allocptr;
	ptr_t limitptr;
	ptr_t alloc2ptr;
	ptr_t limit2ptr;
#endif
    } shm;

    struct scratch scratch;

    int fail_count; /* How many allocation failures? */

    /* 0...# of app procs is variable, 1...fixed */
    int np_fixed;
    /* 0...Stop by signal, 1...Stop by polling */
    int use_polling;

    int spmd_start_count;
    void *(*spmd_func)(void*);
    void *spmd_arg;

    ptr_t leader_stackbottom;

    /* used by SGC_thr_create, SGC_pthread_create */
    struct GC_sh_funcarg {
	void * (*func)(void *);
	void *arg;
	int flags; /* creation flag. SGC_DETACHED/SGC_MAIN_THREAD */
	PSM_mutex_t ml;
    } thread_funcarg;

    struct envp_list appproc_envps;
    struct envp_list gcproc_envps;

    /* Lock the shared memory */
    PSM_mutex_t alloc_ml;

    struct obj_kind _obj_kinds[MAXOBJKINDS];

    /* for Load Distributing */
    struct {
# if DO_LOAD_DIST == 3
	/* Following are used if DO_OPT_NSB == 0 */
	/* For termination detection. */
	long busy_count; /* Num of non-idle processes */
	PSM_mutex_t busy_count_ml;
	long nonempty_count; /* Num of non-empty stack(queue)s */
	PSM_mutex_t nonempty_count_ml;

	/* Following are used if DO_OPT_NSB == 1 */
	/* For termination detection. */
	PSM_mutex_t dterm_ml;
	int flag_modified; 
	int all_finished;
# endif
# if DO_LOAD_DIST == 4 || DO_LOAD_DIST == 5
	/* Following are used if DO_OPT_NSB == 0 */
	/* For termination detection. */
	int busy_count; /* Num of non-idle processes */
	PSM_mutex_t busy_count_ml;
	
	/* Following are used if DO_OPT_NSB == 1 */
	/* For termination detection. */
	PSM_mutex_t dterm_ml;
	int flag_modified; 
	int all_finished;
# endif
    } ld;

    struct {
#ifdef DO_PRECLAIM
	/* For heapblock iterator (Used in parallel reclaimation) */
	bottom_index * hbiter_index_p;
	int hbiter_j;
	PSM_mutex_t hbiter_ml;
	int hbiter_ddd;
	/* For parallel reclaimation */
	PSM_mutex_t freehblk_ml;
	PSM_mutex_t hdr_fl_ml;
	
	int bicounter;
	PSM_mutex_t bicounter_ml;
	/* Used for parallel reclaim */
	PSM_mutex_t link_rl_ml;
#endif
    } pr;

    struct {
	PSM_mutex_t print_ml;
	PSM_mutex_t exitprint_ml;
	int gcing;
    } debug;
    
    bottom_index * _all_bottom_indices;

#ifndef SGC_BALANCED_HDR
    hdr * _hdr_free_list;
#endif

    struct hblk **reclaim_list[MAXOBJKINDS]; /* global reclaim list */
    PSM_mutex_t reclaim_list_ml;

#ifdef FIND_LEAK
    ptr_t *freelist[MAXOBJKINDS]; /* global freelist */
    PSM_mutex_t freelist_ml;
#endif


  struct hblk * _hblkfreelist;
				/* List of completely empty heap blocks	*/
				/* Linked through hb_next field of 	*/
				/* header structure associated with	*/
				/* block.				*/
  struct hblk *_savhbp;  /* heap block preceding next */
                         /* block to be examined by   */
		 	 /* GC_allochblk.                */
  struct hblk * _freehblk_ptr;  /* Search position hint for GC_freehblk */

  word _n_heap_sects;	/* Number of sections currently in heap. */
  word _root_size;   /* Total size of registered root sections */

  ptr_t _least_plausible_heap_addr;
  ptr_t _greatest_plausible_heap_addr;
			/* Bounds on the heap.  Guaranteed valid	*/
			/* Likely to include future heap expansion.	*/

    struct {
/* Pointers to individual tables.  We replace one table by another by 	*/
/* switching these pointers. 						*/
	word * old_normal_bl;
		/* Nonstack false references seen at last full		*/
		/* collection.						*/
	word * incomplete_normal_bl;
		/* Nonstack false references seen since last		*/
		/* full collection.					*/
	word * old_stack_bl;
	word * incomplete_stack_bl;
	word total_black_listed;
			/* Number of bytes on stack blacklist. 	*/
	word black_list_spacing;

			/* Average number of bytes between blacklisted	*/
			/* blocks. Approximate.				*/
	PSM_mutex_t ml;
    } bl;

    PSM_barrier_t gcp_barrier;
    PSM_barrier_t gcp_exit_barrier;
    PSM_barrier_t appp_barrier;
    sh_bsbar_t appp_bsbar;

    /* one of GC/SPMD processes. only valid during a collection. */
    GC_thread gcleader_envp;

    sh_bssem_t sem;
    sh_blocksync_t blocksync;

    int gcp_command; /* tells what GC processes should do. (collect/exit) */
    int spmdp_command;
    GC_thread requestor_envp;
    GC_thread_id mainpid; /* Main process (thread) ID. */
#ifdef SDSM_TEST
    int current_pe;
    int current_userpe;
#endif

    /* structure for debug log (global) */
#ifdef DO_MARKLOG
    struct g_marklog *marklogp;
#endif
#ifdef DO_PICLOG
    struct g_piclog *piclogp;
#endif
#ifdef DO_LDLOG
    struct g_ldlog *ldlogp;
#endif
#ifdef DO_ALLOCLOG
    struct g_alloclog *alloclogp;
#endif
#ifdef DO_GCRLOG
    struct g_gcrlog *gcrlogp;
#endif
#ifdef DO_MSLOG
    struct g_mslog *mslogp;
#endif
#ifdef USE_DEMOWIN
    struct demowin *demowinp;
#endif
#ifdef PROCESSOR_BIND_SUPPORTED
    struct {
	int count;
	PSM_mutex_t ml;
    } bind_support;
#endif /* PROCESSOR_BIND_SUPPORTED */
#if defined SGC_TLS_IMPORT_PUBLIC_FUNCTION || \
    defined SGC_TLS_IMPORT_EACH_FUNCTION
    int tls_key;
#endif
    struct {
	unsigned int finalization_failures;
	/* Number of finalization requests that failed for lack of memory. */

	/* static */
	struct disappearing_link **dl_head;

	/* static */
	signed_word log_dl_table_size;
	/* Binary log of				*/
	/* current size of array pointed to by dl_head.	*/
	/* -1 ==> size is 0.				*/

	word dl_entries;/* Number of entries currently in disappearing	*/
	                /* link table.					*/

	PSM_mutex_t dl_ml;

	/* static */
	struct finalizable_object **fo_head;

	struct finalizable_object * finalize_now;
	/* List of objects that should be finalized now.	*/

	/* static */
	signed_word log_fo_table_size;

	PSM_mutex_t fo_ml;

	word fo_entries;
    } fo;
};

extern GC_FAR struct _GC_arrays GC_arrays;
extern struct _GC_arrays * GC_arraysp;

/**** local (common) data */
/* definition of GC_my_envp is in gc_tls.h */

# define GC_my_pid GC_my_envp->pid
# define GC_my_debug_state GC_my_envp->debug_state
# define GC_my_app_or_gc GC_my_envp->app_or_gc
# define GC_my_scratch GC_my_envp->scratch
# define GC_l_pid(envp) envp->pid
# define GC_l_debug_state(envp) envp->debug_state
# define GC_l_app_or_gc(envp) envp->app_or_gc
# define GC_l_scratch(envp) envp->scratch
# define GC_my_sem GC_my_envp->sem
# define GC_l_sem(envp) envp->sem
# define GC_my_initialized GC_my_envp->initialized
# define GC_l_initialized(envp) envp->initialized
# define GC_my_terminated GC_my_envp->terminated
# define GC_l_terminated(envp) envp->terminated
# define GC_my_mark_stack GC_my_envp->mark_stack
# define GC_l_mark_stack(envp) envp->mark_stack
# define GC_my_mark_stack_too_small GC_my_envp->_mark_stack_too_small
# define GC_l_mark_stack_too_small(envp) envp->_mark_stack_too_small
# define GC_my_mark_state GC_my_envp->_mark_state
# define GC_my_objects_are_marked GC_my_envp->_objects_are_marked
# define GC_my_marklogp GC_my_envp->marklogp
# define GC_l_marklogp(envp) envp->marklogp
# define GC_my_piclogp GC_my_envp->piclogp
# define GC_l_piclogp(envp) envp->piclogp
# define GC_my_cachelogp GC_my_envp->cachelogp
# define GC_l_cachelogp(envp) envp->cachelogp

/**** local (app thread) data */
# define GC_mya_appenvp (&GC_my_envp->app) /* Used by App process */
# define GC_la_appenvp(envp) (&envp->app)

# define GC_mya_app_stat GC_mya_appenvp->stat
# define GC_la_app_stat(envp) GC_la_appenvp(envp)->stat

# define GC_mya_gcregvar GC_mya_appenvp->gcregvar
# define GC_la_gcregvar(envp) GC_la_appenvp(envp)->gcregvar
# define GC_mya_stackbottom GC_mya_appenvp->_stackbottom
# define GC_la_stackbottom(envp) GC_la_appenvp(envp)->_stackbottom

# define GC_mya_freelist GC_mya_appenvp->freelist
# define GC_mya_reclaim_list GC_mya_appenvp->reclaim_list
# define GC_la_freelist(envp) GC_la_appenvp(envp)->freelist
# define GC_la_reclaim_list(envp) GC_la_appenvp(envp)->reclaim_list

# define GC_mya_pr (GC_mya_appenvp->pr)
# define GC_la_pr(envp) (GC_la_appenvp(envp)->pr)

# define GC_mya_flags GC_mya_appenvp->flags
# define GC_la_flags(envp) GC_la_appenvp(envp)->flags
# define GC_mya_polling_flag GC_mya_appenvp->polling_flag
# define GC_la_polling_flag(envp) GC_la_appenvp(envp)->polling_flag
# define GC_mya_alloc_ml GC_mya_appenvp->alloc_ml
# define GC_la_alloc_ml(envp) GC_la_appenvp(envp)->alloc_ml
# define GC_mya_barrier_stage GC_mya_appenvp->barrier_stage
# define GC_la_barrier_stage(envp) GC_la_appenvp(envp)->barrier_stage

# define GC_mya_static_roots GC_mya_appenvp->static_roots
# define GC_mya_n_root_sets GC_mya_appenvp->n_root_sets
# define GC_la_n_root_sets(envp) GC_la_appenvp(envp)->n_root_sets

# define GC_mya_rootcopy GC_mya_appenvp->rootcopy
# define GC_la_rootcopy(envp) GC_la_appenvp(envp)->rootcopy
# define GC_mya_alloclogp GC_mya_appenvp->alloclogp
# define GC_la_alloclogp(envp) GC_la_appenvp(envp)->alloclogp

/**** local (gc thread) data */
# define GC_lg_envp(pno) GC_g_gcproc_envps[pno]

# define GC_myg_gcenvp (&GC_my_envp->gc) /* Used by GC process */
# define GC_lg_gcenvp(envp) (&envp->gc)
# define GC_myg_pno GC_myg_gcenvp->pno
# define GC_lg_pno(envp) GC_lg_gcenvp(envp)->pno

# define GC_myg_gc_stat GC_myg_gcenvp->stat
# define GC_lg_gc_stat(envp) GC_lg_gcenvp(envp)->stat

# define GC_myg_sem GC_myg_gcenvp->sem
# define GC_lg_sem(envp) GC_lg_gcenvp(envp)->sem

/* For load balancing */
# define GC_lg_ld(envp) (GC_lg_gcenvp(envp)->ld)
# define GC_myg_ld (GC_myg_gcenvp->ld)
# define GC_lg_smq(envp) GC_lg_ld(envp).smq
# define GC_myg_smq GC_myg_ld.smq

# define GC_myg_pr (GC_myg_gcenvp->pr)
# define GC_lg_pr(envp) (GC_lg_gcenvp(envp)->pr)

# ifdef SGC_BALANCED_HDR
#   define GC_myg_hdr_free_list GC_myg_gcenvp->_hdr_free_list
#   define GC_myg_hdr_free_list_ml GC_myg_gcenvp->_hdr_free_list_ml
#   define GC_myg_n_free_hdr GC_myg_gcenvp->_n_free_hdr
#   define GC_lg_hdr_free_list(envp) GC_lg_gcenvp(envp)->_hdr_free_list
#   define GC_lg_hdr_free_list_ml(envp) GC_lg_gcenvp(envp)->_hdr_free_list_ml
#   define GC_lg_n_free_hdr(envp) GC_lg_gcenvp(envp)->_n_free_hdr
# endif

# ifdef DO_REQ_REMOTE
#   define GC_myg_reqbuf GC_myg_gcenvp->reqbuf
#   define GC_lg_reqbuf(envp) GC_lg_gcenvp(envp)->reqbuf
# endif

# define GC_myg_spmd_start_flag GC_myg_gcenvp->spmd_start_flag
# define GC_lg_spmd_start_flag(envp) GC_lg_gcenvp(envp)->spmd_start_flag

# define GC_myg_ldlogp GC_myg_gcenvp->ldlogp
# define GC_lg_ldlogp(envp) GC_lg_gcenvp(envp)->ldlogp
# define GC_myg_mslogp GC_myg_gcenvp->mslogp
# define GC_lg_mslogp(envp) GC_lg_gcenvp(envp)->mslogp

/**** global data */
# define GC_fail_count GC_arraysp->fail_count
# define GC_obj_kinds GC_arraysp->_obj_kinds

# define GC_valid_offsets GC_arraysp->_valid_offsets
# define GC_modws_valid_offsets GC_arraysp->_modws_valid_offsets
# ifdef STUBBORN_ALLOC
#    define GC_changed_pages GC_arraysp->_changed_pages
#    define GC_prev_changed_pages GC_arraysp->_prev_changed_pages
# endif

# define GC_obj_map GC_arraysp->_obj_map
# define GC_invalid_map GC_arraysp->_invalid_map
# define GC_obj_map_ml GC_arraysp->_obj_map_ml

# define GC_last_heap_addr GC_arraysp->_last_heap_addr
# define GC_prev_heap_addr GC_arraysp->_prev_heap_addr

# define GC_sum_gc_stat GC_arraysp->sum_gc_stat
# define GC_sum_app_stat GC_arraysp->sum_app_stat

/* for compatibility */
# define GC_words_allocd GC_sum_app_stat.words_allocd
# define GC_words_allocd_before_gc GC_sum_app_stat.words_allocd_before_gc

# define GC_heapsize GC_arraysp->_heapsize
# define GC_max_heapsize GC_arraysp->_max_heapsize

# define GC_heap_sects GC_arraysp->_heap_sects
# define GC_heap_sects_ml GC_arraysp->_heap_sects_ml
# define GC_last_stack GC_arraysp->_last_stack
# ifdef MSWIN32
#   define GC_heap_bases GC_arraysp->_heap_bases
# endif
# define GC_all_nils GC_arraysp->_all_nils
# define GC_top_index GC_arraysp->_top_index

# if defined(PROC_VDB) || defined(MPROTECT_VDB)
#   define GC_grungy_pages GC_arraysp->_grungy_pages
# endif
# ifdef MERGE_SIZES
#   define GC_size_map GC_arraysp->_size_map
#   define GC_size_map_ml GC_arraysp->_size_map_ml
# endif

# define GC_all_bottom_indices GC_arraysp->_all_bottom_indices

# define GC_hblkfreelist GC_arraysp->_hblkfreelist
# define GC_savhbp GC_arraysp->_savhbp
# define GC_freehblk_ptr GC_arraysp->_freehblk_ptr

# define beginGC_arrays ((ptr_t)(GC_arraysp))
# define endGC_arrays (((ptr_t)(GC_arraysp)) + (sizeof GC_arrays))

# define GC_n_heap_sects GC_arraysp->_n_heap_sects
# define GC_root_size GC_arraysp->_root_size
# define GC_least_plausible_heap_addr GC_arraysp->_least_plausible_heap_addr
# define GC_greatest_plausible_heap_addr GC_arraysp->_greatest_plausible_heap_addr

# define GC_scan_ptr GC_arraysp->_scan_ptr
# define GC_scan_ptr_flag GC_arraysp->_scan_ptr_flag
# define GC_scan_ptr_ml GC_arraysp->_scan_ptr_ml

# define GC_old_normal_bl GC_arraysp->bl.old_normal_bl
# define GC_incomplete_normal_bl GC_arraysp->bl.incomplete_normal_bl
# define GC_old_stack_bl GC_arraysp->bl.old_stack_bl
# define GC_incomplete_stack_bl GC_arraysp->bl.incomplete_stack_bl
# define GC_total_black_listed GC_arraysp->bl.total_black_listed
# define GC_black_list_spacing GC_arraysp->bl.black_list_spacing
# define GC_black_list_ml GC_arraysp->bl.ml

# define GC_leader_stackbottom GC_arraysp->leader_stackbottom
# define GC_np_fixed GC_arraysp->np_fixed
# define GC_use_polling GC_arraysp->use_polling
# define GC_blocksync GC_arraysp->blocksync
# define GC_gcp_command GC_arraysp->gcp_command
# define GC_spmdp_command GC_arraysp->spmdp_command
# define GC_gcp_barrier GC_arraysp->gcp_barrier
# define GC_gcp_exit_barrier GC_arraysp->gcp_exit_barrier
# define GC_appp_barrier GC_arraysp->appp_barrier
# define GC_appp_bsbar GC_arraysp->appp_bsbar
# define GC_g_gcleader_envp GC_arraysp->gcleader_envp
# define GC_requestor_envp GC_arraysp->requestor_envp
# define GC_g_mainpid GC_arraysp->mainpid
# define GC_gcp_exit_barrier GC_arraysp->gcp_exit_barrier
# define GC_g_sem GC_arraysp->sem
# define GC_g_join_sem GC_arraysp->join_sem
# define GC_g_create_sem GC_arraysp->create_sem
# define GC_g_alloc_ml GC_arraysp->alloc_ml
# define GC_g_scratch GC_arraysp->scratch
# define GC_spmd_start_count GC_arraysp->spmd_start_count
# define GC_spmd_func GC_arraysp->spmd_func
# define GC_spmd_arg GC_arraysp->spmd_arg
# define GC_thread_funcarg GC_arraysp->thread_funcarg

# define GC_g_ld (GC_arraysp->ld) /* For load balancing */
# define GC_g_pr (GC_arraysp->pr)
# define GC_g_debug (GC_arraysp->debug)
# define GC_g_shm GC_arraysp->shm

# define GC_g_reclaim_list GC_arraysp->reclaim_list
# define GC_g_reclaim_list_ml GC_arraysp->reclaim_list_ml
# define GC_g_freelist GC_arraysp->freelist
# define GC_g_freelist_ml GC_arraysp->freelist_ml

# ifndef SGC_BALANCED_HDR
#   define GC_hdr_free_list GC_arraysp->_hdr_free_list
# endif

# define GC_g_marklogp GC_arraysp->marklogp
# define GC_g_gcrlogp GC_arraysp->gcrlogp
# define GC_g_ldlogp GC_arraysp->ldlogp
# define GC_g_alloclogp GC_arraysp->alloclogp
# define GC_g_piclogp GC_arraysp->piclogp
# define GC_g_mslogp GC_arraysp->mslogp
# define GC_g_demowinp GC_arraysp->demowinp
# define GC_g_bind_support GC_arraysp->bind_support

# define GC_g_appproc_envps GC_arraysp->appproc_envps
# define GC_g_gcproc_envps GC_arraysp->gcproc_envps

#define GC_g_log_dl_table_size GC_arraysp->fo.log_dl_table_size
#define GC_g_dl_head GC_arraysp->fo.dl_head
#define GC_g_dl_entries GC_arraysp->fo.dl_entries
#define GC_g_dl_ml GC_arraysp->fo.dl_ml
#define GC_g_log_fo_table_size GC_arraysp->fo.log_fo_table_size
#define GC_g_fo_head GC_arraysp->fo.fo_head
#define GC_g_fo_entries GC_arraysp->fo.fo_entries
#define GC_g_fo_ml GC_arraysp->fo.fo_ml
#define GC_g_finalization_failures GC_arraysp->fo.finalization_failures
#define GC_g_finalize_now GC_arraysp->fo.finalize_now

extern int GC_n_kinds;

# ifdef MSWIN32
extern word GC_n_heap_bases;	/* See GC_heap_bases.	*/
# endif

extern GC_bool GC_is_initialized;		/* GC_init() has been run.	*/

extern int GC_incremental;  /* Using incremental/generational collection. */

extern GC_bool GC_dirty_maintained;/* Dirty bits are being maintained, 	*/
				/* either for incremental collection,	*/
				/* or to limit the root set.		*/

extern GC_bool GC_debugging_started;	/* GC_debug_malloc has been called. */ 


/*  Marks are in a reserved area in                          */
/*  each heap block.  Each word has one mark bit associated  */
/*  with it. Only those corresponding to the beginning of an */
/*  object are used.                                         */

/* Mark bit perations */

/*
 * Retrieve, set, clear the mark bit corresponding
 * to the nth word in a given heap block.
 *
 * (Recall that bit n corresponds to object beginning at word n
 * relative to the beginning of the block, including unused words)
 */
/* MT-Unsafe */

# define mark_bit_from_hdr(hhdr,n) (((hhdr)->hb_marks[divWORDSZ(n)] \
			    >> (modWORDSZ(n))) & (word)1)
# define set_mark_bit_from_hdr(hhdr,n) (hhdr)->hb_marks[divWORDSZ(n)] \
				|= (word)1 << modWORDSZ(n)

# define clear_mark_bit_from_hdr(hhdr,n) (hhdr)->hb_marks[divWORDSZ(n)] \
				&= ~((word)1 << modWORDSZ(n))

/* Important internal collector routines */

void GC_apply_to_all_blocks GC_PROTO((void (*fn)(struct hblk *h, word client_data), word client_data));
			/* Invoke fn(hbp, client_data) for each 	*/
			/* allocated heap block.			*/
struct hblk * GC_next_block GC_PROTO((struct hblk * h));
/* void GC_mark_init(); */
void GC_clear_marks GC_PROTO((SGC_MYENV_ARG_DECL0));	/* Clear mark bits for all heap objects. */
void GC_invalidate_mark_state GC_PROTO((SGC_MYENV_ARG_DECL0));	/* Tell the marker that	marked 	   */
					/* objects may point to	unmarked   */
					/* ones, and roots may point to	   */
					/* unmarked objects.		   */
					/* Reset mark stack.		   */
void GC_mark_from_mark_stack GC_PROTO((SGC_MYENV_ARG_DECL0)); /* Mark from everything on the mark stack. */
				/* Return after about one pages worth of   */
				/* work.				   */
GC_bool GC_mark_stack_empty GC_PROTO((SGC_MYENV_ARG_DECL0));
GC_bool GC_mark_some GC_PROTO((SGC_MYENV_ARG_DECL0));	/* Perform about one pages worth of marking	*/
			/* work of whatever kind is needed.  Returns	*/
			/* quickly if no collection is in progress.	*/
			/* Return TRUE if mark phase finished.		*/
void GC_push_all GC_PROTO((SGC_MYENV_ARG_DECL1 ptr_t b, ptr_t t));	/* Push everything in a range 		*/
				/* onto mark stack.			*/
void GC_push_dirty GC_PROTO((SGC_MYENV_ARG_DECL1 ptr_t b, ptr_t t,
			     int (*dirty_fn)(struct hblk * h),
			     void (*push_fn)(SGC_MYENV_ARG_DECL1 ptr_t bottom, ptr_t top)));      /* Push all possibly changed	 	*/
				  /* subintervals of [b,t) onto		*/
				  /* mark stack.			*/
#ifndef SMALL_CONFIG
  void GC_push_conditional GC_PROTO((SGC_MYENV_ARG_DECL1 ptr_t bottom, ptr_t top, int all)); 
#else
# define GC_push_conditional(b, t, all) GC_push_all(b, t)
#endif
                                /* Do either of the above, depending	*/
				/* on the third arg.			*/
void GC_push_all_stack GC_PROTO((SGC_MYENV_ARG_DECL1 ptr_t bottom, ptr_t top));
				    /* As above, but consider		*/
				    /*  interior pointers as valid  	*/
void GC_push_roots GC_PROTO((SGC_MYENV_ARG_DECL0)); 
/* void GC_push_roots( GC_bool all ); */ /* Push all or dirty roots.	*/
extern void (*GC_push_other_roots) GC_PROTO((SGC_MYENV_ARG_DECL0));
			/* Push system or application specific roots	*/
			/* onto the mark stack.  In some environments	*/
			/* (e.g. threads environments) this is		*/
			/* predfined to be non-zero.  A client supplied */
			/* replacement should also call the original	*/
			/* function.					*/
void GC_push_regs GC_PROTO((void)); 
			/* Push register contents onto mark stack.	*/
/*void GC_remark();*/	/* Mark from all marked objects.  Used	*/
		 	/* only if we had to drop something.	*/
void GC_push_one GC_PROTO((word p)); 
				/* If p points to an object, mark it	*/
				/* and push contents on the mark stack	*/
void GC_push_one_checked GC_PROTO((SGC_MYENV_ARG_DECL1 register word p, register GC_bool interior_ptrs)); 
/*void GC_push_one_checked(p); *//* Ditto, omits plausibility test	*/
void GC_push_marked GC_PROTO((SGC_MYENV_ARG_DECL1 struct hblk *h, register hdr *hhdr)); 
		/* Push contents of all marked objects in h onto	*/
		/* mark stack.						*/
#ifdef SMALL_CONFIG
# define GC_push_next_marked_dirty(h) GC_push_next_marked(h)
#else
struct hblk *GC_push_next_marked_dirty GC_PROTO((SGC_MYENV_ARG_DECL1 struct hblk *h)); 
		/* Invoke GC_push_marked on next dirty block above h.	*/
		/* Return a pointer just past the end of this block.	*/
#endif /* !SMALL_CONFIG */
struct hblk *GC_push_next_marked GC_PROTO((SGC_MYENV_ARG_DECL1 struct hblk *h)); 
		/* Ditto, but also mark from clean pages.	*/
struct hblk *GC_push_next_marked_uncollectable GC_PROTO((SGC_MYENV_ARG_DECL1 struct hblk *h)); 
		/* Ditto, but mark only from uncollectable pages.	*/
/*GC_bool GC_stopped_mark(); *//* Stop world and mark from all roots	*/
			/* and rescuers.			*/
void GC_clear_hdr_marks GC_PROTO((register hdr *hhdr)); /* Clear the mark bits in a header */
void GC_add_roots_inner GC_PROTO((SGC_MYENV_ARG_DECL1 char *b, char *e, GC_bool tmp)); 
GC_bool GC_is_static_root GC_PROTO((SGC_MYENV_ARG_DECL1 ptr_t p)); 
		/* Is the address p in one of the registered static	*/
		/* root sections?					*/
void GC_register_dynamic_libraries GC_PROTO((SGC_MYENV_ARG_DECL0)); 
		/* Add dynamic library data sections to the root set. */

/* Machine dependent startup routines */
ptr_t GC_get_stack_base GC_PROTO((void)); 
void GC_register_data_segments GC_PROTO((SGC_MYENV_ARG_DECL0)); 

/* Black listing: */
void GC_bl_init GC_PROTO((void)); 
# ifndef ALL_INTERIOR_POINTERS
    void GC_add_to_black_list_normal GC_PROTO((word p)); 
			/* Register bits as a possible future false	*/
			/* reference from the heap or static data	*/
#   define GC_ADD_TO_BLACK_LIST_NORMAL(bits) GC_add_to_black_list_normal(bits)
# else
#   define GC_ADD_TO_BLACK_LIST_NORMAL(bits) GC_add_to_black_list_stack(bits)
# endif

void GC_add_to_black_list_stack GC_PROTO((word p)); 
struct hblk *GC_is_black_listed GC_PROTO((struct hblk *h, word len)); 
			/* If there are likely to be false references	*/
			/* to a block starting at h of the indicated    */
			/* length, then return the next plausible	*/
			/* starting location for h that might avoid	*/
			/* these false references.			*/
void GC_promote_black_lists GC_PROTO((void)); 
			/* Declare an end to a black listing phase.	*/
void GC_unpromote_black_lists GC_PROTO((void)); 
			/* Approximately undo the effect of the above.	*/
			/* This actually loses some information, but	*/
			/* only in a reasonably safe way.		*/
word GC_number_stack_black_listed GC_PROTO((struct hblk *start, struct hblk *endp1)); 
			/* Return the number of (stack) blacklisted	*/
			/* blocks in the range for statistical		*/
			/* purposes.					*/
		 	
ptr_t GC_scratch_alloc GC_PROTO((SGC_MYENV_ARG_DECL1 register word bytes)); 
				/* GC internal memory allocation for	*/
				/* small objects.  Deallocation is not  */
				/* possible.				*/
ptr_t GC_scratch_g_alloc GC_PROTO((register word bytes)); 
	
/* Heap block layout maps: */			
void GC_invalidate_map GC_PROTO((hdr *hhdr)); 
				/* Remove the object map associated	*/
				/* with the block.  This identifies	*/
				/* the block as invalid to the mark	*/
				/* routines.				*/
GC_bool GC_add_map_entry GC_PROTO((word sz)); 
				/* Add a heap block map for objects of	*/
				/* size sz to obj_map.			*/
				/* Return FALSE on failure.		*/
void GC_register_displacement_inner GC_PROTO((word offset)); 
				/* Version of GC_register_displacement	*/
				/* that assumes lock is already held	*/
				/* and signals are already disabled.	*/

/*  hblk allocation: */		
void GC_new_hblk GC_PROTO((SGC_MYENV_ARG_DECL1 register word sz, int kind)); 
				/* Allocate a new heap block, and build */
				/* a free list in it.			*/				
struct hblk *GC_allochblk GC_PROTO((SGC_MYENV_ARG_DECL1 word sz, int kind, unsigned char flags)); 
				/* Allocate a heap block, clear it if	*/
				/* for composite objects, inform	*/
				/* the marker that block is valid	*/
				/* for objects of indicated size.	*/
				/* sz < 0 ==> atomic.			*/ 
void GC_freehblk GC_PROTO((register struct hblk *p)); 
				/* Deallocate a heap block and mark it  */
				/* as invalid.				*/
				
/*  Misc GC: */
/*void GC_init_inner();*/
GC_bool GC_expand_hp_inner GC_PROTO((SGC_MYENV_ARG_DECL1 word n)); 
/*void GC_start_reclaim(abort_if_found);*/
				/* Restore unmarked objects to free	*/
				/* lists, or (if abort_if_found is	*/
				/* TRUE) report them.			*/
				/* Sweeping of small object pages is	*/
				/* largely deferred.			*/
void GC_continue_reclaim GC_PROTO((SGC_MYENV_ARG_DECL1 word sz, int kind)); 
				/* Sweep pages of the given size and	*/
				/* kind, as long as possible, and	*/
				/* as long as the corr. free list is    */
				/* empty.				*/
void GC_reclaim_or_delete_all GC_PROTO((SGC_MYENV_ARG_DECL0)); 
				/* Arrange for all reclaim lists to be	*/
				/* empty.  Judiciously choose between	*/
				/* sweeping and discarding each page.	*/
/*GC_bool GC_reclaim_all(GC_stop_func f);*/
				/* Reclaim all blocks.  Abort (in a	*/
				/* consistent state) if f returns TRUE. */
GC_bool GC_block_empty GC_PROTO((register hdr *hhdr)); /* Block completely unmarked? 	*/
GC_bool GC_never_stop_func GC_PROTO((void)); /* Returns FALSE.		*/
/*GC_bool GC_try_to_collect_inner(GC_stop_func \f);*/
				/* Collect; caller must have acquired	*/
				/* lock and disabled signals.		*/
				/* Collection is aborted if f returns	*/
				/* TRUE.  Returns TRUE if it completes	*/
				/* successfully.			*/
# define GC_gcollect_inner() \
	(void) GC_try_to_collect_inner(GC_never_stop_func)
/*void GC_finish_collection();*/	/* Finish collection.  Mark bits are	*/
				/* consistent and lock is still held.	*/
GC_bool GC_collect_or_expand GC_PROTO((SGC_MYENV_ARG_DECL1 word needed_blocks, GC_bool ignore_off_page)); 
				/* Collect or expand heap in an attempt */
				/* make the indicated number of free	*/
				/* blocks available.  Should be called	*/
				/* until the blocks are available or	*/
				/* until it fails by returning FALSE.	*/
void GC_init GC_PROTO((void));			/* Initialize collector.		*/
/*void GC_collect_a_little_inner(int n); */
				/* Do n units worth of garbage 		*/
				/* collection work, if appropriate.	*/
				/* A unit is an amount appropriate for  */
				/* HBLKSIZE bytes of allocation.	*/
ptr_t GC_generic_malloc GC_PROTO((SGC_MYENV_ARG_DECL1 register word lb, register int k)); 
				/* Allocate an object of the given	*/
				/* kind.  By default, there are only	*/
				/* two kinds: composite, and atomic.	*/
				/* We claim it's possible for clever	*/
				/* client code that understands GC	*/
				/* internals to add more, e.g. to	*/
				/* communicate object layout info	*/
				/* to the collector.			*/
ptr_t GC_generic_malloc_inner GC_PROTO((SGC_MYENV_ARG_DECL1 register word lb, register int k)); 
				/* Ditto, but I already hold lock, etc.	*/
ptr_t GC_generic_malloc_words_small GC_PROTO((SGC_MYENV_ARG_DECL1 register word lw, register int k)); 
				/* As above, but size in units of words */
				/* Bypasses MERGE_SIZES.  Assumes	*/
				/* words <= MAXOBJSZ.			*/
ptr_t GC_generic_malloc_inner_ignore_off_page GC_PROTO((SGC_MYENV_ARG_DECL1 register size_t lb, register int k)); 
				/* Allocate an object, where		*/
				/* the client guarantees that there	*/
				/* will always be a pointer to the 	*/
				/* beginning of the object while the	*/
				/* object is live.			*/
ptr_t GC_allocobj GC_PROTO((SGC_MYENV_ARG_DECL1 word sz, int kind)); 
				/* Make the indicated 			*/
				/* free list nonempty, and return its	*/
				/* head.				*/

void GC_init_headers GC_PROTO((void)); 
GC_bool GC_install_header GC_PROTO((SGC_MYENV_ARG_DECL1 register struct hblk *h)); 
				/* Install a header for block h.	*/
				/* Return FALSE on failure.		*/
GC_bool GC_install_counts GC_PROTO((SGC_MYENV_ARG_DECL1 register struct hblk *h, register word sz)); 
				/* Set up forwarding counts for block	*/
				/* h of size sz.			*/
				/* Return FALSE on failure.		*/
void GC_remove_header GC_PROTO((register struct hblk *h)); 
				/* Remove the header for block h.	*/
void GC_remove_counts GC_PROTO((register struct hblk *h, register word sz)); 
				/* Remove forwarding counts for h.	*/
hdr *GC_find_header GC_PROTO((ptr_t h)); 
			/* Debugging only.			*/

void GC_finalize_gc GC_PROTO((SGC_MYENV_ARG_DECL0)); 
			/* Perform all indicated finalization actions	*/
			/* on unmarked objects.				*/
			/* Unreachable finalizable objects are enqueued	*/
			/* for processing by GC_invoke_finalizers.	*/
			/* Invoked with lock.				*/
int GC_invoke_finalizers GC_PROTO((void)); 
			 	/* Run eligible finalizers.	*/
				/* Invoked without lock.	*/	
			
void GC_add_to_heap GC_PROTO((SGC_MYENV_ARG_DECL1 struct hblk *p, word bytes)); 
			/* Add a HBLKSIZE aligned chunk to the heap.	*/

void GC_print_obj GC_PROTO((ptr_t p)); 
			/* P points to somewhere inside an object with	*/
			/* debugging info.  Print a human readable	*/
			/* description of the object to stderr.		*/
extern void (*GC_check_heap) GC_PROTO((void));
			/* Check that all objects in the heap with 	*/
			/* debugging info are intact.  Print 		*/
			/* descriptions of any that are not.		*/

/* Virtual dirty bit implementation:		*/
/* Each implementation exports the following:	*/
void GC_read_dirty GC_PROTO((SGC_MYENV_ARG_DECL0)); /* Retrieve dirty bits.	*/
GC_bool GC_page_was_dirty GC_PROTO((struct hblk *h)); 
			/* Read retrieved dirty bits.	*/
GC_bool GC_page_was_ever_dirty GC_PROTO((struct hblk *h)); 
			/* Could the page contain valid heap pointers?	*/
void GC_is_fresh GC_PROTO((struct hblk *h, word n)); 
			/* Assert the region currently contains no	*/
			/* valid pointers.				*/
void GC_write_hint GC_PROTO((struct hblk *h)); 
			/* h is about to be written.	*/
void GC_dirty_init GC_PROTO((SGC_MYENV_ARG_DECL0)); 

/* Slow/general mark bit manipulation: */
GC_bool GC_is_marked GC_PROTO((ptr_t p)); 
void GC_clear_mark_bit GC_PROTO((ptr_t p)); 
void GC_set_mark_bit GC_PROTO((ptr_t p)); 

/* Stubborn objects: */
/*void GC_read_changed();*/	/* Analogous to GC_read_dirty */
/*GC_bool GC_page_was_changed(h);*/	/* Analogous to GC_page_was_dirty */
/*void GC_clean_changing_list();*/	/* Collect obsolete changing list entries */
/*void GC_stubborn_init();*/

/* Debugging print routines: */
void GC_print_block_list GC_PROTO((void)); 
void GC_print_hblkfreelist GC_PROTO((void)); 
void GC_print_heap_sects GC_PROTO((void)); 
void GC_print_static_roots GC_PROTO((SGC_MYENV_ARG_DECL0)); 
void GC_dump GC_PROTO((void)); 

/* Make arguments appear live to compiler */
void GC_noop GC_PROTO((void *dummy, ...)); 

/* Logging and diagnostic output: 	*/
void GC_printf GC_PROTO((char * format, long, long, long, long, long, long));
			/* A version of printf that doesn't allocate,	*/
			/* is restricted to long arguments, and		*/
			/* (unfortunately) doesn't use varargs for	*/
			/* portability.  Restricted to 6 args and	*/
			/* 1K total output length.			*/
			/* (We use sprintf.  Hopefully that doesn't	*/
			/* allocate for long arguments.)  		*/
# define GC_printf0(f) GC_printf(f, 0l, 0l, 0l, 0l, 0l, 0l)
# define GC_printf1(f,a) GC_printf(f, (long)a, 0l, 0l, 0l, 0l, 0l)
# define GC_printf2(f,a,b) GC_printf(f, (long)a, (long)b, 0l, 0l, 0l, 0l)
# define GC_printf3(f,a,b,c) GC_printf(f, (long)a, (long)b, (long)c, 0l, 0l, 0l)
# define GC_printf4(f,a,b,c,d) GC_printf(f, (long)a, (long)b, (long)c, \
					    (long)d, 0l, 0l)
# define GC_printf5(f,a,b,c,d,e) GC_printf(f, (long)a, (long)b, (long)c, \
					      (long)d, (long)e, 0l)
# define GC_printf6(f,a,b,c,d,e,g) GC_printf(f, (long)a, (long)b, (long)c, \
						(long)d, (long)e, (long)g)

void GC_err_printf GC_PROTO((char *format, long int a, long int b, long int c, long int d, long int e, long int f)); 
# define GC_err_printf0(f) GC_err_puts(f)
# define GC_err_printf1(f,a) GC_err_printf(f, (long)a, 0l, 0l, 0l, 0l, 0l)
# define GC_err_printf2(f,a,b) GC_err_printf(f, (long)a, (long)b, 0l, 0l, 0l, 0l)
# define GC_err_printf3(f,a,b,c) GC_err_printf(f, (long)a, (long)b, (long)c, \
						  0l, 0l, 0l)
# define GC_err_printf4(f,a,b,c,d) GC_err_printf(f, (long)a, (long)b, \
						    (long)c, (long)d, 0l, 0l)
# define GC_err_printf5(f,a,b,c,d,e) GC_err_printf(f, (long)a, (long)b, \
						      (long)c, (long)d, \
						      (long)e, 0l)
# define GC_err_printf6(f,a,b,c,d,e,g) GC_err_printf(f, (long)a, (long)b, \
							(long)c, (long)d, \
							(long)e, (long)g)
			/* Ditto, writes to stderr.			*/
			
void GC_err_puts GC_PROTO((char *s));
			/* Write s to stderr, don't buffer, don't add	*/
			/* newlines, don't ...				*/

#include "proto.h"

# endif /* GC_PRIVATE_H */
