#!/bin/sh
#
# makehelper.sh: make helper script.
#
# Some input environment variables must have been defined.
# For details, see the definition of MKHELPER in Makefile.
#
# [usage]
#   makehelper.sh var1 [var2 ...]
#
# This script outputs the values of specified variables(var1...)
# to stdout.
#

C_OPT_FLAGS=-O2
#C_OPT_FLAGS=

C_DEBUG_FLAGS=-g
#C_DEBUG_FLAGS=
#-xpg -pg for profiling

C_WARN_FLAGS=
#C_WARN_FLAGS=-Wall -Wstrict-prototypes -Wmissing-prototypes -Wmissing-declarations -Wshadow
#C_WARN_FLAGS=-fullwarn # for mips cc



GCCVER=`$CC -v 2>&1 | grep gcc 2>/dev/null`
if [ "$GCCVER" != "" ]; then
  ISGCC=1
else
  ISGCC=0
fi

SGC_OS_TYPE=`uname -s`
SGC_CPU_TYPE=`uname -p`

if [ "$SGC_CPU_TYPE" = "unknown" ]; then
# On some OS(typically Linux), uname -p doesn't return meaningful result.
  if [ `uname -m` = "i686" ]; then
    SGC_CPU_TYPE="i386"
  fi
fi

if [ "$SGC_CPU_TYPE" = "sparc" ]; then
  if [ "$ISGCC" = "1" ]; then
    if [ "$SGC_WITH_STHREADS" = "NO" ]; then
      C_REGVAR_FLAGS=-ffixed-%g6
    else
      C_REGVAR_FLAGS=-ffixed-%l7\ -mflat
    fi
    C_ABI_FLAGS=-Wa,-xarch=v8plusa
  else
    C_OPT_FLAGS=-xO2
    C_REGVAR_FLAGS=
    C_ABI_FLAGS=-xarch=v8plusa
  fi
  ASM_ABI_FLAGS=-xarch=v8plusa
elif [ "$SGC_CPU_TYPE" = "mips" ]; then
  if [ "$ISGCC" = "1" ]; then
    C_REGVAR_FLAGS=-ffixed-%23
    if [ "$SGC_OBJECT_64BIT" = "YES" ]; then
      C_ABI_FLAGS=-mabi=64
      ASM_ABI_FLAGS=-mips4\ -64
    else
      C_ABI_FLAGS=-mips2
      ASM_ABI_FLAGS=-mips2
    fi
  else
    C_DEBUG_FLAGS=-g3
    C_REGVAR_FLAGS=
    if [ "$SGC_OBJECT_64BIT" = "YES" ]; then
      C_ABI_FLAGS=-mips4\ -64
      ASM_ABI_FLAGS=-mips4\ -64
    else
      C_ABI_FLAGS=-mips2
      ASM_ABI_FLAGS=-mips2
    fi
  fi
elif [ "$SGC_CPU_TYPE" = "i386" ]; then
  if [ "$ISGCC" = "1" ]; then
    C_REGVAR_FLAGS=-ffixed-%ebx
    C_ABI_FLAGS=-m486
  else
# not tested yet...
    C_OPT_FLAGS=-xO2
    C_REGVAR_FLAGS=
    C_ABI_FLAGS=-xarch=386
  fi
  ASM_ABI_FLAGS=
elif [ "$SGC_CPU_TYPE" = "hppa" ]; then
  if [ "$ISGCC" = "1" ]; then
    C_REGVAR_FLAGS=
    C_ABI_FLAGS=
    ASM_ABI_FLAGS=
  else
    C_REGVAR_FLAGS=
    C_OPT_FLAGS=
    if [ "$OBJECT_64BIT" = "YES" ]; then
      C_ABI_FLAGS=+DA2.0W
      ASM_ABI_FLAGS=+DA2.0W
    else
      C_ABI_FLAGS=+DA1.0
      ASM_ABI_FLAGS=+DA1.0
    fi
  fi
fi

# setup compile options.
C_ETC_FLAGS="$C_OPT_FLAGS $C_DEBUG_FLAGS $C_WARN_FLAGS $C_ABI_FLAGS -D$SGC_THREAD_PKG"

if [ "$SGC_TLS_METHOD" = "SGC_TLS_REGISTER_VARIABLE" -o "$SGC_WITH_STHREADS" = "YES" ]; then
  C_ETC_FLAGS="$C_ETC_FLAGS $C_REGVAR_FLAGS"
fi

if [ "$SGC_WITH_STHREADS" = "YES" ]; then
# In StackThreads mode, $SGC_TLS_METHOD is ignored.
  C_ETC_FLAGS="$C_ETC_FLAGS -DSGC_TLS_STHREADS_MODE -I$ST_DIR/include"
else
  C_ETC_FLAGS="$C_ETC_FLAGS -D$SGC_TLS_METHOD"
fi

# how to make machine dependent files, such as mach_dep.o, psm.o.
if [ "$SGC_CPU_TYPE" = "sparc" ]; then
    MACHDEP_SRCTYPE=ASM
    ASMFILE_ID=sparc
elif [ "$SGC_CPU_TYPE" = "mips" ]; then
  if [ "$SGC_OS_TYPE" = "irix" -o "$SGC_OS_TYPE" = "IRIX" -o "$SGC_OS_TYPE" = "IRIX64" ]; then
    MACHDEP_SRCTYPE=ASM
    ASMFILE_ID=mips_sgi
  fi
elif [ "$SGC_CPU_TYPE" = "i386" ]; then
    MACHDEP_SRCTYPE=C
    ASMFILE_ID=i386
elif [ "$SGC_CPU_TYPE" = "hppa" ]; then
    MACHDEP_SRCTYPE=C
    ASMFILE_ID=hppa
fi

if [ "$MACHDEP_SRCTYPE" = "ASM" ]; then
    MAKE_MACH_DEP_O="$AS $ASM_ABI_FLAGS -o mach_dep.o ${ASMFILE_ID}_mach_dep.s"
elif [ "$MACHDEP_SRCTYPE" = "C" ]; then
    MAKE_MACH_DEP_O="$CC $C_ETC_FLAGS -c -o mach_dep.o mach_dep.c"
fi


########

for i in $*
do
#  eval echo "  " $0: $i = '$'$i 1>&2
  eval echo '$'$i
done

