/*
 * Copyright (c) 1998-1999 by Toshio Endo. All rights reserved.
 *
 * THIS MATERIAL IS PROVIDED AS IS, WITH ABSOLUTELY NO WARRANTY EXPRESSED
 * OR IMPLIED.  ANY USE IS AT YOUR OWN RISK.
 *
 * Permission is hereby granted to use or copy this program
 * for any purpose,  provided the above notices are retained on all copies.
 * Permission to modify the code and to distribute modified code is granted,
 * provided the above notices are retained, and a notice that the code was
 * modified is included with the above copyright notice.
 *
 */

# if defined(mips) || defined(__mips)
#    define MIPS
# endif

# if defined(sparc) || defined(__sparc)
#   define SPARC
# endif

#include "perfc_util.h"

#ifdef MIPS

#ifndef R10000
#define R10000
#endif

#include <stdio.h>
#include <sys/types.h>
/*#include <sys/stat.h>*/
#include <sys/procfs.h>
#include <sys/hwperftypes.h>
#include <sys/hwperfmacros.h>
#include <fcntl.h>
#include <unistd.h>

/* In some OS version, these macro are not defined. */
#if !defined HWPERF_CNT0MIN
#  define HWPERF_CNT0MIN 0
#endif
#if !defined HWPERF_CNT1MIN
#  define HWPERF_CNT1MIN 16
#endif
#if !defined HWPERF_CNT0MAX
#  define HWPERF_CNT0MAX 16
#endif
#if !defined HWPERF_CNT1MAX
#  define HWPERF_CNT1MAX 32
#endif

int PERFC_open(PERFC_handler_t *hp, int e0, int e1)
{
    int i;
    int pid = getpid();
    char fn[64];
    hwperf_profevctrarg_t evctr_args;

    sprintf(fn, "/proc/%05d", pid);
    hp->fd = open(fn, O_RDWR);
    if (hp->fd < 0) {
	fprintf(stderr, "[PREFC_open]open %s failed\n", fn);
	return -1;
    }

    if (e0 < 0 || e0 >= HWPERF_CNT0MAX - HWPERF_CNT0MIN) {
	fprintf(stderr, "[PREFC_open]invalid event no %d\n", e0);
	return -1;
    }

    if (e1 < 0 || e1 >= HWPERF_CNT1MAX - HWPERF_CNT1MIN) {
	fprintf(stderr, "[PREFC_open]invalid event no %d\n", e1);
	return -1;
    }

    hp->e0 = e0;
    hp->e1 = e1;

    for (i = HWPERF_CNT0MIN; i < HWPERF_CNT0MAX; i++) {
	if (i == e0 + HWPERF_CNT0MIN) {
	    evctr_args.hwp_evctrargs.hwp_evctrl[i].hwperf_creg.hwp_mode = HWPERF_CNTEN_U;
	    evctr_args.hwp_evctrargs.hwp_evctrl[i].hwperf_creg.hwp_ie = 1;
	    evctr_args.hwp_evctrargs.hwp_evctrl[i].hwperf_creg.hwp_ev = i;
	    evctr_args.hwp_ovflw_freq[i] = 0;
	} else {
	    evctr_args.hwp_evctrargs.hwp_evctrl[i].hwperf_spec = 0;
	    evctr_args.hwp_ovflw_freq[i] = 0;
	}
    }

    for (i = HWPERF_CNT1MIN; i < HWPERF_CNT1MAX; i++) {
	if (i == e1 + HWPERF_CNT1MIN) {
	    evctr_args.hwp_evctrargs.hwp_evctrl[i].hwperf_creg.hwp_mode = HWPERF_CNTEN_U;
	    evctr_args.hwp_evctrargs.hwp_evctrl[i].hwperf_creg.hwp_ie = 1;
	    evctr_args.hwp_evctrargs.hwp_evctrl[i].hwperf_creg.hwp_ev = i - HWPERF_CNT1MIN;
	    evctr_args.hwp_ovflw_freq[i] = 0;
	} else {
	    evctr_args.hwp_evctrargs.hwp_evctrl[i].hwperf_spec = 0;
	    evctr_args.hwp_ovflw_freq[i] = 0;
	}
    }

    evctr_args.hwp_ovflw_sig = 0;
    hp->gen = ioctl(hp->fd, PIOCENEVCTRS, (void *)&evctr_args);
    if (hp->gen < 0) {
	fprintf(stderr, "failed to acquire counters\n");
	return -1;
    }

    return 0;
}

int PERFC_read(PERFC_handler_t *hp,
	       PERFC_counter_t *pc0, PERFC_counter_t *pc1)
{
    int gen2;
    hwperf_cntr_t cnts;

    /* now read the counter values */
    if ((gen2 = ioctl(hp->fd, PIOCGETEVCTRS, (void *)&cnts)) < 0) {
	fprintf(stderr, "PIOCGETEVCTRS returns error\n");
	return -1;
    }
    
    /* generation number should be the same */
    if (hp->gen != gen2) {
	fprintf(stderr, "program lost event counters0\n");
	return -1;
    }

    memcpy(pc0, &cnts.hwp_evctr[hp->e0 + HWPERF_CNT0MIN],
	   sizeof(long long int));
    memcpy(pc1, &cnts.hwp_evctr[hp->e1 + HWPERF_CNT1MIN],
	   sizeof(long long int));

    return 0;
}

int PERFC_close(PERFC_handler_t *hp)
{
    if (hp->fd < 0) {
	return -1;
    }

    if ((ioctl(hp->fd, PIOCRELEVCTRS)) < 0) {
	fprintf(stderr, "prioctl PIOCRELEVCTRS returns error\n");
	return -1;
    }

    close(hp->fd);
    return 0;
}

#elif defined SPARC

/* Ultra SPARC version is fragile... */

/* Notice(1): /dev/pic device driver must have been installed. */
/* Notice(2): `processor_bind' before PERFC_open is strongly recommended. */

#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

#define PIC_FILENAME "/dev/pic"

int PERFC_open(PERFC_handler_t *hp, int e0, int e1)
{
    char fn[64];
    strcpy(fn, PIC_FILENAME);
    hp->fd = open(fn, O_RDONLY);
    if (hp->fd < 0) {
	fprintf(stderr, "[PREFC_open] open %s failed\n", fn);
	return -1;
    }
    if (ioctl(hp->fd, PIC_SET_EVENT0, e0) != 0) {
	fprintf(stderr, "[PREFC_open] ioctl failed\n");
	close(hp->fd);
	return -1;
    }
    if (ioctl(hp->fd, PIC_SET_EVENT1, e1) != 0) {
	fprintf(stderr, "[PREFC_open] ioctl failed\n");
	close(hp->fd);
	return -1;
    }
    hp->e0 = e0;
    hp->e1 = e1;
    return 0;
}

int PERFC_read(PERFC_handler_t *hp,
	       PERFC_counter_t *pc0, PERFC_counter_t *pc1)
{
    RDPIC01(*pc0, *pc1);
}

int PERFC_close(PERFC_handler_t *hp)
{
    if (hp->fd < 0) {
	return -1;
    }
    if (close(hp->fd) < 0) {
	fprintf(stderr, "PERFC_close: close(fd %d) fail! errno %d \n",
		hp->fd, errno);
    }
    return 0;
}

#else

/* not supported architecture */

int PERFC_open(PERFC_handler_t *hp, int e0, int e1)
{
    return -1;
}

int PERFC_read(PERFC_handler_t *hp,
	       PERFC_counter_t *pc0, PERFC_counter_t *pc1)
{
    return -1;
}

int PERFC_close(PERFC_handler_t *hp)
{
    return -1;
}

#endif
