/*
 * Copyright (c) 1998 by Toshio Endo. All rights reserved.
 *
 * THIS MATERIAL IS PROVIDED AS IS, WITH ABSOLUTELY NO WARRANTY EXPRESSED
 * OR IMPLIED.  ANY USE IS AT YOUR OWN RISK.
 *
 * Permission is hereby granted to use or copy this program
 * for any purpose,  provided the above notices are retained on all copies.
 * Permission to modify the code and to distribute modified code is granted,
 * provided the above notices are retained, and a notice that the code was
 * modified is included with the above copyright notice.
 *
 */

/* Ultra SPARC performance counter utility */

/* Definitions of IOCTL commands */

#ifndef _ENDO_PICIO_H
#define	_ENDO_PICIO_H

#define PICIOC ('p' << 8)

#define PIC_WRITE_PCR (PICIOC | 0x01)
#define PIC_READ_PCR (PICIOC | 0x02)
#define PIC_SET_EVENT0 (PICIOC | 0x11)
#define PIC_SET_EVENT1 (PICIOC | 0x12)

#define PIC_READ_PSTATE (PICIOC | 0x21)

/* Definitions of PIC event id */
/* Use as IOCTL argument */

/* For PIC0 */
#define PICEVENT0_CYCLE_CNT (0x0)
#define PICEVENT0_INSTR_CNT (0x1)
#define PICEVENT0_DISPATCH0_IC_MISS (0x2)
#define PICEVENT0_DISPATCH0_STOREBUF (0x3)
#define PICEVENT0_IC_REF (0x8)
#define PICEVENT0_DC_RD (0x9)
#define PICEVENT0_DC_WR (0xa)
#define PICEVENT0_LOAD_USE (0xb)
#define PICEVENT0_EC_REF (0xc)
#define PICEVENT0_EC_WRITE_HIT_RDO (0xd)
#define PICEVENT0_EC_SNOOP_INV (0xe)
#define PICEVENT0_EC_RD_HIT (0xf)

/* For PIC1 */
#define PICEVENT1_CYCLE_CNT (0x0)
#define PICEVENT1_INSTR_CNT (0x1)
#define PICEVENT1_DISPATCH0_MISPRED (0x2)
#define PICEVENT1_DISPATCH0_FP_USE (0x3)
#define PICEVENT1_IC_HIT (0x8)
#define PICEVENT1_DC_RD_HIT (0x9)
#define PICEVENT1_DC_WR_HIT (0xa)
#define PICEVENT1_LOAD_USE_RAW (0xb)
#define PICEVENT1_EC_HIT (0xc)
#define PICEVENT1_EC_WB (0xd)
#define PICEVENT1_EC_SNOOP_CB (0xe)
#define PICEVENT1_EC_IC_HIT (0xf)

/* Definitions of counter management macro */

#ifdef __GNUC__ /* for gcc */

#define RDPIC0(n) asm volatile("rd %%pic, %0" : "=r" (n) : )
#define RDPIC1(n) asm volatile ( \
        "rd %%pic, %0\n\tsrlx %0, 32, %0" : "=r" (n) : )
#define RDPIC01(n0, n1) asm volatile ( \
        "rd %%pic, %0\n\tsrlx %0, 32, %1" : "=r" (n0), "=r" (n1) : )
#define RDPIC(lln) asm volatile("rd %%pic, %0" : "=r" (lln) : )
#define WRPIC(lln) asm volatile("wr %0, %%pic" : : "r" (lln))

#define PIC01_0(lln, n0) n0 = lln
#define PIC01_1(lln, n1) asm volatile ( \
        "srlx %1, 32, %0" : "=r" (n1) : "r" (lln))

#define RDPCR(n) asm volatile("rd %%pcr, %0" : "=r" (n) : )
#define WRPCR(n) asm volatile("wr %0, %%pcr" : : "r" (n))

#define RDPSTATE(n) asm volatile("rdpr %%pstate, %0" : "=r" (n) : )

#define MEMBAR() \
    asm volatile("membar #StoreLoad | #Sync" : : )

#else /* for cc */

static unsigned long rdpic0(unsigned long *p)
{
    asm ("rd %pic, %l1");
    asm ("ld [%fp+68],%l0");
    asm ("st %l1,[%l0+0]");
}

static unsigned long rdpic1(unsigned long *p)
{
    asm ("rd %pic, %l1");
    asm ("srlx %l1, 32, %l1");
    asm ("ld [%fp+68],%l0");
    asm ("st %l1,[%l0+0]");
}

static unsigned long rdpcr(unsigned long *p)
{
    asm ("rd %pcr, %l1");
    asm ("ld [%fp+68],%l0");
    asm ("st %l1,[%l0+0]");
}

static void wrpic(unsigned long n)
{
    asm ("ld [%fp+68],%l0");
    asm ("wr %l0, %pic");
}

static void wrpcr(unsigned long n)
{
    asm ("ld [%fp+68],%l0");
    asm ("wr %l0, %pcr");
}

#define RDPIC0(n) {rdpic0(&n);}
#define RDPIC1(n) {rdpic1(&n);}
#define RDPIC01(n0, n1) {rdpic0(&n0);rdpic1(&n1);}
#define RDPCR(n) {rdpcr(&n);}
#define WRPIC(n) {wrpic(n);}
#define WRPCR(n) {wrpcr(n);}

#endif

#endif /* _ENDO_PICIO_H */
