/*
 * Copyright (c) 1998-1999 by Toshio Endo, Kenjiro Taura. All rights reserved.
 * Copyright (c) 1998-1999 by Mitsubishi Research Institute.  All rights reserved.
 * Copyright (c) 1998-1999 by Information-technology Promotion Agency.  All rights reserved.
 *
 * THIS MATERIAL IS PROVIDED AS IS, WITH ABSOLUTELY NO WARRANTY EXPRESSED
 * OR IMPLIED.  ANY USE IS AT YOUR OWN RISK.
 *
 * Permission is hereby granted to use or copy this program
 * for any purpose,  provided the above notices are retained on all copies.
 * Permission to modify the code and to distribute modified code is granted,
 * provided the above notices are retained, and a notice that the code was
 * modified is included with the above copyright notice.
 *
 */
/* Endo, August 17, 1999 */

/* include this file immediately after `gcconfig.h'. */

# ifndef SGCCONFIG_H
# define SGCCONFIG_H

/* currently SGC doesn't support this... */
# ifdef DYNAMIC_LOADING
#   undef DYNAMIC_LOADING
# endif

/* currently SGC doesn't support incremental GC */
# ifdef PCR_VDB
#   undef PCR_VDB
# endif
# ifdef PROC_VDB
#   undef PROC_VDB
# endif
# ifdef MPROTECT_VDB
#   undef MPROTECT_VDB
# endif
# ifndef DEFAULT_VDB
#   define DEFAULT_VDB
# endif

/* these definitions are mainly for os_dep.c, which is almost */
/* unchanged from the original version. */
# if defined(SUNOS5) && !defined(SOLARIS_THREADS)
#   define SOLARIS_THREADS
# endif
# if defined(IRIX5) && !defined(IRIX_THREADS)
#  define IRIX_THREADS
# endif
# if defined(LINUX) && !defined(LINUX_THREADS)
#   define LINUX_THREADS
# endif
/* in SGC, THREADS is always defined */
# ifndef THREADS
#   define THREADS
# endif

/********************************** User configration part BEGIN */
/* Configuration for experiments */

# define DO_ATOMIC_MARK   /* Must be defined in parallel marking */

/* DO_LOAD_DIST determines task distribution strategy. */
/* not-defined ... do not distribute task */
/* 3 ... Use stealable mark queues */
/* 4 ... Steal objects from vectim's mark stack directly(Cilk-like method) */
/* 5 ... Cilk-like method with watermark */
# define DO_LOAD_DIST 5
# define DO_PRECLAIM 1   /* Parallel reclaimation */

/* #define DO_REQ_REMOTE */

/* DO_OPT_SLO determines how large objects are managed. */
/* not-defined ... same as small objects */
/* 1 ... split large objects when pushed. DO_LOAD_DIST must be 3/4/5. */
/* 2 ... halve large objects when stolen. DO_LOAD_DIST must be 4/5. */
# define DO_OPT_SLO 2
# define DO_OPT_MOS 1 /* Mark with Opt Sync (Test-Compare&swap) */
# define DO_OPT_NSB 1 /* Non-serializing barrier */

# define DO_OPT_SLQ 0 /* Skip locked queue */

/* If SGC_BALANCED_HDR is defined, hblkhdrs are allocated by */
/* GC processes. All GC processes has approximately as many hdrs */
/* as others. */
/* SGC_BALANCED_HDR may improve marking speed on NUMA machines. */
# define SGC_BALANCED_HDR

/* MARKLOG: log about GC-speed = (GC-workload/GC-time) */
/* DO_MARKLOG=1 .. simple, DO_MARKLOG=2 .. detail */
#define DO_MARKLOG 1

/* PICLOG: log with performance counter. (currently sparc/mips only) */
/*#define DO_PICLOG*/

/* LDLOG: log about time-breakdown in each collection. */
/* #define DO_LDLOG */

/* ALLOCLOG: log about the frequency of GC_malloc by user. */
/* #define DO_ALLOCLOG */

/* GCRLOG: log about rate of GC-time to application-time. require MARKLOG */
#define DO_GCRLOG

/* MSLOG: log about mark stack usage. */
/* #define DO_MSLOG */

/* CACHELOG: log about predicted(approximate) cache misses */
/*#define DO_CACHELOG*/

#define USE_PERFC

/* for future optimisation. don't mind! */
/* 99.2.18 endo */
/*# define LOCAL_HDR_COPY*/

/* for temporary statistics. don't mind! */
/* 99.2.18 endo */
/* #define COUNT_CROWD */

/*#define SHM_FIXED*/

/*# define PSM_STANDALONE*/

/* #define DO_PROCESSOR_BIND */

/********************************** User configration part END */


#if !defined SPARC && !defined MIPS && defined DO_PICLOG
#  undef DO_PICLOG
#endif

#if !defined USE_PERFC && defined DO_PICLOG
#  undef DO_PICLOG
#endif

#if defined SGC_TLS_GLOBAL_VARIABLE
/* In process(fork) version, we must mmap to share heap. */
#  define SGC_HEAP_MMAP_SHARED
#else
/* In thread version, no special action is required to share heap. */
#  define SGC_HEAP_IMPLICITLY_SHARED
#endif

#if defined SGC_HEAP_MMAP_SHARED && !defined SHM_FIXED
#  define SHM_FIXED
#endif

/* available thread package */

#if defined IRIX5 || defined SUNOS5 || defined LINUX || defined HP_PA
#  define SIGNAL_SUPPORTED
#endif

#if defined SIGNAL_SUPPORTED && defined SDSM_TEST
#  undef SIGNAL_SUPPORTED
#endif

#if defined IRIX5 || defined SUNOS5 || defined LINUX || defined HP_PA
#  define TIMEVAL_SUPPORTED
#endif

#if defined IRIX5 && defined SGC_PTHREAD
/* IRIX 6.4 supports pthread_setconcurrency, */
/* instead of PTHREAD_SCOPE_SYSTEM. */
#  define PTHREAD_SETCONC_SUPPORTED
#  define PTHREAD_SETSCOPE_NOTSUPPORTED
#endif

#if defined PTHREAD_SETCONC_SUPPORTED || defined SGC_SOLARIS_THREAD || \
    defined SGC_SOLARIS_THREAD_PTHREAD
#  define SETCONC_SUPPORTED
#endif

#if defined SUNOS5 || defined IRIX5
#  define PROCESSOR_BIND_SUPPORTED
#endif

# endif /* SGCCONFIG_H */
