/*
 * Copyright (c) 1998 by Toshio Endo, Kenjiro Taura. All rights reserved.
 * Copyright (c) 1998 by Mitsubishi Research Institute.  All rights reserved.
 * Copyright (c) 1998 by Information-technology Promotion Agency.  All rights reserved.
 *
 * THIS MATERIAL IS PROVIDED AS IS, WITH ABSOLUTELY NO WARRANTY EXPRESSED
 * OR IMPLIED.  ANY USE IS AT YOUR OWN RISK.
 *
 * Permission is hereby granted to use or copy this program
 * for any purpose,  provided the above notices are retained on all copies.
 * Permission to modify the code and to distribute modified code is granted,
 * provided the above notices are retained, and a notice that the code was
 * modified is included with the above copyright notice.
 *
 */
/* Endo, July 21, 1999 */

#include <stdio.h>
#include "gc_priv.h"
#ifdef SIGNAL_SUPPORTED
#include <signal.h>
#endif

/* Configuration (valid only in this file) */

/* #define USE_BLOCKING_BARRIER */
/* #define USE_DEBUG_HANDLER */

#ifndef SILENT
/*#  define SGC_DEBUGPRINT */
#endif

#ifdef SGC_BALANCED_HDR
#define NPAGE_FOR_HDR_ALLOC 8
#endif

/* Define some constants */
#define GCP_COM_COLLECTION 0
#define GCP_COM_EXIT 1
#define GCP_COM_FILLHDR 2 /* Used if SGC_BALANCED_HDR */

#define SPMDP_COM_COLLECTION 0
#define SPMDP_COM_FILLHDR 2 /* Used if SGC_BALANCED_HDR */

#ifdef SIGNAL_SUPPORTED
#  define SGC_SIG_STOPWORLD SIGPWR
#endif

/* Define some functions */
#ifdef SIGNAL_SUPPORTED
void GC_sh_stop_handler GC_PROTO((int));
void GC_sh_error_handler GC_PROTO((int));
void GC_sh_int_handler GC_PROTO((int));
int GC_setup_debug_handler GC_PROTO((void));
int GC_setup_stop_handler GC_PROTO((void));
#endif /* SIGNAL_SUPPORTED */

int GC_marksweep_local GC_PROTO((SGC_MYENV_ARG_DECL0));
void SGC_abort GC_PROTO((SGC_MYENV_ARG_DECL0));
int GC_clear_app_counters GC_PROTO((SGC_MYENV_ARG_DECL0));


/* moved from psm.h. should I change the name? */
void PSM_mutex_lock_polling_inner(SGC_MYENV_ARG_DECL1 
				  PSM_mutex_t *mutex_p,
				  void (*polling_func)(SGC_MYENV_ARG_DECL0))
{
    while (PSM_mutex_trylock(mutex_p) == 0) {
	(*polling_func)(SGC_MYENV_ARG0);
    }
}

/* */
/* Send a signal to a App/SPMD process. */
int GC_notify_async_to_appproc(GC_thread envp)
{
    GC_thread_id pid;
    if (!envp) return -1;

    if (GC_use_polling == 0) {
#ifdef SIGNAL_SUPPORTED
	pid = GC_l_pid(envp);
#ifdef SGC_DEBUGPRINT
	GC_err_printf1("Invoker will send a signal to pid = %d\n", pid);
#endif
	if (GC_sh_sendsig(pid, SGC_SIG_STOPWORLD) < 0) {
	    /* no such a thread! the thread may be finished... */
	    return -1;
	}
#endif
    }
    else {
	GC_la_polling_flag(envp) = 1;
    }

    return 0;
}

/* Send a signal to all App/SPMD processes except caller. */
/* Returns the number of found threads. */
int GC_notify_async_to_all_appprocs(SGC_MYENV_ARG_DECL1 GC_bool exceptme)
{
    DECLARE_MYENV
    GC_thread envp;
    int sendcount = 0;

    for (envp = GC_g_appproc_envps.head; envp; envp = envp->next) {
	if (GC_l_initialized(envp) == 0) continue;
	if (GC_l_terminated(envp) != 0) continue;
	if (exceptme) {
	    /* Skip self */
	    if (envp == GC_my_envp) continue;
	}

	if (GC_notify_async_to_appproc(envp) < 0) {
	    /* failed to send a signal to the thread corresponding to envp. */
	    if (envp) {
		/* We regard that thread `terminated'. */
		GC_l_terminated(envp) = 1;
#if 1 || defined SGC_DEBUGPRINT
		GC_err_printf1("Invoker found thread %d terminated.\n",
			       GC_l_pid(envp));
#endif
	    }
	}
	else {
	    sendcount++;
	}
    }
#ifdef SGC_DEBUGPRINT
    GC_err_printf1("Invoker sent a stop signal to %d procs(except me).\n", sendcount);
#endif

    return sendcount;
}

int GC_awake_proc(GC_thread envp)
{
    GC_bssem_signal(&GC_l_sem(envp));
    return 0;
}

/* Send a semaphore signal to all processes in the list. */
int GC_awake_all_procs(struct envp_list *list)
{
    GC_thread envp;
    int sendcount = 0;
#ifdef SGC_DEBUGPRINT
    struct timeval st, et;
    gettimeofday(&st, NULL);
#endif

    for (envp = list->head; envp; envp = envp->next) {
	if (GC_l_initialized(envp) == 0) continue;
	if (GC_l_terminated(envp) != 0) continue;
	GC_awake_proc(envp);
	sendcount++;
    }
#ifdef SGC_DEBUGPRINT
    gettimeofday(&et, NULL);
    {
	char buf[64];
	sprintf(buf, "awaked %d procs.\n", sendcount);
	GC_err_printf0(buf);
	sprintf(buf, "It took %d msec.\n",
		(et.tv_sec-st.tv_sec)*1000 + (et.tv_usec-st.tv_usec)/1000);
	GC_err_printf0(buf);
    }
#endif

    return 0;
}

int GC_awake_all_appprocs()
{
    return GC_awake_all_procs(&GC_g_appproc_envps);
}

int GC_awake_all_gcprocs()
{
    return GC_awake_all_procs(&GC_g_gcproc_envps);
}

/* */
void GC_proc_envps_init(struct envp_list *list)
{
    list->head = NULL;
    PSM_mutex_init(&list->ml);
}

void GC_appproc_envps_init()
{
    GC_proc_envps_init(&GC_g_appproc_envps);
}

void GC_gcproc_envps_init()
{
    GC_proc_envps_init(&GC_g_gcproc_envps);
}

/* */
int GC_count_proc_envps(struct envp_list *list)
{
    int n = 0;
    GC_thread envp;
    PSM_mutex_lock(&list->ml);
    for (envp = list->head; envp; envp = envp->next) {
	if (GC_l_initialized(envp) == 0) continue;
	if (GC_l_terminated(envp) != 0) continue;
	n++;
    }
    PSM_mutex_unlock(&list->ml);
    return n;
}

int GC_count_appproc_envps()
{
    return GC_count_proc_envps(&GC_g_appproc_envps);
}

int GC_count_gcproc_envps()
{
    return GC_count_proc_envps(&GC_g_gcproc_envps);
}

/* */
int GC_is_appproc_envps_empty()
{
    return (GC_g_appproc_envps.head == NULL);
}

int GC_is_gcproc_envps_empty()
{
    return (GC_g_gcproc_envps.head == NULL);
}

/* */
/* only one process gets 1, others 0. */
int GC_is_appproc_envp_head(GC_thread envp)
{
    return (GC_g_appproc_envps.head == envp);
}

/* only one process gets 1, others 0. */
int GC_is_gcproc_envp_head(GC_thread envp)
{
    return (GC_g_gcproc_envps.head == envp);
}

/* */
int GC_add_proc_envp(struct envp_list *list, GC_thread envp)
{
    PSM_mutex_lock(&list->ml);
    envp->next = list->head;
    list->head = envp;
    PSM_mutex_unlock(&list->ml);
#if 0 || defined SGC_DEBUGPRINT
    GC_err_printf2("added envp, pid = %d. I'm %d\n", 
		   GC_l_pid(envp), GC_sh_getpid());
#endif
    return 0;
}

int GC_add_appproc_envp(GC_thread envp)
{
#if 0 || defined SGC_DEBUGPRINT
    if (GC_l_pid(envp) != GC_sh_getpid()) {
	GC_err_printf2("!BEFORE_APPP_ADD! envp broken %08x, pid %d\n",
		       envp, GC_l_pid(envp));
    }
#endif
    return GC_add_proc_envp(&GC_g_appproc_envps, envp);
}

int GC_add_gcproc_envp(GC_thread envp)
{
    GC_add_proc_envp(&GC_g_gcproc_envps, envp);
    return 0;
}

/* */
GC_thread GC_lookup_proc_envp(struct envp_list *list, GC_thread_id pid)
{
    GC_thread envp;

    PSM_mutex_lock(&list->ml);
    for (envp = list->head; envp; envp = envp->next) {
	if (GC_l_pid(envp) == pid) {
	    PSM_mutex_unlock(&list->ml);
	    return envp;
	}
    }
    PSM_mutex_unlock(&list->ml);
#if 1 || defined SGC_DEBUGPRINT
    GC_err_printf1("we have no envp, whose pid = %d.\nregisterd pid are ", 
		   pid);
    PSM_mutex_lock(&list->ml);
    for (envp = list->head; envp; envp = envp->next) {
	GC_err_printf1("%d, ", GC_l_pid(envp));
    }
    PSM_mutex_unlock(&list->ml);
    GC_err_printf0("\n");
#endif
    return NULL;
}

GC_thread GC_lookup_appproc_envp(GC_thread_id pid)
{
    return GC_lookup_proc_envp(&GC_g_appproc_envps, pid);
}

GC_thread GC_lookup_gcproc_envp(GC_thread_id pid)
{
    return GC_lookup_proc_envp(&GC_g_gcproc_envps, pid);
}

/* */
int GC_remove_proc_envp(struct envp_list *list, GC_thread removep)
{
    GC_thread envp;
    GC_thread prevenvp;

    /* TODO: reuse of env structure */

    PSM_mutex_lock(&list->ml);
    if (!list->head) {
	PSM_mutex_unlock(&list->ml);
	return -1;
    }

    if (list->head == removep) {
	list->head = list->head->next;
	PSM_mutex_unlock(&list->ml);
	return 0;
    }

    prevenvp = list->head;
    for (envp = prevenvp->next; envp; envp = envp->next) {
	if (envp == removep) {
	    prevenvp->next = envp->next;
	    PSM_mutex_unlock(&list->ml);
	    return 0;
	}
	prevenvp = envp;
    }

    PSM_mutex_unlock(&list->ml);
    return -1;
}

int GC_remove_appproc_envp(GC_thread removep)
{
    return GC_remove_proc_envp(&GC_g_appproc_envps, removep);
}

int GC_remove_gcproc_envp(GC_thread removep)
{
    return GC_remove_proc_envp(&GC_g_gcproc_envps, removep);
}

/* */
GC_thread GC_alloc_env()
{
    GC_thread envp;
    envp = (GC_thread)GET_MEM(sizeof(struct GC_Thread_Rep));
    if (envp == NULL) {
	GC_err_printf0("[GC_alloc_env] Cannot allocate GC_thread.\n");
	GC_sh_exit_process();
    }
    GC_touch_memory_region((char *)envp, sizeof(struct GC_Thread_Rep));
#ifdef SGC_DEBUGPRINT
    GC_err_printf3("####### pid %d allocd GC_thread %08x -- %08x\n",
		   GC_sh_getpid(), envp, (envp+1));
    GC_err_printf2("####### pid %d gcproc_info starts at %08x\n",
		   GC_sh_getpid(), &envp->gc);
    GC_err_printf2("####### pid %d appproc_info starts at %08x\n",
		   GC_sh_getpid(), &envp->app);
#endif
    return envp;
}

int GC_init_proc_common(GC_thread envp)
{
    GC_l_initialized(envp) = 0;
    GC_l_terminated(envp) = 0;
    GC_l_pid(envp) = GC_sh_getpid();
#if 0 || defined SGC_DEBUGPRINT
    GC_err_printf2("set pid: envp %08x, pid %d\n",
		   envp, GC_l_pid(envp));
#endif
    GC_l_debug_state(envp) = 0;
    GC_scratch_alloc_init(&GC_l_scratch(envp));

    GC_bssem_init(&GC_l_sem(envp), &GC_blocksync);
#if 1 || defined SGC_DEBUGPRINT
    if (GC_l_pid(envp) != GC_sh_getpid()) {
	GC_err_printf2("!B! envp broken %08x, pid %d\n", envp, GC_l_pid(envp));
    }
#endif
    GC_bssem_clear(&GC_l_sem(envp));
#if 1 || defined SGC_DEBUGPRINT
    if (GC_l_pid(envp) != GC_sh_getpid()) {
	GC_err_printf2("!C! envp broken %08x, pid %d\n", envp, GC_l_pid(envp));
    }
#endif

    /* Allocate mark stack. */
    GC_mark_stack_init(envp);
    GC_MARKLOG_INIT_LOCAL(envp);
    GC_PICLOG_INIT_LOCAL(envp);
    GC_CACHELOG_INIT_LOCAL(envp);

    return 0;
}

int GC_init_gcenv(GC_thread envp)
{
    int ok;
    ptr_t *fl;
    struct hblk **rlist;
    GC_thread p;

#if 0 || defined SGC_DEBUGPRINT
    GC_err_printf2("GC_INIT_GCENV: (pid %d) envp %08x\n",
		   GC_l_pid(envp), envp);
#endif
#if 1 || defined SGC_DEBUGPRINT
    if (GC_l_pid(envp) != GC_sh_getpid()) {
	GC_err_printf2("!INIT_GCENV! envp broken %08x, pid %d\n",
		       envp, GC_l_pid(envp));
    }
#endif

#ifdef DO_REQ_REMOTE
    GC_reqbuf_init(envp);
#endif
#ifdef SGC_BALANCED_HDR
    GC_balanced_hdr_init(envp);
#endif

#ifdef SIGNAL_SUPPORTED
    if (GC_sh_is_signalhandler_separate()) {
	/* each process must setup signal handler */
	GC_setup_debug_handler();
    }
#endif

# ifdef DO_LOAD_DIST
    GC_ld_init_local(envp);
# endif
    GC_LDLOG_INIT_LOCAL(envp);
    GC_MSLOG_INIT_LOCAL(envp);

#ifdef FIND_LEAK
    {
	GC_alloc_flrl(envp, NORMAL);
	fl = (ptr_t *)
	  GC_scratch_generic_alloc((MAXOBJSZ+1)*sizeof(ptr_t),
				   &GC_l_scratch(envp));
	BZERO(fl, (MAXOBJSZ+1)*sizeof(ptr_t));
	GC_lg_pr(envp).freelist_top[NORMAL] = fl;
	
	GC_alloc_flrl(envp, PTRFREE);
	fl = (ptr_t *)
	  GC_scratch_generic_alloc((MAXOBJSZ+1)*sizeof(ptr_t),
				   &GC_l_scratch(envp));
	BZERO(fl, (MAXOBJSZ+1)*sizeof(ptr_t));
	GC_lg_pr(envp).freelist_top[PTRFREE] = fl;

	GC_alloc_flrl(envp, UNCOLLECTABLE);
	fl = (ptr_t *)
	  GC_scratch_generic_alloc((MAXOBJSZ+1)*sizeof(ptr_t),
				   &GC_l_scratch(envp));
	BZERO(fl, (MAXOBJSZ+1)*sizeof(ptr_t));
	GC_lg_pr(envp).freelist_top[UNCOLLECTABLE] = fl;
# ifdef ATOMIC_UNCOLLECTABLE
	GC_alloc_flrl(envp, AUNCOLLECTABLE);
	fl = (ptr_t *)
	  GC_scratch_generic_alloc((MAXOBJSZ+1)*sizeof(ptr_t),
				   &GC_l_scratch(envp));
	BZERO(fl, (MAXOBJSZ+1)*sizeof(ptr_t));
	GC_lg_pr(envp).freelist_top[AUNCOLLECTABLE] = fl;
# endif
# ifdef STUBBORN_ALLOC
	GC_alloc_flrl(envp, STUBBORN);
	fl = (ptr_t *)
	  GC_scratch_generic_alloc((MAXOBJSZ+1)*sizeof(ptr_t),
				   &GC_l_scratch(envp));
	BZERO(fl, (MAXOBJSZ+1)*sizeof(ptr_t));
	GC_lg_pr(envp).freelist_top[STUBBORN] = fl;
# endif
    }

#endif /* FIND_LEAK */

    return 0;
}

/* Work of GC process */
void GC_gcproc_mainloop(SGC_MYENV_ARG0)
SGC_MYENV_ARG_DECL2
{
    DECLARE_MYENV
    int com;
    while (1) {
#if 0 && defined SGC_DEBUGPRINT
	{
	    char buf[64];
	    sprintf(buf, "GC proc start waiting... envp %08x.\n",
		    GC_my_envp);
	    GC_err_printf0(buf);
	}
#endif
	/* Wait signal. */
	GC_bssem_wait(&GC_my_sem);
	/* For next signal. */
	GC_bssem_clear(&GC_my_sem);
	if (GC_gcp_command == GCP_COM_EXIT) {
	    /* Termination of system. */
	    break;
	}
#ifdef SGC_BALANCED_HDR
	else if (GC_gcp_command == GCP_COM_FILLHDR) {
	    GC_fill_hdr_free_list(SGC_MYENV_ARG0);
	    PSM_barrier_sync(&GC_gcp_barrier);
	    if (GC_is_gcproc_envp_head(GC_my_envp)) {
		GC_awake_proc(GC_requestor_envp);
	    }
	}
#endif
	else {
	    /* Do a collection */
#if 0 && defined SGC_DEBUGPRINT
	    {
		char buf[64];
		sprintf(buf,
			"GC proc %08x received semsignal and start GC work\n",
			GC_my_envp);
		GC_err_printf0(buf);
	    }
#endif
	    if (GC_my_envp == GC_g_gcproc_envps.head) {
		/* determine gc leader thread */
		GC_g_gcleader_envp = GC_my_envp;
	    }
	    PSM_barrier_sync(&GC_gcp_barrier);
	    GC_marksweep_local(SGC_MYENV_ARG0); /* gc main routine */
	}
    }
}

/* called by each (App) process. */
/* don't use GC_my_envp in this function. */
int GC_alloc_flrl(GC_thread envp, int ok)
{
    ptr_t *fl;
    struct hblk **rlist;

    fl = (ptr_t *)
	GC_scratch_generic_alloc((MAXOBJSZ+1)*sizeof(ptr_t),
				 &GC_l_scratch(envp));
    BZERO(fl, (MAXOBJSZ+1)*sizeof(ptr_t));
    GC_la_freelist(envp)[ok] = fl;

    rlist = (struct hblk **)
	GC_scratch_generic_alloc((MAXOBJSZ+1)*sizeof(struct hblk *),
				 &GC_l_scratch(envp));
    BZERO(rlist, (MAXOBJSZ+1)*sizeof(struct hblk *));
    GC_la_reclaim_list(envp)[ok] = rlist;

    return 0;
}

# if defined(__STDC__) || defined(__cplusplus)
int GC_alloc_grl(int ok)
#else
int GC_alloc_grl(ok)
     int ok;
#endif
{
    ptr_t *fl;
    struct hblk **rlist;

    rlist = (struct hblk **)
	GC_scratch_g_alloc((MAXOBJSZ+1)*sizeof(struct hblk *));
    BZERO(rlist, (MAXOBJSZ+1)*sizeof(struct hblk *));
    GC_g_reclaim_list[ok] = rlist;
    PSM_mutex_init(&GC_g_reclaim_list_ml);

#ifdef FIND_LEAK
    fl = (ptr_t *)
	GC_scratch_g_alloc((MAXOBJSZ+1)*sizeof(ptr_t));
    BZERO(fl, (MAXOBJSZ+1)*sizeof(ptr_t));
    GC_g_freelist[ok] = fl;
    PSM_mutex_init(&GC_g_freelist_ml);
#endif /* FIND_LEAK */

    return 0;
}

int GC_init_appenv(GC_thread envp)
{
#if 0 || defined SGC_DEBUGPRINT
    GC_err_printf2("GC_INIT_APPENV: (pid %d) envp %08x\n",
		   GC_l_pid(envp), envp);
#endif
    if (GC_l_pid(envp) != GC_sh_getpid()) {
	GC_err_printf2("!INIT_APPENV! envp broken %08x, pid %d\n",
		       envp, GC_l_pid(envp));
    }

    /* Set up free lists. */
    GC_alloc_flrl(envp, NORMAL);
    GC_alloc_flrl(envp, PTRFREE);
    GC_alloc_flrl(envp, UNCOLLECTABLE);
# ifdef ATOMIC_UNCOLLECTABLE
    GC_alloc_flrl(envp, AUNCOLLECTABLE);
# endif
# ifdef STUBBORN_ALLOC
    GC_alloc_flrl(envp, STUBBORN);
# endif
    PSM_mutex_init(&GC_la_alloc_ml(envp));

#ifdef SIGNAL_SUPPORTED
    if (GC_sh_is_signalhandler_separate()) {
	/* each process must setup signal handler */
	GC_setup_debug_handler();
	if (GC_use_polling == 0) {
	    GC_setup_stop_handler();
	}
    }
#endif
    GC_la_polling_flag(envp) = 0;
    GC_la_n_root_sets(envp) = 0;
    GC_la_barrier_stage(envp) = 0;

    GC_ALLOCLOG_INIT_LOCAL(envp);

    return 0;
}

/* don't use GC_my_envp in this function. */
GC_thread GC_init_appproc(ptr_t my_stackbottom, int detached)
{
    GC_thread envp;
    /* Init this process as App proc. */
    envp = GC_alloc_env();

    GC_l_app_or_gc(envp) = 0;
    GC_la_stackbottom(envp) = my_stackbottom;
    GC_la_flags(envp) = (detached)? SGC_DETACHED: 0;

    GC_init_proc_common(envp);
    GC_init_appenv(envp);

    GC_add_appproc_envp(envp);

    return envp;
}

/* don't use GC_my_envp in this function. */
GC_thread GC_init_gcproc()
{
    GC_thread envp;

    /* Init this process as GC proc. */
#if defined PROCESSOR_BIND_SUPPORTED && defined DO_PROCESSOR_BIND
    {
	int n;
	int pr;
	PSM_mutex_lock(&GC_g_bind_support.ml);
	n = GC_g_bind_support.count;
	GC_g_bind_support.count++;
	PSM_mutex_unlock(&GC_g_bind_support.ml);
	/* try to bind this process to `n'th processor */
	pr = GC_sh_processor_bind(n);
#ifdef SGC_DEBUGPRINT
	if (pr >= 0) {
	    GC_err_printf2("OK to bind pid %d to processor %d.\n",
			   GC_sh_getpid(), pr);
	}
v	else {
	    GC_err_printf1("NG to bind pid %d to processor.\n",
			   GC_sh_getpid());
	}
#endif
    }
#endif

    envp = GC_alloc_env();
    GC_l_app_or_gc(envp) = 1;

    GC_init_proc_common(envp);
    GC_init_gcenv(envp);

    GC_add_gcproc_envp(envp);

    return envp;
}

/* don't use GC_my_envp in this function. */
GC_thread GC_init_spmdproc(ptr_t my_stackbottom, int detached)
{
    GC_thread envp;

    /* Init this process as SPMD proc. */
#if defined PROCESSOR_BIND_SUPPORTED && defined DO_PROCESSOR_BIND
    {
	int n;
	int pr;
	PSM_mutex_lock(&GC_g_bind_support.ml);
	n = GC_g_bind_support.count;
	GC_g_bind_support.count++;
	PSM_mutex_unlock(&GC_g_bind_support.ml);
	/* try to bind this process to `n'th processor */
	pr = GC_sh_processor_bind(n);
#ifdef SGC_DEBUGPRINT
	if (pr >= 0) {
	    GC_err_printf2("OK to bind pid %d to processor %d.\n",
			   GC_sh_getpid(), pr);
	}
	else {
	    GC_err_printf1("NG to bind pid %d to processor.\n",
			   GC_sh_getpid());
	}
#endif
    }
#endif

    envp = GC_alloc_env();

    GC_l_app_or_gc(envp) = -1;
    GC_la_stackbottom(envp) = my_stackbottom;
    GC_la_flags(envp) = (detached)? SGC_DETACHED: 0;

    GC_init_proc_common(envp);
    GC_init_gcenv(envp);
    GC_init_appenv(envp);

    GC_add_appproc_envp(envp);
    GC_add_gcproc_envp(envp);

    return envp;
}

void *GC_gcproc_wrapper(void *arg)
{
    DECLARE_MYENV2
    void *oldreg = SGC_save_regvar(); /* Save register variable */
    GC_thread envp;

    PSM_mutex_lock(&GC_g_alloc_ml);
    envp = GC_init_gcproc();
    PSM_mutex_unlock(&GC_g_alloc_ml);

#if SGC_EXPLICIT_INIT_LOCAL == 1
    SGC_gcproc_init_hook();
#endif
    /* now access to GC_my_envp is possible. */
    SGC_set_myenvp(envp);
    GC_my_envp = envp;
    GC_my_initialized = 1;

    PSM_barrier_sync(&GC_gcp_barrier);

    if (GC_is_gcproc_envp_head(GC_my_envp)) {
	/* Send a signal to parent process(App process). */
	GC_bssem_signal(&GC_g_sem);
    }

    /* set process id */
    GC_myg_pno = (long)arg;

    /* GC process main loop */
    GC_gcproc_mainloop(SGC_MYENV_ARG0);
    SGC_restore_regvar(oldreg); /* Restore register variable */

    return NULL;
}

int GC_create_gcprocs(int n_gcprocs)
{
    int i;

    if (GC_np_fixed != 0) {
	GC_err_printf0("[GC_create_gcprocs] logic error.\n");
	return -1;
    }

    GC_bssem_clear(&GC_g_sem);
    PSM_barrier_init(&GC_gcp_barrier, n_gcprocs);
    for (i = 0; i < n_gcprocs; i++) {
	GC_sh_create_process(GC_gcproc_wrapper, (void *)0);
    }
    /* Wait initialization of all GC processes. */
    GC_bssem_wait(&GC_g_sem);
    return 0;
}

int GC_kill_gcprocs()
{
    GC_thread envp;
    int sendcount = 0;
    struct envp_list *list;

    /* When GC process receives a signal next time, it exits. */
    GC_gcp_command = GCP_COM_EXIT;
    GC_awake_all_gcprocs();

    for (envp = GC_g_gcproc_envps.head; envp; envp = envp->next) {
	if (GC_l_initialized(envp) == 0) {
	    GC_err_printf0("[GC_kill_gcprocs] found non-initialized proc. \n");
	    continue;
	}
	if (GC_l_terminated(envp) != 0) continue;
	GC_sh_wait(GC_l_pid(envp));
    }
    return 0;
}

/* called by each app/SPMD process. */
/* don't use GC_my_envp in this function. */
GC_thread GC_process_creation_hook(ptr_t my_stackbottom, int flags)
{
    GC_thread envp;
    if (GC_np_fixed == 0) {
	envp = GC_init_appproc(my_stackbottom, flags);
    }
    else {
	envp = GC_init_spmdproc(my_stackbottom, flags);
    }

#if SGC_EXPLICIT_INIT_LOCAL == 0    
    /* setup GC_my_envp. */
    SGC_init_local();
#endif

    return envp;
}

/*******************************************************************/
#ifdef SIGNAL_SUPPORTED
int GC_setup_stop_handler()
{
    struct sigaction act;

    if (GC_use_polling) {
	return 0;
    }

#if defined SGC_DEBUGPRINT
    GC_err_printf1("P/TID %d setup stop handler.\n",GC_sh_getpid());
#endif

    sigaction(SGC_SIG_STOPWORLD, NULL, &act);
    act.sa_flags |= SA_RESTART;
    act.sa_handler = GC_sh_stop_handler;
    sigaction(SGC_SIG_STOPWORLD, &act, NULL);

    return 0;
}

void GC_reset_stop_handler()
{
    struct sigaction act;

    sigaction(SGC_SIG_STOPWORLD, NULL, &act);
    act.sa_flags &= ~SA_RESTART;
    act.sa_handler = SIG_DFL;
    sigaction(SGC_SIG_STOPWORLD, &act, NULL);
}

/* For debug */
int GC_setup_debug_handler()
{
#ifdef USE_DEBUG_HANDLER
    struct sigaction act;

    sigaction(SIGBUS, NULL, &act);
    act.sa_flags |= SA_RESTART;
    act.sa_handler = GC_sh_error_handler;
    sigaction(SIGBUS, &act, NULL);
    
    sigaction(SIGSEGV, NULL, &act);
    act.sa_flags |= SA_RESTART;
    act.sa_handler = GC_sh_error_handler;
    sigaction(SIGSEGV, &act, NULL);
    
    sigaction(SIGILL, NULL, &act);
    act.sa_flags |= SA_RESTART;
    act.sa_handler = GC_sh_error_handler;
    sigaction(SIGILL, &act, NULL);

    sigaction(SIGINT, NULL, &act);
    act.sa_flags |= SA_RESTART;
    act.sa_handler = GC_sh_int_handler;
    sigaction(SIGINT, &act, NULL);
#endif /* USE_DEBUG_HANDLER */
    return 0;
}

/* For debug */
void GC_reset_debug_handler()
{
#ifdef USE_DEBUG_HANDLER
    struct sigaction act;

    sigaction(SIGBUS, NULL, &act);
    act.sa_flags &= ~SA_RESTART;
    act.sa_handler = SIG_DFL;
    sigaction(SIGBUS, &act, NULL);
    
    sigaction(SIGSEGV, NULL, &act);
    act.sa_flags &= ~SA_RESTART;
    act.sa_handler = SIG_DFL;
    sigaction(SIGSEGV, &act, NULL);
    
    sigaction(SIGILL, NULL, &act);
    act.sa_flags &= ~SA_RESTART;
    act.sa_handler = SIG_DFL;
    sigaction(SIGILL, &act, NULL);

    sigaction(SIGINT, NULL, &act);
    act.sa_flags &= ~SA_RESTART;
    act.sa_handler = SIG_DFL;
    sigaction(SIGINT, &act, NULL);
#endif /* USE_DEBUG_HANDLER */
}
#endif

#   ifdef STUBBORN_ALLOC
void GC_stubborn_init0(){
    GC_stubborn_init();
}
#endif

/*******************************************************************/
/* Should be called only once among all processes. */
int GC_system_init(int n_gcprocs, SGC_attr_t *attr)
{
    volatile word *chp;
    int shm_size;
    int use_polling;
    int np_fixed;

    if (attr) {
	shm_size = attr->mmapsize;
	use_polling = attr->use_polling;
	np_fixed = attr->np_fixed;
    }
    else {
	shm_size = 0;
	use_polling = 0;
	np_fixed = 0;
    }
#ifndef SIGNAL_SUPPORTED
    use_polling = 1;
#endif
#ifndef ADDRS_SEPARATE
    shm_size = 0; /* ignored in thread version */
#endif

    GC_sh_get_mem_init(shm_size);
#ifdef SGC_DEBUGPRINT
    GC_err_printf2("*GC_arraysp is %08x -- %08x\n",
		   (void*)GC_arraysp, (void*)(GC_arraysp+1));
#endif
    GC_scratch_alloc_init(&GC_g_scratch);

    GC_use_polling = use_polling;
    GC_np_fixed = np_fixed;

    GC_thrpkg_init(n_gcprocs);
    GC_g_mainpid = GC_sh_getpid();
    atexit(GC_sh_exitfunc); /* UNIX only... How do we do in Windows? */

    /* user visible structure, which contains GC_dont_gc, GC_gc_no.. */
    GC_publicinfop = (struct _GC_publicinfo *)
	GC_scratch_g_alloc(sizeof(struct _GC_publicinfo));
    
    GC_appproc_envps_init();
    GC_gcproc_envps_init();

    /*** Setup synchronization variables... */
    {
	int maxsem;
	int maxbar;
	maxsem = 100;
	maxbar = 1;
	GC_blocksync_init(&GC_blocksync, maxsem, maxbar);
    }

    GC_bssem_init(&GC_g_sem, &GC_blocksync);

    GC_init_obj_kinds();

    PSM_mutex_init(&GC_g_alloc_ml);
    PSM_mutex_init(&GC_g_debug.print_ml);
    PSM_mutex_init(&GC_g_debug.exitprint_ml);
    GC_g_debug.gcing = 0;
#ifdef DO_PRECLAIM
    PSM_mutex_init(&GC_g_pr.freehblk_ml); 
    PSM_mutex_init(&GC_g_pr.hdr_fl_ml); 
#endif
    PSM_mutex_init(&GC_thread_funcarg.ml); 
    PSM_mutex_init(&GC_scan_ptr_ml);
    PSM_mutex_init(&GC_heap_sects_ml);

    /*** Setup miscellaneous variables... */
    GC_fail_count = 0;
    GC_all_bottom_indices = 0;
    GC_hblkfreelist = 0;
    GC_savhbp = (struct hblk *)0;
    GC_freehblk_ptr = 0;
    GC_n_heap_sects = 0;
    GC_root_size = 0;
    GC_least_plausible_heap_addr = (ptr_t)ONES;
    GC_greatest_plausible_heap_addr = 0;
    GC_dont_gc = 0;
    GC_dont_expand = 0;
    GC_non_gc_bytes = 0;
    GC_gc_no = 0;
    GC_full_freq = 4;
    GC_free_space_divisor = 4;
    GC_quiet = 0;
    GC_spmd_start_count = 1;
#ifdef PROCESSOR_BIND_SUPPORTED
    GC_g_bind_support.count = 0;
    PSM_mutex_init(&GC_g_bind_support.ml);
#endif /* PROCESSOR_BIND_SUPPORTED */
#ifdef SDSM_TEST
    GC_arraysp->current_pe = 0;
    GC_arraysp->current_userpe = 1;
#endif
    GC_g_finalization_failures = 0;
    GC_g_dl_head = 0;
    GC_g_log_dl_table_size = -1;
    GC_g_dl_entries = 0;
    PSM_mutex_init(&GC_g_dl_ml);
    GC_g_fo_head = 0;
    GC_g_finalize_now = 0;
    GC_g_log_fo_table_size = -1;
    GC_g_fo_entries = 0;
    PSM_mutex_init(&GC_g_fo_ml);
    

    if  (sizeof (ptr_t) != sizeof(word)) {
        ABORT("sizeof (ptr_t) != sizeof(word)\n");
    }
    if  (sizeof (signed_word) != sizeof(word)) {
        ABORT("sizeof (signed_word) != sizeof(word)\n");
    }
    if  (sizeof (struct hblk) != HBLKSIZE) {
        ABORT("sizeof (struct hblk) != HBLKSIZE\n");
    }

    GC_init_headers();
    GC_bl_init();

    GC_register_displacement_inner(0L);
#   ifdef MERGE_SIZES
    GC_init_size_map();
#   endif
#   ifdef STUBBORN_ALLOC
    GC_stubborn_init0();
#   endif

    /* Set up global reclaim lists. */
    GC_alloc_grl(NORMAL);
    GC_alloc_grl(PTRFREE);
    GC_alloc_grl(UNCOLLECTABLE);
# ifdef ATOMIC_UNCOLLECTABLE
    GC_alloc_grl(AUNCOLLECTABLE);
# endif
# ifdef STUBBORN_ALLOC
    GC_alloc_grl(STUBBORN);
# endif
    GC_obj_map_init();

    GC_LDLOG_INIT();
    GC_ALLOCLOG_INIT();
    GC_PICLOG_INIT();
    GC_MARKLOG_INIT();
    GC_GCRLOG_INIT();
    GC_MSLOG_INIT();
#ifdef USE_DEMOWIN
    GC_demowin_init();
#endif

    /* Setup stack bottom of leader process (this process). */
#ifdef SGC_DEBUGPRINT
# ifdef STACKBOTTOM
    GC_err_printf0("STACKBOTTOM is defined\n");
# endif
# ifdef HEURISTIC1
    GC_err_printf0("HEURISTIC1 is defined\n");
# endif
# ifdef HEURISTIC2
    GC_err_printf0("HEURISTIC2 is defined\n");
# endif
#endif
    GC_leader_stackbottom = GC_get_stack_base();

#ifdef SIGNAL_SUPPORTED
    if (GC_sh_is_signalhandler_separate() == 0) {
	GC_setup_debug_handler();
	if (GC_use_polling == 0) {
	    GC_setup_stop_handler();
	}
    }
#endif

#ifdef SDSM_TEST
    /* broadcast global variables */
    S_broadcast_nowait((int *)&GC_arraysp);
    S_broadcast_nowait((int *)&GC_publicinfop);
#endif

    if (GC_np_fixed == 0) {
	/* Fork all GC procs */
	GC_create_gcprocs(n_gcprocs);
    }

    GC_is_initialized = TRUE;
#ifdef SDSM_TEST
    /* broadcast global variables */
    S_broadcast_nowait((int *)&GC_is_initialized);
#endif

    return 0;
}

/****** API visible to user */
int SGC_attr_init(SGC_attr_t *attr)
{
    BZERO((void*)attr, sizeof(SGC_attr_t));
    return 0;
}

/****** API visible to user */
/* Should be called only once among all processes. */
int SGC_init(int n_gcprocs, SGC_attr_t *attr)
{
    int rcd;
    rcd = GC_system_init(n_gcprocs, attr);
    if (rcd < 0) {
	return rcd;
    }

    /* initialize this process */
    GC_process_creation_hook((ptr_t)GC_leader_stackbottom, SGC_MAIN_THREAD);

    return 0;
}

void SGC_disable_stop_world()
{
    GC_thread envp;
    envp = GC_lookup_appproc_envp(GC_sh_getpid());
    if (envp) {
	PSM_mutex_lock_polling(&GC_la_alloc_ml(envp), SGC_polling);
    }
}

void SGC_enable_stop_world()
{
    GC_thread envp;
    envp = GC_lookup_appproc_envp(GC_sh_getpid());
    if (envp) {
	PSM_mutex_unlock(&GC_la_alloc_ml(envp));
    }
}

/* called by each App/SPMD process. */
/* If SGC_EXPLICIT_INIT_LOCAL == 1 (See gc_tls.h), */
/* All application processes(threads) must call this function explicitly */
/* before any GC_malloc. */
/* If SGC_EXPLICIT_INIT_LOCAL == 0, this function is called implicitly. */
/* When this function is called, setting to GC_my_envp must have been */
/* enabled. (In some environment, this condition may not hold immediately */
/* after thread creation.) */
int SGC_init_local()
{
    DECLARE_MYENV2
    GC_thread envp;
    void *gcreg;
    GC_thread_id pid;

    pid = GC_sh_getpid();
    /* this mutex-lock avoids accidental concurrent execution of */
    /* this app process and gc process. */
    PSM_mutex_lock(&GC_g_alloc_ml);
    envp = GC_lookup_appproc_envp(pid);

    if (envp == NULL) {
	GC_err_printf1("[SGC_init_local]failed to initialize thread(id %d).\n",
		       pid);
	while (1) ;
    }

    SGC_set_myenvp(envp);
    GC_my_envp = envp;

    /* Now, access to GC_my_envp is enabled. */
    /* Remember current register variable in preparation for */
    /* future stop handler. */
    gcreg = SGC_save_regvar();
    GC_la_gcregvar(envp) = gcreg;

    if (GC_l_initialized(envp)) {
	/* second invocation? */
	PSM_mutex_unlock(&GC_g_alloc_ml);
	return 0;
    }

#ifdef SGC_DEBUGPRINT
    GC_err_printf0("SGC_init_local: 1st\n");
#endif
    /* Add initial guess of root sets.  Do this first, since sbrk(0)	*/
    /* mightbe used.							*/
    /* [Parallel] this should be much earlier... */
    GC_register_data_segments(SGC_MYENV_ARG0);

    /* now this process can use GC API. */
    GC_l_initialized(envp) = 1;
    PSM_mutex_unlock(&GC_g_alloc_ml);

    return 0;
}

/* returns 1 if no more App/SPMD proc is alive. */
/* don't use GC_my_envp in this function. */
int GC_terminate_proc_inner(GC_thread_id pid)
{
    int endflag;
    GC_thread envp;

    envp = GC_lookup_appproc_envp(pid);
    if (envp == NULL) {
	return -1;
    }

    GC_ALLOCLOG_END_LOCAL(envp);

    GC_yield_all_blocks(envp);
#ifdef FIND_LEAK
    GC_yield_freelist(envp);
#endif /* FIND_LEAK */
    GC_yield_reclaim_list(envp);
    
    if (GC_np_fixed == 0) {
	GC_remove_appproc_envp(envp);
    }
    else {
	GC_remove_gcproc_envp(envp);
	GC_remove_appproc_envp(envp);
    }
    endflag = (GC_is_appproc_envps_empty());

#ifdef SGC_DEBUGPRINT
    GC_err_printf1("$$$ pid %d finished $$$\n", pid);
    if (endflag) {
	GC_err_printf0("$$$ now all processes finished $$$\n");
    }
#endif
    return endflag;
}

int GC_terminate_proc(GC_thread_id pid)
{
    int rc;
    PSM_mutex_lock_polling(&GC_g_alloc_ml, SGC_polling);
    rc = GC_terminate_proc_inner(pid);
    PSM_mutex_unlock(&GC_g_alloc_ml);
    return rc;
}

/****** API visible to user */
/* called by main thread before its termination. */
/* Only for statistics. */
int SGC_finalize_log()
{
#ifdef USE_DEMOWIN
    GC_demowin_end();
#endif
    GC_GCRLOG_END();
    GC_ALLOCLOG_END();
}

int GC_system_terminate()
{
    if (GC_np_fixed == 0) {
	GC_kill_gcprocs();
    }

    GC_blocksync_end(&GC_blocksync);
    GC_sh_get_mem_end();

#ifdef SIGNAL_SUPPORTED
    GC_reset_stop_handler();
    GC_reset_debug_handler();
#endif

    return 0;
}

/* program termination function registered by atexit().  */
void GC_sh_exitfunc()
{
    if (GC_sh_getpid() == GC_g_mainpid) {
	GC_system_terminate();
    }
    return;
}

/****** API visible to user */
/* called by each App/SPMD process. */
/* don't use GC_my_envp in this function. */
int SGC_finalize()
{
    GC_err_printf0("SGC_finalize is obsolete!\n");
    return 0;
}

/****** API visible to user */
/* called by one thread. */
int SGC_finalize_all()
{
    GC_err_printf0("SGC_finalize_all is obsolete!\n");
    return 0;
}

/* assumes lock to GC_g_alloc_ml. */
int GC_cleanup_finished_detached_procs()
{
    GC_thread envp;

    for (envp = GC_g_appproc_envps.head; envp; envp = envp->next) {
	if (GC_l_initialized(envp) == 0) continue;
	if (GC_l_terminated(envp) != 0) {
	    if (GC_la_flags(envp) & SGC_DETACHED) {
		/* remove this thread. */
#if defined SGC_DEBUGPRINT
		GC_err_printf1("Invoker cleans up terminated detached thread %d.\n",
			       GC_l_pid(envp));
#endif
		GC_terminate_proc_inner(GC_l_pid(envp));
		/* retry threads-loop. this is simple, but slow... */
		envp = GC_g_appproc_envps.head;
	    }
	    else {
#if defined SGC_DEBUGPRINT
		GC_err_printf1("Invoker ignores terminated non-detached thread %d.\n",
			       GC_l_pid(envp));
#endif
	    }
	}
    }

    return 0;
}

/*******************************************************************/
#ifdef SGC_BALANCED_HDR

int GC_balanced_hdr_init(GC_thread envp)
{
    GC_lg_hdr_free_list(envp) = NULL;
    PSM_mutex_init(&GC_lg_hdr_free_list_ml(envp));
    GC_lg_n_free_hdr(envp) = 0;
    return 0;
}

/* Called by GC or SPMD process. */
int GC_fill_hdr_free_list(SGC_MYENV_ARG0)
SGC_MYENV_ARG_DECL2
{
    DECLARE_MYENV
    int pagesize = GETPAGESIZE();
    int n;
    int size;
    hdr *p;
    int i;

    n = (int)(pagesize*NPAGE_FOR_HDR_ALLOC/sizeof(hdr));
    if (n < 4) {
	n = 4;
    }
#ifdef SGC_DEBUGPRINT
    GC_err_printf2("gc/spmd %08x allocates %d hhdrs.\n", GC_my_envp, n);
#endif

    p = (hdr *)GET_MEM(n * sizeof(hdr));
    if (p == NULL) {
	return -1;
    }

    /* Localize the allocated pages. */
    /* In Origin2000, memory pages are local to the processor */
    /* that accessed the pages first. */
    GC_touch_memory_region((char *)p, n * sizeof(hdr));

    /* Link n headers to hdr_free_list. */
    for (i = 0; i < n; i++) {
	if (i < n-1) {
	    p[i].hb_next = (struct hblk *) &p[i+1];
	}
	else {
	    p[n-1].hb_next = (struct hblk *) GC_myg_hdr_free_list;
	}
	p[i].hdrhome_envp = GC_my_envp;
    }

    GC_myg_hdr_free_list = &p[0];
    GC_myg_n_free_hdr += n;
#ifdef SGC_DEBUGPRINT
    GC_err_printf3("SGC_BALANCED_HDR: GC proc %08x allocd %d hblks (total %d).\n",
		   GC_my_envp, n, GC_myg_n_free_hdr);
#endif
    return 0;
}

/* Called by App or SPMD process. */
/* Send a request GC or SPMD process to allocate some hblkhdrs. */
int GC_request_fill_hdr_free_list(SGC_MYENV_ARG0)
SGC_MYENV_ARG_DECL2
{
    DECLARE_MYENV
    if (GC_np_fixed == 0) {
	/* request to GC process */
	GC_gcp_command = GCP_COM_FILLHDR;	
	GC_requestor_envp = GC_my_envp;
	GC_awake_all_gcprocs();
	GC_bssem_wait(&GC_my_sem);
    }
    else {
	int ngcp;
	/* request to SPMD process */
	GC_spmdp_command = SPMDP_COM_FILLHDR;	
	GC_requestor_envp = GC_my_envp;

	ngcp = GC_count_gcproc_envps();
	PSM_barrier_init(&GC_gcp_barrier, ngcp);

	/* Stop all SPMD process by signal. */
	GC_notify_async_to_all_appprocs(SGC_MYENV_ARG1 TRUE);
	GC_fill_hdr_free_list(SGC_MYENV_ARG0);
	PSM_barrier_sync(&GC_gcp_barrier);

	PSM_barrier_destroy(&GC_gcp_barrier);
    }
}

/* MT-Unsafe */
/* Assumes lock on GC_g_alloc_ml is acquired. */
/* called by App/SPMD proc. */
hdr * GC_alloc_balanced_hdr(SGC_MYENV_ARG_DECL1 struct hblk * h)
{
    GC_thread envp;
    int max_n_hdr;
    GC_thread target_envp;
    hdr * result;

    /* Find the richest process. */
    target_envp = NULL;
    max_n_hdr = 0;

    for (envp = GC_g_gcproc_envps.head; envp; envp = envp->next) {
	if (GC_l_initialized(envp) == 0) continue;
	if (GC_l_terminated(envp) != 0) continue;

	if (GC_lg_n_free_hdr(envp) > max_n_hdr) {
	    max_n_hdr = GC_lg_n_free_hdr(envp);
	    target_envp = envp;
	}
    }

    if (max_n_hdr == 0) {
	/* No GC process has free hblkhdrs. */
	GC_request_fill_hdr_free_list(SGC_MYENV_ARG0);

	/* Now all GC processes have free hblkhdrs. */
	/* Choose one of them. */
	target_envp = GC_g_appproc_envps.head;
    }

    PSM_mutex_lock(&GC_lg_hdr_free_list_ml(target_envp));
    if (GC_lg_hdr_free_list(target_envp) == 0) {
	PSM_mutex_unlock(&GC_lg_hdr_free_list_ml(target_envp));
	/* failed! */
	return NULL;
    }

    result = GC_lg_hdr_free_list(target_envp);
    GC_lg_hdr_free_list(target_envp) = (hdr *) (result -> hb_next);
    GC_lg_n_free_hdr(target_envp)--;
    PSM_mutex_unlock(&GC_lg_hdr_free_list_ml(target_envp));

    return(result);
}

/* MT-Safe */
/* called by App/GC/SPMD proc. */
void GC_free_balanced_hdr(hdr *hhdr, struct hblk *h)
{
    GC_thread envp;
    envp = hhdr->hdrhome_envp;

    MUTEX_LOCK_LOG(&GC_lg_hdr_free_list_ml(envp), GC_LDK_WHFL);
    hhdr -> hb_next = (struct hblk *) GC_lg_hdr_free_list(envp);
    GC_lg_hdr_free_list(envp) = hhdr;
    GC_lg_n_free_hdr(envp)++;
    PSM_mutex_unlock(&GC_lg_hdr_free_list_ml(envp));
}

#endif /* SGC_BALANCED_HDR */

ptr_t GC_approx_sp GC_PROTO((void)); 

/* This function is called by App/SPMD proc. */
void GC_push_all_stack_contents(SGC_MYENV_ARG0)
SGC_MYENV_ARG_DECL2
{
    DECLARE_MYENV
# ifdef STACK_GROWS_DOWN
    GC_push_all_contents(SGC_MYENV_ARG1  GC_approx_sp(), GC_mya_stackbottom );
# else
    GC_push_all_contents(SGC_MYENV_ARG1  GC_mya_stackbottom, GC_approx_sp() );
# endif

    return;
}

/* Simple barrier synchronization for App processes. */
/* This function must be called only one process. */
/* WARNING: Strictly speaking, this is not a barrier. */
/* Other processes do not wait for the termination of */
/* this function. */
int GC_wait_for_all_appprocs(long nextstage)
{
    GC_thread envp;
    long volatile *p;

    while (1) {
	for (envp = GC_g_appproc_envps.head; envp; envp = envp->next) {
	    if (GC_l_initialized(envp) == 0) continue;
	    if (GC_l_terminated(envp) != 0) continue;

	    p = &GC_la_barrier_stage(envp);
	    if (*p != nextstage) {
		break;
	    }
	}
	if (!envp) {
	    /* all processes reached */
	    break;
	}
    }

    return 0;
}

/* Task of App processes in one GC phase. */
int GC_gcwork_by_appproc(SGC_MYENV_ARG_DECL1 GC_bool invoker)
{
    DECLARE_MYENV

    /* setup mark state and flush mark stack. */
    GC_invalidate_mark_state(SGC_MYENV_ARG0);
#ifdef DO_REQ_REMOTE
    GC_myg_reqbuf.validsize = 0;
#endif
    GC_my_objects_are_marked = FALSE;

    GC_PICLOG_INIT_PHASE_LOCAL();
    GC_emit_roots(SGC_MYENV_ARG0);
    GC_PICLOG_END_PHASE_LOCAL();

    GC_bssem_clear(&GC_my_sem);

#if 1
    GC_mya_barrier_stage = 1;
#else
    /* Barrier of App procs should be blocking-barrier. */
#ifdef USE_BLOCKING_BARRIER
    GC_bsbar_sync(&GC_appp_bsbar, invoker);
#else
    PSM_barrier_sync(&GC_appp_barrier);
#endif
#endif

    if (invoker) {
#if 1
	GC_wait_for_all_appprocs(1);
#endif
	/* Wake up GC processes */
	GC_gcp_command = GCP_COM_COLLECTION;
	GC_awake_all_gcprocs();
    }
    /* Wait to be awaked by GC processes. */
    GC_bssem_wait(&GC_my_sem);

    /* clear counters */
    GC_clear_app_counters(SGC_MYENV_ARG0);

#if 1
    GC_mya_barrier_stage = 2;
    /* Barrier of App procs should be blocking-barrier. */
#else
#ifdef USE_BLOCKING_BARRIER
    GC_bsbar_sync(&GC_appp_bsbar, invoker);
#else
    PSM_barrier_sync(&GC_appp_barrier);
#endif
#endif
    if (invoker) {
	GC_wait_for_all_appprocs(2);
	GC_sumup_app_counters();
    }

    /* Restart application... */
    return 0;
}

/* Task of SPMD processes in one GC phase. */
int GC_gcwork_by_spmdproc(SGC_MYENV_ARG_DECL1 GC_bool invoker)
{
    DECLARE_MYENV

    /* setup mark state and flush mark stack. */
    GC_invalidate_mark_state(SGC_MYENV_ARG0);
#ifdef DO_REQ_REMOTE
    GC_myg_reqbuf.validsize = 0;
#endif
    GC_my_objects_are_marked = FALSE;

    GC_PICLOG_INIT_PHASE_LOCAL();
    GC_emit_roots(SGC_MYENV_ARG0);
    GC_PICLOG_END_PHASE_LOCAL();

    if (invoker) {
	/* determine gc leader thread */
	GC_g_gcleader_envp = GC_my_envp;
    }

    PSM_barrier_sync(&GC_gcp_barrier);
    GC_marksweep_local(SGC_MYENV_ARG0); /* gc main routine */
    PSM_barrier_sync(&GC_gcp_barrier);

    /* clear counters */
    GC_clear_app_counters(SGC_MYENV_ARG0);
    PSM_barrier_sync(&GC_gcp_barrier);
    if (invoker) {
	GC_sumup_app_counters();
    }

    /* Restart application... */
    return 0;
}

void GC_async_work(SGC_MYENV_ARG0)
SGC_MYENV_ARG_DECL2
{
    if (GC_np_fixed == 0) {
	GC_gcwork_by_appproc(SGC_MYENV_ARG1 FALSE);
    }
    else {
	switch (GC_spmdp_command) {
	case SPMDP_COM_COLLECTION:
	    GC_gcwork_by_spmdproc(SGC_MYENV_ARG1 FALSE);
	    break;
	case SPMDP_COM_FILLHDR:
#ifdef SGC_BALANCED_HDR
	    GC_fill_hdr_free_list(SGC_MYENV_ARG0);
	    PSM_barrier_sync(&GC_gcp_barrier);
#endif
	    break;
	}
    }
    return;
}

#ifdef SIGNAL_SUPPORTED
void GC_sh_stop_handler(sig)
  int sig;
{
    DECLARE_MYENV2
    void *osreg;
    void *gcreg;
    GC_thread envp;

    osreg = SGC_save_regvar(); /* save register variable of os */
#ifdef SGC_DEBUGPRINT
    GC_err_printf1("STOP_HANDLER, osreg = %08x\n", osreg);
#endif
    envp = GC_lookup_appproc_envp(GC_sh_getpid());
    gcreg = GC_la_gcregvar(envp);
    SGC_restore_regvar(gcreg); /* restore register variable for GC */
    GC_my_envp = envp;
    /* now access to GC_my_envp is possible. */

#ifdef SGC_DEBUGPRINT
    GC_err_printf3("STOP_HANDLER, newreg = %08x, myenvp = %08x, pid %d\n",
		   gcreg, GC_my_envp, GC_sh_getpid());
    fflush(stderr);
#endif

    GC_async_work(SGC_MYENV_ARG0);

    SGC_restore_regvar(osreg); /* restore register variable of os */
    /* now returning from signal handler is possible */
    return;
}

void GC_sh_error_handler(sig)
  int sig;
{
    DECLARE_MYENV2
    int myid;
    void *osreg;
    void *gcreg;
    GC_thread envp;

    osreg = SGC_save_regvar(); /* save register variable of os */
    envp = GC_lookup_appproc_envp(GC_sh_getpid());
    gcreg = GC_la_gcregvar(envp);
    SGC_restore_regvar(gcreg); /* restore register variable for GC */
    /* now access to GC_my_envp is possible. */

    myid = GC_sh_getpid();
    GC_err_printf2("!!!!! P/TID %d received SIGNAL %d.\n",
	    myid, sig);

    PSM_mutex_lock(&GC_g_debug.exitprint_ml);

    GC_err_printf2("!!!!! P/TID %d received SIGNAL %d and exit..\n",
	    myid, sig);
    GC_err_printf3("!!!!! P/TID %d proc kind %c: debug_state %d\n",
	    myid, "AG"[GC_my_app_or_gc], GC_my_debug_state);
    fflush(stderr);
    PSM_mutex_unlock(&GC_g_debug.exitprint_ml);

    GC_sh_exitfunc();
    abort();
}

void GC_sh_int_handler(sig)
  int sig;
{ 
    GC_err_printf0("\nreceived SIGINT.\n");
    GC_sh_exitfunc();
    abort();
}

#endif

void SGC_polling_inner(SGC_MYENV_ARG0)
SGC_MYENV_ARG_DECL2
{
    DECLARE_MYENV
    if (GC_mya_polling_flag == 0) {
	return;
    }
    if (GC_use_polling == 0) {
	return;
    }
    GC_async_work(SGC_MYENV_ARG0);
    GC_mya_polling_flag = 0;
    return;
}

/********** API visible to user */
void SGC_polling()
{
    DECLARE_MYENV
    DECLARE_MYENV3
    if (GC_mya_polling_flag) {
	SGC_polling_inner(SGC_MYENV_ARG0);
    }
    return;
}

/*******************************************************************/
/* clear app_stat after GC */
int GC_clear_app_counters(SGC_MYENV_ARG0)
SGC_MYENV_ARG_DECL2
{
    DECLARE_MYENV
    appproc_stat_t *l_statp;
    l_statp = &GC_mya_app_stat;

    l_statp->words_allocd_before_gc +=
	l_statp->words_allocd;
    l_statp->non_gc_bytes_at_gc =
	l_statp->non_gc_bytes;
    l_statp->words_allocd = 0;
    l_statp->words_wasted = 0;
    l_statp->mem_freed = 0;
    return 0;
}

int GC_sumup_app_counters()
{
    GC_thread envp;
    appproc_stat_t *l_statp;

    GC_sum_app_stat.words_allocd = 0;
    GC_sum_app_stat.words_wasted = 0;
    GC_sum_app_stat.words_finalized = 0;
    GC_sum_app_stat.non_gc_bytes_at_gc = 0;
    GC_sum_app_stat.non_gc_bytes = 0;
    GC_sum_app_stat.mem_freed = 0;
    GC_sum_app_stat.words_allocd_before_gc = 0;

    for (envp = GC_g_appproc_envps.head; envp; envp = envp->next) {
	if (GC_l_initialized(envp) == 0) continue;
	if (GC_l_terminated(envp) != 0) continue;
	l_statp = &GC_la_app_stat(envp);
	GC_sum_app_stat.words_allocd += 
	    l_statp->words_allocd;
	GC_sum_app_stat.words_wasted += 
	    l_statp->words_wasted;
	GC_sum_app_stat.words_finalized += 
	    l_statp->words_finalized;
	GC_sum_app_stat.non_gc_bytes_at_gc += 
	    l_statp->non_gc_bytes_at_gc;
	GC_sum_app_stat.non_gc_bytes += 
	    l_statp->non_gc_bytes;
	GC_sum_app_stat.mem_freed += 
	    l_statp->mem_freed;
	GC_sum_app_stat.words_allocd_before_gc += 
	    l_statp->words_allocd_before_gc;
    }
    return 0;
}

int GC_sumup_gc_counters()
{
    GC_thread envp;
    gcproc_stat_t *l_statp;

    GC_sum_gc_stat.mem_found = 0;
# ifdef GATHERSTATS
    GC_sum_gc_stat.composite_in_use = 0;
    GC_sum_gc_stat.atomic_in_use = 0;
# endif
    for (envp = GC_g_gcproc_envps.head; envp; envp = envp->next) {
	if (GC_l_initialized(envp) == 0) continue;
	if (GC_l_terminated(envp) != 0) continue;
	l_statp = &GC_lg_gc_stat(envp);
	GC_sum_gc_stat.mem_found += l_statp->mem_found;
# ifdef GATHERSTATS
	GC_sum_gc_stat.composite_in_use += l_statp->composite_in_use;
	GC_sum_gc_stat.atomic_in_use += l_statp->atomic_in_use;
# endif
    }
    return 0;
}

/*******************************************************************/
/* Called by one App/SPMD proc(GC invoker). */
/* This avoids interrupt during GC_malloc. */
int GC_lock_all_la_alloc(SGC_MYENV_ARG0)
SGC_MYENV_ARG_DECL2
{
    DECLARE_MYENV
    GC_thread envp;

    for (envp = GC_g_appproc_envps.head; envp; envp = envp->next) {
	/* Skip self */
	if (GC_l_initialized(envp) == 0) continue;
	if (GC_l_terminated(envp) != 0) continue;
	if (envp == GC_my_envp) continue;
	PSM_mutex_lock(&GC_la_alloc_ml(envp));
    }
    return 0;
}

/* Called by one of App/SPMD proc(GC invoker). */
int GC_unlock_all_la_alloc(SGC_MYENV_ARG0)
SGC_MYENV_ARG_DECL2
{
    DECLARE_MYENV
    GC_thread envp;

    for (envp = GC_g_appproc_envps.head; envp; envp = envp->next) {
	/* Skip self */
	if (GC_l_initialized(envp) == 0) continue;
	if (GC_l_terminated(envp) != 0) continue;
	if (envp == GC_my_envp) continue;
	PSM_mutex_unlock(&GC_la_alloc_ml(envp));
    }
    return 0;
}

/*******************************************************************/

/* called by one of App processes. */
/* assumes lock. NOT MT-safe. */
int SGC_gcollect_inner(SGC_MYENV_ARG0)
SGC_MYENV_ARG_DECL2
{
    GC_lock_all_la_alloc(SGC_MYENV_ARG0);
    GC_GCRLOG_INIT_GC();
    GC_GCRLOG_SET_CLOCK(0);

    {
	GC_sumup_app_counters();
#   ifdef SGC_DEBUGPRINT
	GC_dump_list_stat();
#   endif
#   ifdef PRINTSTATS
	GC_printf2(
	   "Initiating full world-stop collection %lu after %ld allocd bytes\n",
	   (unsigned long) GC_gc_no+1,
	   (long)WORDS_TO_BYTES(GC_sum_app_stat.words_allocd));
#   endif

	GC_promote_black_lists();
	GC_clear_marks(SGC_MYENV_ARG0);
#   ifdef SAVE_CALL_CHAIN
        GC_save_callers(GC_last_stack);
#   endif
    }

    if (GC_np_fixed == 0) {
	int nappp;
	nappp = GC_count_appproc_envps();
#ifdef USE_BLOCKING_BARRIER
	GC_bsbar_init(&GC_appp_bsbar, &GC_blocksync, nappp);
#else
	PSM_barrier_init(&GC_appp_barrier, nappp);
#endif
	/* Stop-the-world */
	GC_notify_async_to_all_appprocs(SGC_MYENV_ARG1 TRUE);
	/* self work */
	GC_gcwork_by_appproc(SGC_MYENV_ARG1 TRUE);
#ifdef USE_BLOCKING_BARRIER
	GC_bsbar_end(&GC_appp_bsbar);
#else
	PSM_barrier_destroy(&GC_appp_barrier);
#endif
	GC_cleanup_finished_detached_procs();
    }
    else {
	int ngcp;
	ngcp = GC_count_gcproc_envps();
	PSM_barrier_init(&GC_gcp_barrier, ngcp);
	/* Stop-the-world */
	GC_spmdp_command = SPMDP_COM_COLLECTION;
	GC_notify_async_to_all_appprocs(SGC_MYENV_ARG1 TRUE);
	/* self work */
	GC_gcwork_by_spmdproc(SGC_MYENV_ARG1 TRUE);
	PSM_barrier_destroy(&GC_gcp_barrier);
	GC_cleanup_finished_detached_procs();
    }

    GC_GCRLOG_SET_CLOCK(3);
    GC_GCRLOG_END_GC();
    GC_unlock_all_la_alloc(SGC_MYENV_ARG0);

    return 0;
}

/* called by GC/SPMD processes. */
void GC_stopped_mark_init_local(SGC_MYENV_ARG0)
SGC_MYENV_ARG_DECL2
{
    DECLARE_MYENV

    if (GC_np_fixed == 0) {
	/* setup mark state and flush mark stack. */
	GC_invalidate_mark_state(SGC_MYENV_ARG0);

#ifdef DO_REQ_REMOTE
	GC_myg_reqbuf.validsize = 0;
#endif
	GC_my_objects_are_marked = FALSE;
    }

    GC_clear_a_few_frames();
    GC_noop(0,0,0,0,0,0);

#ifdef LOCAL_HDR_COPY
    {
	bottom_index *org;
	bottom_index *copy;
	bottom_index **prevcopy;
	int i;

	/* copy all_bottom_indices list */
	prevcopy = &GC_myg_gcenvp->_all_bottom_indices_copy;
	for (org = GC_all_bottom_indices; org; org = org->asc_link) {
	    copy = (bottom_index*)
		GC_scratch_alloc(SGC_MYENV_ARG1 (word)(sizeof (bottom_index)));
	    BCOPY(org, copy, sizeof(bottom_index));
	    for (i = 0; i < BOTTOM_SZ; i++) {
		if (!IS_FORWARDING_ADDR_OR_NIL(copy->index[i]) &&
		    !IS_SOLE_HHDR(copy->index[i])) {
		    copy->index[i] = NOT_COPIED_HDR;
		}
	    }
	    *prevcopy = copy;
	    prevcopy = &copy->asc_link;
	}
    }
#endif
}

ptr_t GC_stopped_mark_local(SGC_MYENV_ARG0)
SGC_MYENV_ARG_DECL2
{
    DECLARE_MYENV
    GC_stopped_mark_init_local(SGC_MYENV_ARG0);
    for (;;) {
	for(;;) {
	    if (GC_mark_some(SGC_MYENV_ARG0)) {
		break;
	    }
# ifdef DO_LOAD_DIST
	    /* if I have many works, export some of them */
	    GC_PICLOG_CHANGE_MODE(2);
	    GC_push_overload(SGC_MYENV_ARG0);
	    GC_PICLOG_CHANGE_MODE(0);
# endif
	}
# ifdef DO_LOAD_DIST
	/* Get works from other processes */
	GC_PICLOG_CHANGE_MODE(2);
	if (GC_get_work_or_terminate(SGC_MYENV_ARG0)) {
	    /* detected termination of mark phase */
	    GC_PICLOG_CHANGE_MODE(0);
	    break;
	}
	GC_PICLOG_CHANGE_MODE(0);
	/* continue marking */
	GC_my_mark_state = MS_ROOTS_PUSHED;
# else
	break;
# endif
#ifdef USE_DEMOWIN
	GC_demowin_ms_update();
#endif
    }
#ifdef USE_DEMOWIN
    GC_demowin_ms_update();
#endif

    return (ptr_t)TRUE;
}

#ifdef COUNT_CROWD
static void print_crowd_count_for_block(h, dummy)
struct hblk *h;
word dummy;
{
    register hdr * hhdr = HDR(h);

    if (hhdr->crowd_max) {
	GC_err_printf1("{%ld}", hhdr->crowd_max);
    }
}
#endif

#if 0 /* for Debug. */
static void destroy_unmarked(struct hblk *hbp, word dummy)
{
    int word_no;
    word *p, *q, *plim;
    word sz;
    hdr *hhdr;
    
    GET_HDR(hbp, hhdr);
    sz = hhdr->hb_sz;
    if (!SMALL_OBJ(WORDS_TO_BYTES(sz))) return;

    p = (word *)(hbp->hb_body);
    word_no = HDR_WORDS;
    plim = (word *)((((word)hbp) + HBLKSIZE)
		   - WORDS_TO_BYTES(sz));

    /* go through all words in block */
	while( p <= plim )  {
	    if( mark_bit_from_hdr(hhdr, word_no) ) {
		p += sz;
	    } else {
		/* Clear object, advance p to next object in the process */
		    q = p + sz;
                    p++; /* Skip link field */
                    while (p < q) {
			/* destroy! */
			*p++ = 0x10000000 | ((word)p & 0x0fffffff);
		    }
	    }
	    word_no += sz;
	}
}
#endif

/* called by all GC/SPMD processes. */
int GC_marksweep_local(SGC_MYENV_ARG0)
SGC_MYENV_ARG_DECL2
{
    DECLARE_MYENV
#   ifdef PRINTSTATS
    CLOCK_TYPE start_time, current_time;
#   endif
    
    if (GC_my_envp == GC_g_gcleader_envp) {
	GC_GCRLOG_SET_CLOCK(1);
    }

#   ifdef PRINTSTATS
    if (GC_my_envp == GC_g_gcleader_envp) {
	GET_TIME(start_time);
	GC_printf1("--> Marking for collection %lu ",
	           (unsigned long) GC_gc_no + 1);
	GC_printf2("after %lu allocd bytes + %lu wasted bytes\n",
	   	   (unsigned long)WORDS_TO_BYTES(GC_sum_app_stat.words_allocd),
	   	   (unsigned long)WORDS_TO_BYTES(GC_sum_app_stat.words_wasted));
    }
#   endif

    /*****/
    if (GC_my_envp == GC_g_gcleader_envp) {
# ifdef USE_DEMOWIN
	GC_demowin_init_phase();
# endif
# ifdef DO_LOAD_DIST
	GC_ld_init_phase();
# endif
	GC_MARKLOG_INIT_PHASE();
	GC_MSLOG_INIT_PHASE();
	GC_CACHELOG_INIT_PHASE();
    }

   
    PSM_barrier_sync(&GC_gcp_barrier);
    
#   ifdef GATHERSTATS
    /* Counters reflect currently marked objects: reset here */
    GC_myg_gc_stat.composite_in_use = 0;
    GC_myg_gc_stat.atomic_in_use = 0;
#   endif

    if (GC_my_envp == GC_g_gcleader_envp) {
	GC_g_debug.gcing = 1;
	GC_LDLOG_INIT_PHASE();
    }

    /* Mark phase */
    PSM_barrier_sync(&GC_gcp_barrier);
    GC_PICLOG_INIT_PHASE_LOCAL();
    if (GC_my_envp == GC_g_gcleader_envp) {
	GC_MARKLOG_START_CLOCK();
    }
    GC_stopped_mark_local(SGC_MYENV_ARG0);
    if (GC_my_envp == GC_g_gcleader_envp) {
	GC_MARKLOG_STOP_CLOCK();
    }
    GC_PICLOG_END_PHASE_LOCAL();
#ifdef SDSM_TEST
    {
	char str[256];
	sprintf(str, "[%d:%05x]: ! finished mark\n",
		S_my_pe(), S_my_threadid());
	fprintf(stderr, str);
    }
#endif
    PSM_barrier_sync(&GC_gcp_barrier);

#ifdef COUNT_CROWD
    if (GC_my_envp == GC_g_gcleader_envp) {
	GC_apply_to_all_blocks(print_crowd_count_for_block, (word)0);
    }
#endif

    if (GC_my_envp == GC_g_gcleader_envp) {
	GC_my_mark_state = MS_NONE;
	GC_LDLOG_END_PHASE();
#ifdef SGC_DEBUGPRINT
	{
	    GC_thread envp;
	    for (envp = GC_g_appproc_envps.head; envp; envp = envp->next) {
		if (GC_l_initialized(envp) == 0) continue;
		if (GC_l_terminated(envp) != 0) continue;
		GC_dump_freelist(envp);
	    }
	}
#endif
# ifdef USE_DEMOWIN
	GC_demowin_end_phase();
# endif
    }

    PSM_barrier_sync(&GC_gcp_barrier);

    if (GC_my_envp == GC_g_gcleader_envp) {
	GC_gc_no++;
#   ifdef PRINTSTATS
	GC_sumup_gc_counters();
	GC_printf2("Collection %lu reclaimed %ld bytes",
		   (unsigned long) GC_gc_no - 1,
		   (long)WORDS_TO_BYTES(GC_sum_gc_stat.mem_found));
	GC_printf1(" ---> heapsize = %lu bytes\n",
		   (unsigned long) GC_heapsize);
	/* Printf arguments may be pushed in funny places.  Clear the	*/
	/* space.							*/
	GC_printf0("");
#   endif      

#   ifdef PRINTTIMES
	GET_TIME(current_time);
	GC_printf1("World-stopped marking took %lu msecs\n",
	           MS_TIME_DIFF(current_time,start_time));
#   endif
    }

    if (GC_my_envp == GC_g_gcleader_envp) {
	GC_finalize_gc(SGC_MYENV_ARG0);
#     ifdef STUBBORN_ALLOC
        GC_clean_changing_list();
#     endif
    }

    if (GC_my_envp == GC_g_gcleader_envp) {
	GC_LDLOG_INIT_PHASE();
	GC_MARKLOG_START_CLOCK();
	/* prepare for reclaimation */
	GC_hbiter_start();
    }

    PSM_barrier_sync(&GC_gcp_barrier);
    GC_myg_gc_stat.mem_found = 0;
    GC_clear_lists_local(SGC_MYENV_ARG0);
    PSM_barrier_sync(&GC_gcp_barrier);
    GC_start_reclaim_local(SGC_MYENV_ARG0);
    PSM_barrier_sync(&GC_gcp_barrier);

    if (GC_my_envp == GC_g_gcleader_envp) {
	GC_make_hblkfl();
	GC_hbiter_end();
	GC_MARKLOG_STOP_CLOCK();
	GC_MSLOG_END_PHASE();
	GC_LDLOG_END_PHASE();
	GC_MARKLOG_END_PHASE();
	GC_PICLOG_END_PHASE();
	GC_CACHELOG_END_PHASE();
    }


    if (GC_my_envp == GC_g_gcleader_envp) {
	GC_g_debug.gcing = 0;
	GC_sumup_gc_counters();
#   ifdef PRINTSTATS
	GC_printf2(
		  "Immediately reclaimed %ld bytes in heap of size %lu bytes\n",
	          (long)WORDS_TO_BYTES(GC_sum_gc_stat.mem_found),
	          (unsigned long)GC_heapsize);
	GC_printf2("%lu (atomic) + %lu (composite) collectable bytes in use\n",
	           (unsigned long)WORDS_TO_BYTES(GC_sum_gc_stat.atomic_in_use),
	           (unsigned long)WORDS_TO_BYTES(GC_sum_gc_stat.composite_in_use));
#   endif
    }

#ifdef SGC_DEBUGPRINT
    if (GC_my_envp == GC_g_gcleader_envp) {
	GC_dump_list_stat();
    }
#endif

    PSM_barrier_sync(&GC_gcp_barrier);

    if (GC_my_envp == GC_g_gcleader_envp) {
	GC_GCRLOG_SET_CLOCK(2);
    }

    if (GC_np_fixed == 0) {
	/* NP-variable model */
	if (GC_my_envp == GC_g_gcleader_envp) {
	    /* Awake all App procs. */
	    GC_awake_all_appprocs();
	}
    }

    return 0;
}

/* */
int GC_sh_dummy(int s)
{
    return 0;
}

/* This function may be helpful when memory region should be */
/* on physical memory. */
int GC_touch_memory_region(char *ptr, size_t size)
{
    return 0;
/*
    volatile char *p;
    char c;
    int d = 0;
    int align = GETPAGESIZE();
    for (p = ptr; p < ptr+size; p += align) {
	c = *(p);
	d += c;
	*(p) = c;
    }
    return GC_sh_dummy(d);
*/
}

/* */
/* debug routines */
int GC_dump_list_stat()
{
/*    int sz; */
    struct hblk *hbp;
    hdr * hhdr;
/*    ptr_t fl; */
    int n1, n2, maxn;

    GC_err_printf0("=--------------------------------------------------=\n");
    /* hblk free list */
    n1 = 0;
    n2 = 0;
    maxn = 0;
    hbp = GC_hblkfreelist;
    while (hbp) {
	hhdr = HDR(hbp);
	n1++;
	n2 += hhdr->hb_sz / HBLKSIZE;
	if (hhdr->hb_sz / HBLKSIZE > maxn) {
	    maxn = hhdr->hb_sz / HBLKSIZE;
	}
	hbp = hhdr->hb_next;
    }
    GC_err_printf3("hblk f.l. %d cont-hblks, %d hblk, max %d hblks\n",
	    n1, n2, maxn);

    GC_err_printf0("=--------------------------------------------------=\n");
    return 0;
}

int GC_dump_freelist(GC_thread envp)
{
    register word size;		/* current object size		*/
    int kind;
    ptr_t fl;
    register ptr_t p;	/* pointer to current object	*/

    for (kind = 0; kind < GC_n_kinds; kind++) {
	for (size = 1; size <= MAXOBJSZ; size++) {
	    fl = GC_la_freelist(envp)[kind][size];
	    if (fl == NULL) continue;
	    GC_err_printf3("Proc %08x: Freelist[%d][%d] ->",
			   envp, kind, size);
	    for (p= fl; p != 0; p=obj_link(p)){
		GC_err_printf1("%08x->", p);
	    }
	    GC_err_printf0("\n");
	}
    }

    return 0;
}

void GC_debug_print_envp(SGC_MYENV_ARG0)
SGC_MYENV_ARG_DECL2
{
    DECLARE_MYENV
    GC_err_printf1("My envp = %08x\n", GC_my_envp);
}

void GC_set_debug_state(SGC_MYENV_ARG_DECL1 int ds)
{
    DECLARE_MYENV
    if (GC_my_envp) {
	GC_my_debug_state = ds;
    }
}

GC_thread GC_debug_get_envp(SGC_MYENV_ARG0)
SGC_MYENV_ARG_DECL2
{
    DECLARE_MYENV
    return GC_my_envp;
}

#if 0
int GC_debug_get_reg()
{
    int rc;
#ifdef SPARC
    asm volatile("add %%g7, %%g0, %0" : "=r" (rc) : );
#else
    rc = 0;
#endif
    return rc;
}
#endif

#ifdef SAVE_CALL_CHAIN
int GC_debug_call_chain()
{
    struct callinfo info[NFRAMES];
    GC_save_callers(info);
    GC_print_callers(info);
    return 0;
}
#endif
