/*
 * Copyright (c) 1998 by Toshio Endo, Kenjiro Taura. All rights reserved.
 * Copyright (c) 1998 by Mitsubishi Research Institute.  All rights reserved.
 * Copyright (c) 1998 by Information-technology Promotion Agency.  All rights reserved.
 *
 * THIS MATERIAL IS PROVIDED AS IS, WITH ABSOLUTELY NO WARRANTY EXPRESSED
 * OR IMPLIED.  ANY USE IS AT YOUR OWN RISK.
 *
 * Permission is hereby granted to use or copy this program
 * for any purpose,  provided the above notices are retained on all copies.
 * Permission to modify the code and to distribute modified code is granted,
 * provided the above notices are retained, and a notice that the code was
 * modified is included with the above copyright notice.
 *
 */
/* Endo, June 28, 1999 */

#include "gc_priv.h"
#ifdef SIGNAL_SUPPORTED
#  include <signal.h>
#endif

/*******************************************************************/
/* Functions for absorb difference among thread packages. */
/*******************************************************************/

/* extern struct appproc_entry *GC_alloc_appproc_entry (); */

/*******************************************************************/
#if defined SGC_FORK_THREAD

GC_thread_id GC_sh_fork_getpid()
{
    return (GC_thread_id)getpid();
}

int GC_sh_fork_sendsig(GC_thread_id pid, int sig)
{
    if (kill((pid_t)pid, sig) != 0) {
	return -1;
    }
    return 0;
}

int GC_sh_fork_is_signalhandler_separate()
{
    return 1;
}

int GC_sh_fork_wait(GC_thread_id pid)
{
    return waitpid((pid_t)pid, NULL, 0);
}

int GC_sh_fork_waitany(GC_thread_id *ppid)
{
    if ((*ppid = (GC_thread)wait()) < 0) {
	*ppid = 0;
	return -1;
    }
    return 0;
}

GC_thread_id GC_sh_fork_create_process(void*(*func)(void*), void *arg)
{
    pid_t rcd;
    if ((rcd = fork()) == 0) {
	(*func)(arg);
	exit(0);
    }
    return (GC_thread_id)rcd;
}

void GC_sh_fork_exit_process()
{
    exit(0);
}

/****** API visible to user */
pid_t GC_fork()
{
    pid_t pid;

    pid = fork();
    if (pid == 0) {
	/* Child process */
	/* Stack bottom is the same as that of leader process. */
	GC_process_creation_hook((ptr_t)GC_leader_stackbottom, 0);
    }

    return pid;
}

pid_t GC_wait(int *statp)
{
    pid_t pid = wait(statp);
    if (pid >= 0) {
	GC_terminate_proc((GC_thread_id)pid);
    }
    return pid;
}

pid_t GC_waitpid(pid_t pid, int *stat_loc, int options)
{
    pid_t finished = waitpid(pid, stat_loc, options);
    if (finished >= 0) {
	GC_terminate_proc((GC_thread_id)finished);
    }
    return finished;
}

#endif /* FORK */

/*******************************************************************/
#if defined SGC_SOLARIS_THREAD || \
    defined SGC_SOLARIS_THREAD_PTHREAD

GC_thread_id GC_sh_solaris_thread_getpid()
{
#ifndef SDSM_TEST
    return (GC_thread_id)thr_self();
#else
    return (GC_thread_id)S_my_threadid();
#endif
}

int GC_sh_solaris_thread_sendsig(GC_thread_id pid, int sig)
{
    if (thr_kill((thread_t)pid, sig) != 0) {
	return -1;
    }
    return 0;
}

int GC_sh_solaris_thread_is_signalhandler_separate()
{
    return 0;
}

int GC_sh_solaris_thread_wait(GC_thread_id pid)
{
    return thr_join((thread_t)pid, NULL, NULL);
}

int GC_sh_solaris_thread_waitany(GC_thread_id *ppid)
{
    return thr_join(0, (thread_t *)ppid, NULL);
}

GC_thread_id GC_sh_solaris_thread_create_process(void*(*func)(void*),
						 void *arg)
{
    thread_t tid;
    int rcd;
#ifndef SDSM_TEST
    rcd = thr_create(NULL, 0, func, arg, THR_BOUND, &tid);
#else
    rcd = S_create_thread(GC_arraysp->current_pe, func, arg, &tid);
    GC_arraysp->current_pe++;
#endif
    if (rcd != 0) {
	GC_err_printf0( "ERROR: failed to thr_create !!!\n");
	return -1;
    }
    return (GC_thread_id)tid;
}

void GC_sh_solaris_thread_exit_process()
{
    thr_exit(NULL);
}

int GC_sh_solaris_thread_setconcurrency(int n)
{
    return thr_setconcurrency(n);
}

/**************************/
void *GC_solaris_thread_wrapper(void *dummyarg)
{
    int dummy; /* To calculate stack bottom of this new thread. */
    void *rcd;
    void * (*func)(void *);
    void *rarg;
    int flags;

    void *oldreg = SGC_save_regvar(); /* Save register variable */

    func = GC_thread_funcarg.func;
    rarg = GC_thread_funcarg.arg;
    flags = GC_thread_funcarg.flags;
    /* Unlock funcarg structure. */
    PSM_mutex_unlock(&GC_thread_funcarg.ml);

    /* Setup stack bottom of child thread. */
    /* User start function is executed on the top of THIS stack frame. */
    GC_process_creation_hook((ptr_t)&dummy, flags);

    /* Call user start function */
    rcd = func(rarg);

    /* Possibly return here */
    SGC_restore_regvar(oldreg); /* Restore register variable */

    return rcd;
}

/****** API visible to user */
int GC_thr_create(void *stack_base, size_t stack_size,
              void *(*start_func)(void *), void *arg, long flags,
              void *nt)
{
    int rcd;
    thread_t *new_thread = (thread_t *)nt;

    /* Lock the funcarg structure. */
    /* This will be unlocked by child thread. */
    PSM_mutex_lock(&GC_thread_funcarg.ml);

    GC_thread_funcarg.func = start_func;
    GC_thread_funcarg.arg = arg;
    GC_thread_funcarg.flags = (flags & THR_DETACHED)? SGC_DETACHED: 0;

#ifndef SDSM_TEST
    rcd = thr_create(stack_base, stack_size,
		     GC_solaris_thread_wrapper, NULL,
		     flags, new_thread);
#else
    rcd = S_create_thread(GC_arraysp->current_userpe, SGC_solaris_thread_wrapper, NULL, new_thread);
    GC_arraysp->current_userpe++;
#endif

    return rcd;
}

int GC_thr_join(thread_t target_thread, thread_t *departed,
		void **status)
{
    thread_t tid;
    int rc;
    rc = thr_join(target_thread, &tid, status);
    if (rc == 0) {
	GC_terminate_proc((GC_thread_id)tid);
	if (departed) {
	    *departed = tid;
	}
    }
    return rc;
}

#endif /* SOLARIS_THREAD */

/*******************************************************************/
#if defined SGC_PTHREAD || \
    defined SGC_SOLARIS_THREAD_PTHREAD

GC_thread_id GC_sh_pthread_getpid()
{
#ifndef SDSM_TEST
    return (GC_thread_id)pthread_self();
#else
    return (GC_thread_id)S_my_threadid();
#endif
}

int GC_sh_pthread_sendsig(GC_thread_id pid, int sig)
{
    if (pthread_kill((pthread_t)pid, sig) != 0) {
	return -1;
    }
    return 0;
}

int GC_sh_pthread_is_signalhandler_separate()
{
    return 0;
}

int GC_sh_pthread_wait(GC_thread_id pid)
{
    return pthread_join((pthread_t)pid, NULL);
}

int GC_sh_pthread_waitany(GC_thread_id *ppid)
{
    /* not supported in pthread */
    return -1;
}

GC_thread_id GC_sh_pthread_create_process(void*(*func)(void*), void *arg)
{
    pthread_t tid;
    int rcd;
    pthread_attr_t attr;
    pthread_attr_init(&attr);
#ifndef PTHREAD_SETSCOPE_NOTSUPPORTED
    pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);
#endif
#ifndef SDSM_TEST
    rcd = pthread_create((pthread_t *)&tid, &attr, func, arg);
#else
    rcd = S_create_thread(GC_arraysp->current_pe, func, arg, &tid);
    GC_arraysp->current_pe++;
#endif
    if (rcd != 0) {
	GC_err_printf0( "ERROR: failed to pthread_create!!!\n");
	return -1;
    }
    return (GC_thread_id)tid;
}

void GC_sh_pthread_exit_process()
{
    pthread_exit(NULL);
}

int GC_sh_pthread_setconcurrency(int n)
{
#ifdef PTHREAD_SETCONC_SUPPORTED
    return pthread_setconcurrency(n);
#else
    /* normal pthread package does not support this */
    return -1;
#endif
}

/**************************/
void *GC_pthread_wrapper(void *dummyarg)
{
    int dummy; /* To calculate stack bottom of this new thread. */
    void *rcd;
    void * (*func)(void *);
    void *rarg;
    int flags;

    void *oldreg = SGC_save_regvar(); /* Save register variable */

    func = GC_thread_funcarg.func;
    rarg = GC_thread_funcarg.arg;
    flags = GC_thread_funcarg.flags;
    /* Unlock funcarg structure. */
    PSM_mutex_unlock(&GC_thread_funcarg.ml);

    /* Setup stack bottom of child thread. */
    /* User start function is executed on the top of THIS stack frame. */
    GC_process_creation_hook((ptr_t)&dummy, flags);

    /* Call user start function */
    rcd = func(rarg);

    /* Possibly return here */
    SGC_restore_regvar(oldreg); /* Restore register variable */

    return rcd;
}

/****** API visible to user */
int GC_pthread_create(void *nt, void *pa,
		       void *(*start_func)(void *), void *arg)
{
    int rcd;
    pthread_t *new_thread = (pthread_t *)nt;
    pthread_attr_t *pattr = pa;

    /* Lock the funcarg structure. */
    /* This will be unlocked by child thread. */
    PSM_mutex_lock(&GC_thread_funcarg.ml);

    GC_thread_funcarg.func = start_func;
    GC_thread_funcarg.arg = arg;
    if (pattr) { 
	int detachstate;
	pthread_attr_getdetachstate(pattr, &detachstate);
	GC_thread_funcarg.flags =
	    (detachstate == PTHREAD_CREATE_DETACHED)? SGC_DETACHED: 0;
    }
    else {
	GC_thread_funcarg.flags = 0;
    }

#ifndef SDSM_TEST
    rcd = pthread_create(new_thread, pattr,
			 GC_pthread_wrapper, NULL);
#else
    rcd = S_create_thread(GC_arraysp->current_userpe, GC_pthread_wrapper, NULL, new_thread);
    GC_arraysp->current_userpe++;
    if (GC_arraysp->current_userpe >= S_num_pes()) {
	GC_arraysp->current_userpe = 1;
    }
#endif

    return rcd;
}

int GC_pthread_join(pthread_t target_thread, void **status)
{
    int rc;
    rc = pthread_join(target_thread, status);
    if (rc == 0) {
	GC_terminate_proc((GC_thread_id)target_thread);
    }
    return rc;
}

#endif /* PTHREAD */

/*******************************************************************/
#if defined SGC_SPROC_THREAD

#include <sys/prctl.h> /* for PRDA */
#include <ulocks.h>

#if defined SGC_TLS_IMPORT_PUBLIC_FUNCTION || \
    defined SGC_TLS_IMPORT_EACH_FUNCTION
/* NOTICE: SGC library (with IRIX sproc) reserves a part of PRDA. */
/* Change this value if it conflicts with other module. */
#  define IRIX_PRDA_OFFSET 0
#endif

/* we must specify the maximum number of sproc processes... */
#define IRIX_SPROC_MAX 512

GC_thread_id GC_sh_sproc_getpid()
{
    return (GC_thread_id)getpid();
}

int GC_sh_sproc_sendsig(GC_thread_id pid, int sig)
{
    if (kill((pid_t)pid, sig) != 0) {
	return -1;
    }
    return 0;
}

int GC_sh_sproc_is_signalhandler_separate()
{
    return 1;
}

int GC_sh_sproc_wait(GC_thread_id pid)
{
    return waitpid((pid_t)pid, NULL, 0);
}

int GC_sh_sproc_waitany(GC_thread_id *ppid)
{
    if ((*ppid = (GC_thread_id)wait()) < 0) {
	*ppid = 0;
	return -1;
    }
    return 0;
}

GC_thread_id GC_sh_sproc_create_process(void*(*func)(void*), void *arg)
{
    pid_t rcd;
    rcd = sproc((void (*)(void *))func, PR_SALL, arg);
    if (rcd == -1) {
	GC_err_printf3( "ERROR: failed to sproc(%ld, %d, %d)!!!\n",
		(long)func, PR_SALL, arg);
	return -1;
    }
    return (GC_thread_id)rcd;
}

void GC_sh_sproc_exit_process()
{
    exit(0);
}

/**************************/
void GC_sproc_wrapper(void *dummyarg)
{
    int dummy; /* To calculate stack bottom of this new thread. */
    void *rcd;
    void * (*func)(void *);
    void *rarg;
    int flags;

    void *oldreg = SGC_save_regvar(); /* Save register variable */

    func = GC_thread_funcarg.func;
    rarg = GC_thread_funcarg.arg;
    flags = GC_thread_funcarg.flags;
    /* Unlock funcarg structure. */
    PSM_mutex_unlock(&GC_thread_funcarg.ml);

    /* Setup stack bottom of child thread. */
    /* User start function is executed on the top of THIS stack frame. */
    GC_process_creation_hook((ptr_t)&dummy, flags);

    /* Call user start function */
    rcd = func(rarg);

    /* Possibly return here */
    SGC_restore_regvar(oldreg); /* Restore register variable */

    return;
}

/****** API visible to user */
/* currently, users must call this function with three arguments. */
pid_t GC_sproc(void (*start_func)(void *), unsigned inh, void *arg)
{
    pid_t rcd;

    /* Lock the funcarg structure. */
    /* This will be unlocked by child thread. */
    PSM_mutex_lock(&GC_thread_funcarg.ml);

    GC_thread_funcarg.func = (void *(*)(void *))start_func;
    GC_thread_funcarg.arg = arg;
    GC_thread_funcarg.flags = 0;

    rcd = sproc(GC_sproc_wrapper, inh, NULL);

    return rcd;
}

pid_t GC_wait(int *statp)
{
    pid_t pid = wait(statp);
    if (pid >= 0) {
	GC_terminate_proc((GC_thread_id)pid);
    }
    return pid;
}

pid_t GC_waitpid(pid_t pid, int *stat_loc, int options)
{
    pid_t finished = waitpid(pid, stat_loc, options);
    if (finished >= 0) {
	GC_terminate_proc((GC_thread_id)finished);
    }
    return finished;
}


#endif /* SPROC */

/*******************************************************************/
/* dispatch according to using thread package. */

#ifdef SETCONC_SUPPORTED
int GC_sh_setconcurrency GC_PROTO((int n));
#endif

int GC_thrpkg_init(int n_gcprocs)
{
    SGC_tlskey_create();

#ifdef SETCONC_SUPPORTED
    GC_sh_setconcurrency(n_gcprocs);
#endif

#ifdef SGC_SPROC_THREAD
    if (usconfig(CONF_INITUSERS, IRIX_SPROC_MAX) == -1) {
	GC_err_printf0("WARNING in GC_thrpkg_init() : failed to usconfig().");
    }
#endif
    return 0;
}

GC_thread_id GC_sh_getpid()
{
#if defined SGC_FORK_THREAD
    return GC_sh_fork_getpid();
#endif
#if defined SGC_SOLARIS_THREAD || \
    defined SGC_SOLARIS_THREAD_PTHREAD
    return GC_sh_solaris_thread_getpid();
#endif
#if defined SGC_PTHREAD
    return GC_sh_pthread_getpid();
#endif
#if defined SGC_SPROC_THREAD
    return GC_sh_sproc_getpid();
#endif
}

#ifdef SIGNAL_SUPPORTED
int GC_sh_sendsig(GC_thread_id pid, int sig)
{
#if defined SGC_FORK_THREAD
    return GC_sh_fork_sendsig(pid, sig);
#endif
#if defined SGC_SOLARIS_THREAD || \
    defined SGC_SOLARIS_THREAD_PTHREAD
    return GC_sh_solaris_thread_sendsig(pid, sig);
#endif
#if defined SGC_PTHREAD
    return GC_sh_pthread_sendsig(pid, sig);
#endif
#if defined SGC_SPROC_THREAD
    return GC_sh_sproc_sendsig(pid, sig);
#endif
}
#endif

#ifdef SIGNAL_SUPPORTED
int GC_sh_is_signalhandler_separate()
{
#if defined SGC_FORK_THREAD
    return GC_sh_fork_is_signalhandler_separate();
#endif
#if defined SGC_SOLARIS_THREAD || \
    defined SGC_SOLARIS_THREAD_PTHREAD
    return GC_sh_solaris_thread_is_signalhandler_separate();
#endif
#if defined SGC_PTHREAD
    return GC_sh_pthread_is_signalhandler_separate();
#endif
#if defined SGC_SPROC_THREAD
    return GC_sh_sproc_is_signalhandler_separate();
#endif
}
#endif

int GC_sh_wait(GC_thread_id pid)
{
#if defined SGC_FORK_THREAD
    return GC_sh_fork_wait(pid);
#endif
#if defined SGC_SOLARIS_THREAD || \
    defined SGC_SOLARIS_THREAD_PTHREAD
    return GC_sh_solaris_thread_wait(pid);
#endif
#if defined SGC_PTHREAD
    return GC_sh_pthread_wait(pid);
#endif
#if defined SGC_SPROC_THREAD
    return GC_sh_sproc_wait(pid);
#endif
}

int GC_sh_waitany(GC_thread_id *ppid)
{
#if defined SGC_FORK_THREAD
    return GC_sh_fork_waitany(ppid);
#endif
#if defined SGC_SOLARIS_THREAD || \
    defined SGC_SOLARIS_THREAD_PTHREAD
    return GC_sh_solaris_thread_waitany(ppid);
#endif
#if defined SGC_PTHREAD
    return GC_sh_pthread_waitany(ppid); /* not supported in pthread */
#endif
#if defined SGC_SPROC_THREAD
    return GC_sh_sproc_waitany(ppid);
#endif
}

GC_thread_id GC_sh_create_process(void*(*func)(void*), void *arg)
{
#if defined SGC_FORK_THREAD
    return GC_sh_fork_create_process(func, arg);
#endif
#if defined SGC_SOLARIS_THREAD || \
    defined SGC_SOLARIS_THREAD_PTHREAD
    return GC_sh_solaris_thread_create_process(func, arg);
#endif
#if defined SGC_PTHREAD
    return GC_sh_pthread_create_process(func, arg);
#endif
#if defined SGC_SPROC_THREAD
    return GC_sh_sproc_create_process(func, arg);
#endif
}

void GC_sh_exit_process()
{
#if defined SGC_FORK_THREAD
    GC_sh_fork_exit_process();
    return;
#endif
#if defined SGC_SOLARIS_THREAD || \
    defined SGC_SOLARIS_THREAD_PTHREAD
    GC_sh_solaris_thread_exit_process();
    return;
#endif
#if defined SGC_PTHREAD
    GC_sh_pthread_exit_process();
    return;
#endif
#if defined SGC_SPROC_THREAD
    GC_sh_sproc_exit_process();
    return;
#endif
}

#ifdef SETCONC_SUPPORTED
int GC_sh_setconcurrency(int n)
{
#if defined SGC_SOLARIS_THREAD || \
    defined SGC_SOLARIS_THREAD_PTHREAD
    return GC_sh_solaris_thread_setconcurrency(n);
#elif defined SGC_PTHREAD
    return GC_sh_pthread_setconcurrency(n);
#else
    return -1;
#endif
}
#endif

/* utility API. calls SGC_fork, SGC_thr_create or SGC_pthread_create, */
/* or SGC_sproc according to the specified thread package. */
int SGC_create_process(void*(*func)(void*), void *arg)
{
#if defined SGC_FORK_THREAD
    {
	pid_t id;
	id = GC_fork();
	if (id == 0) {
	    (*func)(arg);
	    exit(0);
	}
	return (int)id;
    }
#endif
#if defined SGC_SOLARIS_THREAD || \
    defined SGC_SOLARIS_THREAD_PTHREAD
    {
	thread_t id;
	if (GC_thr_create(NULL, 0, func, arg, THR_BOUND, &id) != 0) {
	    return -1;
	}
	return (int)id;
    }
#endif
#if defined SGC_PTHREAD
    {
	pthread_attr_t attr;
	pthread_t id;
	pthread_attr_init(&attr);
#ifndef PTHREAD_SETSCOPE_NOTSUPPORTED
	pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);
#endif
	if (GC_pthread_create(&id, &attr, func, arg) != 0) {
	    GC_err_printf0( "ERROR: failed to pthread_create!!!\n");
	    return -1;
	}
	return (int)id;
    }
#endif
#if defined SGC_SPROC_THREAD
    {
	pid_t id;
	id = GC_sproc((void (*)(void *))func, PR_SALL, arg);
	return (int)id;
    }
#endif
}

/* utility API. calls wait, thr_join, pthread_join, */
/* according to `thread_package'. */
int SGC_wait_process(int pid)
{
#if defined SGC_FORK_THREAD || defined SGC_SPROC_THREAD
    {
	int stat;
	if (GC_waitpid(pid, &stat, 0) != -1) {
	    return stat;
	}
	else {
	    return -1;
	}
    }
#endif
#if defined SGC_SOLARIS_THREAD || \
    defined SGC_SOLARIS_THREAD_PTHREAD
    {
	void *stat;
	if (GC_thr_join((thread_t)pid, NULL, &stat) == 0) {
	    return (int)stat;
	}
	else {
	    return -1;
	}
    }
#endif
#if defined SGC_PTHREAD
    {
	void *stat;
	if (GC_pthread_join((pthread_t)pid, &stat) == 0) {
	    return (int)stat;
	}
	else {
	    return -1;
	}
    }
#endif
}

#if defined SGC_TLS_IMPORT_PUBLIC_FUNCTION || \
    defined SGC_TLS_IMPORT_EACH_FUNCTION
GC_thread  SGC_get_myenvp(void)
{
#if defined SGC_SOLARIS_THREAD || \
    defined SGC_SOLARIS_THREAD_PTHREAD
    {
	void *p;
	int err_no = 
	    thr_getspecific((thread_key_t)(GC_arraysp->tls_key), &p);
	if(err_no == 0){
	    return (GC_thread)p;
	}else{
	    GC_err_printf3( "ERROR in SGC_get_myenvp() : file = %s, line = %d, errno = %d\n", __FILE__, __LINE__, err_no);
	    return 0;
	}
    }
#endif
#if defined SGC_PTHREAD
    {
	void *p;
	p = pthread_getspecific((pthread_key_t)(GC_arraysp->tls_key));
	if(p == 0){
	    GC_err_printf3( "ERROR in SGC_get_myenvp() : key = %d, file = %s, line = %d\n", (GC_arraysp->tls_key), __FILE__, __LINE__);
	    return 0;
	}
	return (GC_thread)p;
    }
#endif
#if defined SGC_SPROC_THREAD
    {
	void *p;
	p = *((void **)&PRDA->usr_prda.fill[IRIX_PRDA_OFFSET]);
	if(p == 0){
	    GC_err_printf3( "ERROR in SGC_get_myenvp() : key = %d, file = %s, line = %d\n", (GC_arraysp->tls_key), __FILE__, __LINE__);
	    return 0;
	}
	return (GC_thread )p;
    }
#endif
}

void SGC_set_myenvp(GC_thread  envp)
{
#if defined SGC_SOLARIS_THREAD || \
    defined SGC_SOLARIS_THREAD_PTHREAD
    {
	int err_no = 
	    thr_setspecific((thread_key_t)(GC_arraysp->tls_key), (void *)envp);
	if(err_no == 0){
	    return;
	}else{
	    GC_err_printf3( "ERROR in SGC_set_myenvp() : file = %s, line = %d, errno = %d\n", __FILE__, __LINE__, err_no);
	    return;
	}
    }
#endif
#if defined SGC_PTHREAD
    {
	int err_no = 
	    pthread_setspecific((pthread_key_t)(GC_arraysp->tls_key), (void *)envp);
	if(err_no == 0){
	    return;
	}else{
	    GC_err_printf3( "ERROR in SGC_set_myenvp() : file = %s, line = %d, errno = %d\n", __FILE__, __LINE__, err_no);
	    return;
	}
    }
#endif
#if defined SGC_SPROC_THREAD
    {
	*((void **)&PRDA->usr_prda.fill[IRIX_PRDA_OFFSET]) =
	    (void *)envp;
	return;
    }
#endif
}

void SGC_tlskey_create(void)
{
#if defined SGC_SOLARIS_THREAD || \
    defined SGC_SOLARIS_THREAD_PTHREAD
    {
	int err_no = 
	    thr_keycreate((thread_key_t *)&(GC_arraysp->tls_key), 0);
	if(err_no == 0){
	    return;
	}else{
	    GC_err_printf3( "ERROR in SGC_tlskey_create() : file = %s, line = %d, errno = %d\n", __FILE__, __LINE__, err_no);
	    return;
	}
    }
#endif
#if defined SGC_PTHREAD
    {
	int err_no = 
	    pthread_key_create((pthread_key_t *)&(GC_arraysp->tls_key), 0);
	if(err_no == 0){
	    return;
	}else{
	    GC_err_printf3( "ERROR in SGC_tlskey_create() : file = %s, line = %d, errno = %d\n", __FILE__, __LINE__, err_no);
	    return;
	}
    }
#endif
#if defined SGC_SPROC_THREAD
    {
	/* do nothing */
	return;
    }
#endif
}

#endif /* SGC_TLS_IMPORT_xxx */
